-- vim:ts=4
-- ==========================================================================--
--								"AlphaDelta":								 --
--	 Elementary operations on Types (those which DO NOT REQUIRE EVALUATION): --
--   Alpha- and Delta- (AD-) Equivalence of Aldor Types, Vals, Exprs, DTs	 --
--		 and FBs, to be used in type-checking and evaluation engines:		 --
--						   L.A.Timochouk, 2001--2002						 --
-- ========================================================================= --
module AlphaDelta
	(
		ADE_Res (..),     ADE_Env,  mkADEnv,
		eqTypes,          eqExprs,  eqFBs,   AD_Equatable (adeq),
		elemTypeOf,		  findTypesEnclFPTargs,

		--	TC_Res is to be declared here so it is available for all Eval*
		--		and TypeCheck* modules without explicit qualification:
		TC_Res  (..),	  InferredType,		 noTYPE,      TCR_CL,

		CanBeBottom (..)
	)
where

import DataTypes
import NameBinds
import FButils

import List
import Maybe
import IOExts

modName :: String
modName = "AlphaDelta"


-- =================================================== --
--	The Alpha-Delta-Equivalence Checking Environment:  --
-- =================================================== --
type EqDefNos = [(DefNo, DefNo)]

data ADE_Env  = ADE_Env  {_binds    :: NameBinds,
						  _eqDefNos :: EqDefNos,
						  _remDepth :: Int,
						  _debugMode:: Bool}

--	 "ADE_Res" may involve NameBinds,  but they only make sense for a Positive
--		result (which may mean that the equivalence holds, but only under some
--		restricted binds):
--
data ADE_Res  = Bottom_ADER
			  | No_ADER
			  | Yes_ADER  NameBinds
				deriving  (Show)

-----------------
--	"mkADEnv": --
-----------------
mkADEnv:: NameBinds ->  Int-> Bool -> ADE_Env

mkADEnv       binds		max_ade_depth debug =
	ADE_Env {_binds		= binds,
			 _eqDefNos  = [],
			 _remDepth  = max_ade_depth,
			 _debugMode = debug}


----------------------------
--	"WithBinds" instance: --
----------------------------
instance WithBinds ADE_Env where
	getBinds  adenv 	  = _binds adenv
	setBinds  binds adenv = adenv {_binds = binds}


---------------------------
--	"ANDable"  instance: --
---------------------------
instance ANDable ADE_Res
	where
	--	Any negative arg -- the result is negative:
	(-&&-)   No_ADER     _				= No_ADER
	(-&&-)   _			 No_ADER		= No_ADER

	--	Then, any Bottom arg -- the result is Bottom:
	(-&&-)   Bottom_ADER _				= Bottom_ADER
	(-&&-)   _			 Bottom_ADER	= Bottom_ADER

	--	Two positives:   the result depends on crossing the binds:
	(-&&-)   (Yes_ADER   binds1) (Yes_ADER binds2)  =
		case consolidate binds1 binds2  of
			Just binds3 -> Yes_ADER binds3
			Nothing		-> No_ADER


-------------------------
--	"ORable" instance: --
-------------------------
instance ORable ADE_Res
	where
	--	Two "Yes" results -- the disjunction is "Yes", and we merge the binds:
	--
	(-||-)  (Yes_ADER binds1) (Yes_ADER binds2) = Yes_ADER (binds1 -||- binds2)

	--	Exactly one "Yes" result -- propagated "as is":
	(-||-)  (Yes_ADER binds1) _					= Yes_ADER  binds1
	(-||-)	_				  (Yes_ADER binds2) = Yes_ADER  binds2

	--	No "Yes" results but at least one "Bottom" -- "Bottom":
	(-||-)  Bottom_ADER		  _					= Bottom_ADER
	(-||-)	_				  Bottom_ADER		= Bottom_ADER

	--	Otherwise: two negatives: "No":
	(-||-)	_				  _					= No_ADER


-- ============ --
--	"eqExprs":  --
-- ============ --
--	Checks AD-equivalence   of two given Exprs,  that is, their structural
--		equivalence and the equivalence of all involved Identifiers modulo
--		the provided list of DefNos which are deemed to be equivalent:
--
eqExprs :: ADE_Env  -> AldorExpr -> AldorExpr -> ADE_Res

------------------
--	"VoidExpr": --
------------------
eqExprs    adenv VoidExpr  VoidExpr =  Yes_ADER (getBinds adenv)

------------------
--	"Literal":  --
------------------
eqExprs adenv (Literal  val1) (Literal val2) = eqVals adenv val1 val2

--------------------
--	"Identifier": --
--------------------
--	This is the ONLY point where NameBinds and EqDefs are used.  The names of
--		the Identifiers are not compared, only their bindings are. Again, the
--		def names of DefTarg(s) for this Identifier are ignored,    only type
--		exps and bodies are compared.
--	An Identifier, if it is "%", can be equivalent not only to another Identi-
--		fier, but to other exprs as well, as defined below:
--
eqExprs adenv (Identifier use_no1 qn1) rhs
	--
	--	Take the DISJUNCTION of the general and special cases:
	--
	| otherwise					   = (=||=) [resT, resP1, resP2]
	where
	binds :: NameBinds
	binds =  _binds  adenv

	targs1:: [BindTarget]
	targs1=  lookupBinds binds use_no1

	--
	--	The most common case: "rhs" is also a (general) Identifier:
	--
	resT  :: ADE_Res
	resT  =  case rhs of
		Identifier use_no2 _ ->
			let
				targs2 :: [BindTarget]
				targs2 =  lookupBinds binds use_no2
			in
				--	Of course, if "UseNo"s are the same, it's the same
				--		Identifier. Otherwise, compare the Targets:
				if   proper_eq  use_no1 use_no2
				then Yes_ADER  binds
				else eqLists_Targets adenv use_no1 use_no2 targs1 targs2

		_ ->
			--	If the "rhs" is not an Identifierm say "No" here but then try
			--		special cases:
			No_ADER

	--	NOW:
	--	There are extra rules for "%" Identifiers.
	--	
	--	First of all, any two "%"s must be made equivalent, otherwise no two
	--		Categories can be equivalent. This should not trigger false posi-
	--		tives for ADTs, since in that case we also compare the "Rep"s...
	--
	resP1 :: ADE_Res
	resP1 =  case rhs of
		Identifier _ qn2 ->
			if   qn1 == ["%"] && qn2 == ["%"]
			then Yes_ADER binds
			else No_ADER
		_ ->
			--	This case is not applicable:
			No_ADER

	--	Also, the "%" of an ADT-valued expr (but NOT of a Category-valued one)
	--		is equivalent to that expr:
	--
	--	Pair the "targs" with all ADT-valued exprs they correspond to:
	--
	texprs:: [(AldorExpr, BindTarget)]
	texprs=  foldl getADTexpr [] targs1

	getADTexpr:: [(AldorExpr, BindTarget)] -> BindTarget  ->
				 [(AldorExpr, BindTarget)]

	getADTexpr   curr   t@(PerCentTarg e)  =
		case e of
			Literal (TypeVal (ADT _ _ _)) -> (e,t) : curr
			Add1 _ _					  -> (e,t) : curr
			Add2 _ _ _ 					  -> (e,t) : curr
			_							  -> curr

	getADTexpr   curr   (AddedTarg t)     =  getADTexpr curr t
	getADTexpr   curr   _				  =  curr

	--	Select those "texprs" (if any) which are AD-equivalent to the "rhs":
	--	We actually only need their Targets:
	--
	ress  :: [(ADE_Res, BindTarget)]
	ress  =  map eqExprs2   texprs

	eqExprs2  ::(AldorExpr, BindTarget) -> (ADE_Res, BindTarget)
	eqExprs2    (e, t) =    (eqExprs (adenv {_binds =binds'}) e rhs,  t)
		where
		binds':: NameBinds
		binds' = modifyBinds(_binds adenv) use_no1 [t]

	--	Now select "Yes" and "Bottom" results:
	--
	yess  :: [(ADE_Res, BindTarget)]
	yess  =  [(ader, t) | (ader@(Yes_ADER _), t) <- ress]

	bots  :: [(ADE_Res, BindTarget)]
	bots  =  [(ader, t) | (ader@(Bottom_ADER),t) <- ress]

	--	If there is at least one "yess" element, the result is "Yes"
	--		with restricted binds:
	resP2 :: ADE_Res
	resP2 =  case yess of
		_:_   ->
			let
				ytargs :: [BindTarget]
				ytargs =  [t | (_, t) <- yess]

				ybinds :: NameBinds
				ybinds =  modifyBinds (_binds adenv) use_no1 ytargs
			in
				Yes_ADER  ybinds

		[]	  ->
			--	Otherwise, if there is at least one Bottom, the result
			--		is Bottom, otherwise No:
			case bots of
				_:_ -> Bottom_ADER
				[]  -> No_ADER

	--	That's all!

--
--	And, of course, the equivalence is symmetrical:
--
eqExprs  adenv lhs ident@(Identifier _ _) = eqExprs adenv ident lhs


-----------------
--	"AndList": --
--   "OrList": --
-----------------
--	We use "eqLists" here, i.e. the lists may be equivalent up to a permutation
--		(due to commutativity of the operations and the absense of side-effects
--		in the evaluation of operands), but in order to avoid a factorial comp-
--		lexity of the test, an assumption is made (in "eqLists") that there are
--		no equivalent elements within "fbs2":
--	Empty lists cannot occur here -- so we can reat them as non-equivalent:
--
eqExprs adenv (AndList fbs1) (AndList fbs2) =
	eqLists adenv False eqFBs fbs1 fbs2

eqExprs adenv (OrList  fbs1) (OrList  fbs2) =
	eqLists adenv False eqFBs fbs1 fbs2

--------------------
--	"IfThenElse": --
--------------------
eqExprs adenv (IfThenElse cond1 then1 else1) (IfThenElse cond2 then2 else2) =
	(eqFBs adenv  cond1  cond2) -&&-
	(eqFBs adenv  then1  then2) -&&- 
	(eqFBs adenv  else1  else2)

----------------------
--	"FunctionCall": --
----------------------
--	Here, of course, actual parameters must be compared  in their actual order,
--		but unlike "eqFProtos", the environment for comparison is the same for
--		all FBs involved.
--	For calls to well-known built-in functions, commutativity and associativity
--		laws are applied (so e.g. "1+x" and "x+1" will be considered to be equ-
--		ivalent), but with the following limitations:
--
--
--	First of all, special cases:
--
--eqExprs adenv (FunctionCall
--			 (SimpleFB _ (Literal (BuiltInOp "+" _ _))) )

eqExprs adenv (FunctionCall func_fb1 arg_fbs1)(FunctionCall func_fb2 arg_fbs2) =
	(=&&=) [eqFBs adenv fb1  fb2  |
					   (fb1, fb2)<- zip  (func_fb1:arg_fbs1)(func_fb2:arg_fbs2)]

---------------------
--	"RecordCtor":  --
---------------------
--	Similar to "FunctionCall":
--
eqExprs adenv (RecordCtor fields1)  (RecordCtor fields2) =
	(=&&=) [eqFBs adenv fb1 fb2  |  (fb1, fb2)<- zip fields1 fields2]

-------------------
--	"UnionCtor": --
-------------------
--	The tag expr and the contents must be equivalent  for "UnionCtor"s to be
--		equivalent:
--
eqExprs adenv (UnionCtor tag_expr1 cont_fb1) (UnionCtor   tag_expr2 cont_fb2) =
	(eqExprs  adenv tag_expr1  tag_expr2) -&&-
	(eqFBs    adenv cont_fb1    cont_fb2)

--------------------
--	"UnionCase":  --
--------------------
--	The Unions themselves and the tags (selectors) are compared:
--
eqExprs adenv (UnionCase union1 tag1) (UnionCase union2 tag2) =
	(eqFBs    adenv      union1 union2) -&&-
	(eqExprs  adenv		 tag1   tag2  )

----------------------
--	"RestrictType": --
--	   "ConvertTo": --
----------------------
eqExprs adenv (RestrictType fb1 to_type1)    (RestrictType fb2 to_type2)    =
	(eqFBs adenv fb1 fb2)  -&&- (eqFBs adenv to_type1 to_type2)

eqExprs adenv (ConvertTo    fb1 to_type1)    (ConvertTo    fb2 to_type2)    =
	(eqFBs adenv fb1 fb2)  -&&- (eqFBs adenv to_type1 to_type2)

-----------------------------------------
--	"Add1", "Add2", "With1", "With2":  --
-----------------------------------------
--	These exprs are enumerated, and this helps us to decide equality:
--
eqExprs adenv (Add1   lno lfb)       (Add1  rno rfb)
	| proper_eq lno rno = Yes_ADER (getBinds  adenv)
	| otherwise 		=  eqFBs adenv lfb  rfb


eqExprs adenv (Add2   lno lfb1 lfb2) (Add2  rno rfb1 rfb2)
	| proper_eq lno rno = Yes_ADER (getBinds adenv)
	| otherwise			= (eqFBs adenv lfb1 rfb1) -&&- (eqFBs adenv lfb2 rfb2)


eqExprs adenv (With1  lno lfb)	   (With1 rno rfb)
	| proper_eq lno rno = Yes_ADER (getBinds adenv)
	| otherwise			=  eqFBs adenv lfb  rfb


eqExprs adenv (With2  lno lfb1 lfb2)   (With2 rno rfb1 rfb2)
	| proper_eq lno rno = Yes_ADER (getBinds adenv)
	| otherwise			= (eqFBs adenv lfb1 rfb1) -&&- (eqFBs adenv lfb2 rfb2)


-----------------------
--	All other cases: --
-----------------------
--	Exprs are not equal:
eqExprs _ _ _ = No_ADER


-- =========== --
--	"eqVals":  --
-- =========== --
--	AD-equivalence of "AldorValue"s:
--
eqVals:: ADE_Env  -> AldorValue -> AldorValue -> ADE_Res

-------------------------
--	Elementary Values: --
-------------------------
eqVals   adenv TrivVal		    TrivVal		    =  Yes_ADER  (getBinds adenv)

eqVals   adenv (BooleanVal b1) (BooleanVal b2)  =
	if   b1 == b2
	then Yes_ADER (getBinds adenv)
	else No_ADER

eqVals   adenv (IntegerVal i1) (IntegerVal i2)  =
	if   i1 == i2
	then Yes_ADER (getBinds adenv)
	else No_ADER

eqVals   adenv (FloatVal   f1) (FloatVal   f2)  =
	if   f1 == f2
	then Yes_ADER (getBinds adenv)
	else No_ADER

eqVals   adenv (StringVal  s1) (StringVal  s2)  =
	if   s1 == s2
	then Yes_ADER (getBinds adenv)
	else No_ADER

------------------
--	"TypeVal"s: --
------------------
eqVals adenv (TypeVal t1)  (TypeVal t2)     =  eqTypes adenv t1 t2

--------------------
--	"LambdaVal"s: --
--------------------
--	If their protos match, the bodies are to be compared modulo the original
--		"eq_defnos" PLUS the equivalence relations between  the formal parms
--		defined in those protos.  (The protos are themselves compared modulo
--		gradually increasing "eq_defnos", and the resulting "eq_defnos'" are
--		re-used here):
--
eqVals adenv  (LambdaVal proto1 body1) (LambdaVal proto2 body2) =
	protosOK -&&- (eqFBs adenv' body1  body2)
	where
	(protosOK, eq_defnos')    = eqFProtos   adenv proto1 proto2
	adenv' = adenv {_eqDefNos = eq_defnos'}

--------------------
--	"BuiltInOp"s: --
--------------------
--	BuilnIn functions do not have named formal parms,  so their bodies can be
--		compared modulo the original "eq_defnos". But in fact body comparison
--		is not even necessary:   a built-in function is completely defined by
--		its name and type,  so if they are the same, the bodies should be the
--		same, too:
--
eqVals adenv (BuiltInOp name1 type1 _) (BuiltInOp name2 type2 _) =
	if   name1 == name2
	then eqTypes  adenv type1 type2
	else No_ADER

--------------------
--	"RecordVal"s: --
--------------------
--	Just compared field-by-field, as RecordVals do not carry any explicit type
--		information:
--
eqVals adenv (RecordVal fields1) (RecordVal fields2) =
	if   length fields1 == length fields2
	then (=&&=) [eqVals adenv f1 f2 | (f1,f2) <-zip fields1 fields2]
	else No_ADER

-------------------
--	"UnionVal"s: --
-------------------
--	There is no full UnionType information stored in UnionVals; however, there
--		is a UseNo (a unique system-wide tag)  identifying the field currently
--		contained. So we first compare tags, and then their values.  If any of
--		the tags is "noNUMBER", the comparison FAILS since we don't know which
--		field is referred to:
--
eqVals adenv (UnionVal sel1 cont1) (UnionVal sel2 cont2) =
	(eqExprs adenv     sel1  sel2) -&&-
	(eqVals  adenv    cont1 cont2)

------------------
--	"ADT_Val"s: --
------------------
--	The corresp ADTs must be equivalent themselves, and so must be the values:
--
eqVals adenv (ADT_Val type1 val1 )      (ADT_Val  type2 val2) =
	(eqTypes adenv    type1 type2) -&&- (eqVals   adenv val1 val2)

-----------------------
--	All other cases: --
-----------------------
--	Vals are not equal:
eqVals _ _ _ = No_ADER


-- =========== --
--	"eqTypes": --
-- =========== --
--	AD-equivalence of "AldorType"s:
--
eqTypes :: ADE_Env  -> AldorType-> AldorType -> ADE_Res

------------------------
--	Elementary Types: --
------------------------
eqTypes adenv  TrivT	TrivT	 =  Yes_ADER  (getBinds adenv)
eqTypes adenv  ExitT	ExitT	 =  Yes_ADER  (getBinds adenv)
eqTypes adenv  TypeT	TypeT	 =  Yes_ADER  (getBinds adenv)
eqTypes adenv  BooleanT BooleanT =  Yes_ADER  (getBinds adenv)
eqTypes adenv  IntegerT IntegerT =  Yes_ADER  (getBinds adenv)
eqTypes adenv  FloatT   FloatT   =  Yes_ADER  (getBinds adenv)
eqTypes adenv  StringT  StringT  =  Yes_ADER  (getBinds adenv)

--------------------
--	"FunctionT"s: --
--	"RecordT"s  : --
--	"UnionT"s   : --
--------------------
--	Each of them is actually a sequence of FBlocks, from which type exprs are
--		to be extracted and compared in order,  and the "eq_defnos" are to be
--		modified in this process:   "eqFPLists"/"eqFProtos"  do that, but the
--		resulting new "eq_defnos" is discarded here:
--
eqTypes adenv proto1@(FunctionT _ _) proto2@(FunctionT _ _)   =
	fst (eqFProtos adenv proto1   proto2)

eqTypes adenv (RecordT   fields1) (RecordT fields2) =
	fst (eqFPLists adenv fields1  fields2)

eqTypes adenv (UnionT    fields1) (UnionT  fields2) =
	fst (eqFPLists adenv fields1  fields2)

--------------
--	"ADT"s: --
--------------
--	For ADTs to be equivalent, both Reps and Interfaces must be equivalent.
--	Enumerators help us to by-pass complex checks:
--
eqTypes  adenv (ADT no1 rep_fb1 dt1) (ADT no2 rep_fb2 dt2) =
	if   proper_eq  no1  no2
	then Yes_ADER   (getBinds adenv)
	else (eqFBs adenv rep_fb1 rep_fb2) -&&- (eqDTs adenv dt1 dt2)

--------------------
--	"CategoryT"s: --
--------------------
--	Just equivalence of DTs:
--
eqTypes  adenv    (CategoryT no1 dt1) (CategoryT no2 dt2) =
	if   proper_eq no1 no2
	then Yes_ADER (getBinds  adenv)
	else eqDTs  adenv  dt1   dt2


-----------------------
--	All other cases: --
-----------------------
--	Types are not equal:
eqTypes _ _ _ = No_ADER


-- ========== --
--	"eqFBs":  --
-- ========== --
--	AD-equivalence of FBlocks. We use enumerators whenever possible:
--
eqFBs :: ADE_Env -> AldorFBlock -> AldorFBlock ->  ADE_Res

-----------------
--	"VoidFB"s: --
-----------------
--	They were not supposed to occur at all (the case of two "VoidFB"s as inner
--		FBs in "NestedFB"s was treated explicitly elsewhere) until the Declara-
--		tions (without Def Bodies) were implemented. The case of two "VoidFB"s
--		is now perfectly acceptable:
--
eqFBs adenv (VoidFB _) (VoidFB _) = Yes_ADER (getBinds adenv)

-------------------
--	"SimpleFB"s: --
-------------------
eqFBs adenv (SimpleFB fb_no1  expr1 _)   (SimpleFB fb_no2 expr2 _) =
	if   proper_eq    fb_no1  fb_no2
	then Yes_ADER (getBinds   adenv)
	else eqExprs  adenv expr1 expr2

-------------------
--	"NestedFB"s: --
-------------------
--	Consider explicitly the case when both "innerFB"s are "VoidFB"s (just
--		to eliminate that separate equation, which may mask errors):
--
eqFBs adenv (NestedFB fb_no1 dt1 inner1 _) (NestedFB fb_no2 dt2 inner2 _) =
	if   proper_eq    fb_no1 fb_no2
	then Yes_ADER  (getBinds adenv)
	else dt_res -&&- inner_res
	where
	dt_res, inner_res     :: ADE_Res
	dt_res   =  eqDTs  adenv dt1    dt2
	inner_res=  eqFBs  adenv inner1 inner2

-------------------
--	"FParmFB"s:  --
-------------------
-- Comparison of "FParmFB"s  implemented here  is different from that in
--		"eqFPLists"/"eqFProtos": here the general "eqDTs" is used, which
--		takes the Names into account as well (whereas the former compare
--		type exprs only):
--	XXX: How can this case occur at all? For "FParmFB"s in "FP_Targ"s, a
--		different comparison function is used ("eqFPNo"s);  here we give
--		a warning:
--
eqFBs adenv fb1@(FParmFB  fb_no1 dt1 _)  fb2@(FParmFB fb_no2 dt2 _)  =

	debug_msg adenv ("\nNOTE: \"eqFBs\" called on FParmFBs: "++
					(show fb1)++" and "++(show fb2)++"\n")

	(	if  proper_eq fb_no1 fb_no2
		then Yes_ADER (getBinds   adenv)
		else eqDTs adenv dt1 dt2
 	)

-----------------------
--	All other cases: --
-----------------------
--	FBs are not equal:
eqFBs _ _ _ = No_ADER


-- ============== --
--	"eqFPLists":  --
-- ============== --
--	AD-equivalence  of lists  of FBlocks  ("FParmFB"s only!) over gradually-
--		increasing "eq_defnos". Only type exprs, not names are actually com-
--		pared.  The resulting "eq_defnos" are returned along with the compa-
--		rison result   (but they make sense only if the result was positive,
--		otherwise (No_ADER, []) or (Bottom_ADER, []) is returned):
--
eqFPLists ::    ADE_Env  -> [AldorFBlock] -> [AldorFBlock] ->
				(ADE_Res,   EqDefNos)

eqFPLists adenv []  []   = (Yes_ADER (getBinds adenv), _eqDefNos adenv)

eqFPLists _     ((FParmFB _ _ _) :_)  []  =  (No_ADER, [])
eqFPLists _     []  ((FParmFB _ _ _) :_)  =  (No_ADER, [])

eqFPLists adenv ((fb1@(FParmFB _ _ _)): fbs1) ((fb2@(FParmFB _ _ _)) : fbs2) =
	--
	--	Catch lists of different lengths immediately: in that case there is
	--		no need to compare their elements:
	--
	if length fbs1 == length fbs2
	then
		case  ((eqExprs  adenv texpr1 texpr2) -&&- tailOK)  of
			Bottom_ADER	     -> (Bottom_ADER,  [])
			yes@(Yes_ADER _) -> (yes, eq_defnos'')
			No_ADER			 -> (No_ADER,	   [])
	else
		(No_ADER, [])
	where
	eq_defnos' = (def_no1, def_no2)  : (_eqDefNos adenv)
	(tailOK,  eq_defnos'') =  eqFPLists adenv{_eqDefNos=eq_defnos'} fbs1 fbs2
	--
	--	eq_defnos'' are to be returned; they ontain the "longest" equivalence
	--		list, up to the end of the protos being compared

	GenFDTE def_no1 _ texpr1 _ = getFParm fb1
	GenFDTE def_no2 _ texpr2 _ = getFParm fb2

eqFPLists _ list1 list2 =
	error2 "eqFPLists"
		   ("Not an FParmFB encountered:\n"++(show list1)++"\n"++(show list2))


-- ============= --
--	"eqFProtos": --
-- ============= --
--	AD-equivalence of function types, using "eqFPLists". Different from
--		"eqTypes" on "FunctionT"   in that "eqFProtos" also returns the
--		modified "EqDefs" (in fact, "eqTypes--FunctionT" is implemented
--		via "eqFProtos"):
--
eqFProtos ::  ADE_Env   -> AldorType -> AldorType -> (ADE_Res, EqDefNos)

eqFProtos adenv (FunctionT parms1 retype1) (FunctionT parms2 retype2) =
	case  protosOK of
		Bottom_ADER		-> (Bottom_ADER, [])
		yes@(Yes_ADER _)-> (yes, eq_defnos')
		No_ADER			-> (No_ADER,	 [])
	where
	protosOK			 = parmsOK -&&-  retypesOK
	(parmsOK, eq_defnos')= eqFPLists adenv	parms1  parms2
	retypesOK			 = eqFBs     adenv{_eqDefNos=eq_defnos'} retype1 retype2

eqFProtos _ type1 type2  =
	error2 "eqFProtos"
		  ("Not a FunctionT encountered:\n"++(show type1)++"\n"++(show type2))


-- ========= --
--	"eqDTs": --
-- ========= --
--	AD-equivalence of DefTables.   They are converted into lists and compared
--		using "eqFDTEs", i.e. names declared are also used in comparison. Im-
--		port lists are also compared for equivalence:
--	"True" in "eqFDTEs" means that def names ARE taken into account:
--
eqDTs :: ADE_Env -> DefTable -> DefTable -> ADE_Res

eqDTs adenv dt1 dt2 =
	--	Empty lists should not occur for FDTEs, but can occur for Imports, and
	--		in the latter case the Imports are equivalent:
	--
	(eqLists adenv False (eqFDTEs True) fdtes1   fdtes2  ) -&&-
	(eqLists adenv True   eqExprs       imports1 imports2)
	where
	--	Convert "dt1" and "dt2" into lists of "FlatDTE"s:
	--
	fdtes1,   fdtes2  :: [FlatDTE]
	fdtes1 =  foldDT (:) [] dt1
	fdtes2 =  foldDT (:) [] dt2

	--	Get the lists of Imports:
	--
	imports1, imports2:: [AldorExpr]
	imports1= getImports dt1
	imports2= getImports dt2


-- ============= --
--	"eqLists*":  --
-- ============= --
--	Used for testing AD-equality of lists of various BindTargets. The logic of
--		list processing is the same, only comparison of actual Targets is dif-
--		ferent (given by the first parameter).
--	To reduce the temporal complexity of this function, we assume that  "list2"
--		itself does not contain any AD-equivalent elements, ie  each element of
--		"list1" can be equivalent to no more than 1 element of "list2".    When
--		such an equivalent element is found for a given element of "list1", the
--		former is removed from "list2", and so on (see "eqLists_R" below):
--
------------------------
--	"eqLists": Strong --
------------------------
--	For the lists to be strongly equivalent,   they must be of the same length,
--		and the list of equivalent pairs produced by "eqLists_R" must be of the
--		same length, and the binds of all these pairs must be compatible.
--	The Boolean flag indicates whether two empty lists are considered to be
--		equivalent (this depends on the caller's logic):
--
eqLists:: ADE_Env -> Bool ->
		 (ADE_Env -> a-> a-> ADE_Res) -> [a] -> [a] -> ADE_Res

eqLists   adenv emptyEq _ [] [] =
	if    emptyEq
	then  Yes_ADER (getBinds adenv)
	else  No_ADER

eqLists   adenv  _ eqFunc list1 list2 =
	if    length list2 == n
	then
		if  length eqs == n
		then
			--	All equivalent pairs constructed -- intersect the binds:
			case consolidateList [binds | (_, _, binds) <- eqs] of
				Just cbinds ->   Yes_ADER cbinds
				Nothing		->	 No_ADER
		else
			--	They are simply not equal, or there were Bottoms?
			if  bot_flag
			then
				Bottom_ADER
			else
				No_ADER
	else
		--	"list1" and "list2" are of different lengths:
		No_ADER
	where
	n				= length list1
	(eqs, bot_flag) = eqLists_R adenv eqFunc [] False list1 list2


-------------------------
--	"eqLists_Targets": --
-------------------------
--	Used for AD-equivalence testing of targets of Identifiers with the given
--		"use_no1" and "use_no2".  Before targets are compared, the binds for
--		the corresp "use_no"s are restricted to those targets. The result is
--		"Yes" if there was at least one successful test,  otherwise "Bottom"
--		if there was at least one Botton, otherwise "No". This is a WEAK com-
--		parison. For "Yes", the binds returned  are the UNION  of restricted
--		binds  under which the successful results were obtained.
--	NB: Two empty lists are ALWAYS considered to be NON-EQUIVALENT, since we
--		need explicit matches for merging the binds   (otherwise binds would
--		be too large after merge as described above):
--
eqLists_Targets :: ADE_Env -> UseNo -> UseNo -> [BindTarget] -> [BindTarget] ->
				   ADE_Res

eqLists_Targets    _ _ _ [] []   =  No_ADER

eqLists_Targets    adenv   use_no1  use_no2  targs1  targs2 =
	case eqs of
		[] ->
			--	No equivalent elements found!
			if   bot_flag
			then Bottom_ADER
			else No_ADER

		_  ->
			--	There are equivalent elements,   merge their binds and return
			--		"Yes". XXX: can we assume, for efficiency, that the binds
			--		only differ in targs for "use_no1" and "use_no2", to opti-
			--		mise merging? -- For now, we don't do that:
			--
			Yes_ADER ((=||=) [binds | (_, _, binds) <- eqs])
	where
	eqs      :: [(BindTarget, BindTarget, NameBinds)]
	bot_flag :: Bool

	(eqs, bot_flag) =  eqLists_R adenv eqTargs_R   []  False targs1 targs2

	--
	--	Modified equivalence test which restricts binds:
	--
	eqTargs_R:: ADE_Env -> BindTarget -> BindTarget -> ADE_Res
	eqTargs_R   adenv'     targ1         targ2		=
		eqTargs(restrBinds2 adenv' targ1 targ2)  targ1 targ2

	--
	--	"restrBinds2":
	--
	restrBinds2 :: ADE_Env -> BindTarget -> BindTarget -> ADE_Env
	restrBinds2    adenv'     targ1         targ2      =  setBinds binds2 adenv'
		where
		binds1,  binds2 ::  NameBinds
		binds1 = modifyBinds (getBinds adenv') use_no1 [targ1]
		binds2 = modifyBinds binds1			   use_no2 [targ2]


-------------------
--	"eqLists_R": --
-------------------
--	An internal function, common core of "eqLists" and "eqLists_Weak".  Its
--		result is a list of found equivalent pairs, if any, along with Name
--		Binds under which their equivalence was decided, and a flag indica-
--		ting whether a Bottom was encountered:
--
eqLists_R:: ADE_Env -> (ADE_Env  -> a    ->  a  -> ADE_Res)   ->
		 -- adenv	   eqFunc
			 [(a, a, NameBinds)] -> Bool -> [a] -> [a] ->
		 --  curr_equals		bot_flag	list1  list2
			([(a, a, NameBinds)],   Bool)
		 -- (res_equals         res_bot_flag)

eqLists_R   adenv  eqFunc   curr_eqs  bot_flag (head1:tail1) list2  =  res
	where
	--	"analyse2":
	--
	--	Finds whether there is an element in "list2" which is equivalent to
	--		"head1", and whether there are elements in "list2" whose equiva-
	--		lence  to "head1" is undecided. All elements which are NOT equi-
	--		valent to "head1" are returned as well:
	--
	--	analyse2 :: [a] -> (Maybe (a, NameBinds)) -> Bool -> [a] ->
	--				(Maybe (a, NameBinds), Bool,    [a])
	--				(Equivalent,           Bottom,  NotEquiv)
	--	To reduce the number of comparisons (which may be recursive and
	--		therefore expensive),   once an equivalent element is found,
	--		all others are not tested for equivalence anymore  (so some
	--		Bottoms can be missing as a result, but that's OK):
	--
	analyse2 []            mb_eq bot_flag2 others =
		(mb_eq, bot_flag2, others)
	-- All Done!

	analyse2 (head2:tail2) mb_eq bot_flag2 others =
	-- Generic case...
		case mb_eq of
			Nothing ->
				--	No equivalent element to "head1" was found so far in
				--		"list2", so do the test on "head2":
				--
				case eqFunc adenv head1 head2 of
					Bottom_ADER	->
						--	Set the Bottom flag and continue; "head2" is
						--		moved to "others":
						analyse2 tail2 mb_eq     True  (head2:others)

					No_ADER     ->
						--	Move "head2" to "others", but don't change the
						--		Bottom flag:
						analyse2 tail2 mb_eq bot_flag2 (head2:others)

					Yes_ADER ybinds ->
						--	Found, but continue to fill in "others":
						analyse2 tail2 (Just (head2, ybinds)) bot_flag2 others

			Just _ ->
				--	Move "head2" to "others", as an element equivalent to
				--		 "head2" has already been found:
				analyse2 tail2 mb_eq bot_flag2 (head2:others)

	--	Now the tail-recursive call of "eqLists_R":
	--	Calling "analyse2" with the current our "bot_flag" and Nothing for
	--		equivalent elements of "head1":
	--
	res = case  analyse2 list2 Nothing bot_flag [] of
		(Just (match2, ybinds), bot_flag2, others) ->
			--	Equivalent element found, "others" are non-equivalent ones,
			--		to be compared with "tail1":
			eqLists_R adenv eqFunc ((head1, match2, ybinds):curr_eqs) bot_flag2
					  tail1 others

		(Nothing, bot_flag2, others) ->
			eqLists_R adenv eqFunc curr_eqs bot_flag2 tail1 others


eqLists_R _ _ curr_eqs bot_flag [] _ = (curr_eqs, bot_flag)
	--	LHS is finished -- all done!


------------------------------------------------
--	Extra comparison functions for "eqLists": --
------------------------------------------------
-----------------
--	"eqTargs": --
-----------------
eqTargs :: ADE_Env -> BindTarget -> BindTarget -> ADE_Res

--
--	First of all, Targs with the same ctors:
--
eqTargs    adenv  (DefTarg fdte1) (DefTarg fdte2) =
	--	Hmm... Do we compare them with or without names? Do WITHOUT:
	eqFDTEs False adenv    fdte1  fdte2

eqTargs     adenv (FP_Targ fp1)   (FP_Targ fp2)   =
	eqFPNos adenv (defno1, fbno1) (defno2, fbno2)
	where
	GenFDTE defno1 _ _ _ = getFParm  fp1
	GenFDTE defno2 _ _ _ = getFParm  fp2
	fbno1				 = getFBNo_U fp1
	fbno2				 = getFBNo_U fp2


eqTargs     adenv (PerCentTarg e1) (PerCentTarg e2) =
	eqExprs adenv e1 e2

eqTargs    adenv (BI_Targ bino1)  (BI_Targ bino2)  =
	if     proper_eq   bino1   bino2
	then   Yes_ADER  (getBinds adenv)
	else   No_ADER

eqTargs    adenv (DynTarg qnm1 t1) (DynTarg qnm2 t2) =
	if     qnm1 == qnm2
	then   eqTargs adenv t1 t2
	else   No_ADER

eqTargs    adenv (RepTarg  fb1)   (RepTarg  fb2)   = eqFBs   adenv fb1 fb2

eqTargs    adenv (CondTarg fb1)   (CondTarg fb2)   = eqFBs   adenv fb1 fb2

eqTargs    adenv (AddedTarg t1)   (AddedTarg t2)   = eqTargs adenv t1  t2

--
--	Now: Targs with different ctors may also be equivalent:
--
eqTargs    adenv t1 (AddedTarg    t2) = eqTargs adenv t1 t2
eqTargs    adenv (AddedTarg t1)   t2  = eqTargs adenv t1 t2

--
--	"CondTarg"s should not occur at all:
--
eqTargs _ _ t@(CondTarg _)   = error2 "eqTargs" ((show t)++" encountered")
eqTargs _   t@(CondTarg _) _ = error2 "eqTargs" ((show t)++" encountered")

--
--	All other combinations: Negative result:
--
eqTargs _ _ _ = No_ADER


-----------------
--	"eqFDTEs": --
-----------------
--	Comapres two FDTEs. If their DefNos are the same, they are equivalent.  If
--		not, they may still be equivalent  if the recursive equivalence checks
--		are allowed   (remaining recursion depth is > 0)   and their names and
--		structural components are themselves equivalent. The def names are NOT
--		taken into into account if the "withNames" arg is "False":
--	This function is used for deciding AD-equivalence of DefTables and DefTargs:
--
eqFDTEs:: Bool -> ADE_Env -> FlatDTE -> FlatDTE -> ADE_Res

eqFDTEs   withNames adenv (GenFDTE def_no1 name1 type_expr1 body_fb1)
			              (GenFDTE def_no2 name2 type_expr2 body_fb2)

	| proper_eq def_no1 def_no2 = Yes_ADER (getBinds adenv)

	| _remDepth adenv >= 1 =
		if  withNames && name1 /= name2
		then
			No_ADER
		else
			(eqExprs adenv' type_expr1 type_expr2) -&&-
			(eqFBs   adenv' body_fb1   body_fb2)

	| otherwise = Bottom_ADER
	where
	adenv'= adenv{_remDepth= (_remDepth adenv) - 1}


eqFDTEs  withNames adenv (RenFDTE def_no1 name1 ren_id1)
			   			 (RenFDTE def_no2 name2 ren_id2)

	| proper_eq  def_no1 def_no2 = Yes_ADER  (getBinds adenv)

	| _remDepth adenv >= 1 =
		if   withNames && name1 /= name2
		then No_ADER
		else eqExprs adenv' ren_id1 ren_id2

	| otherwise = Bottom_ADER
	where
	adenv' = adenv{_remDepth= (_remDepth adenv) - 1}


eqFDTEs _ _ _ _ = No_ADER


-----------------
--	"eqFPNo"s: --
-----------------
--	Used to compare "FParmFB"s.  Only enumerators are taken into account, not
--		even types (as two function parms with identical type exprs are still
--		entirely different!):
--
eqFPNos:: ADE_Env -> (DefNo, FBNo ) -> (DefNo, FBNo ) -> ADE_Res
eqFPNos   adenv      (defno1,fbno1)    (defno2,fbno2) =
	if    proper_eq    fbno1 fbno2
	then  --   "DefNo"s must be equal as well!
		if     proper_eq   defno1 defno2
		then   Yes_ADER (getBinds adenv)
		else   error2 "eqFPNos" "Integrity error"
	else
		No_ADER


-- ======================= --
--	"AD_Equatable" Class:  --
-- ======================= --
--	Provides the "adeq" equivalence test which generalises "eqTypes", "eqExprs"
--		and "eqFBs":
--
class    AD_Equatable a where
	adeq :: ADE_Env-> a -> a -> ADE_Res

instance AD_Equatable AldorType   where
	adeq = eqTypes

instance AD_Equatable AldorExpr   where
	adeq = eqExprs

instance AD_Equatable AldorFBlock where
	adeq = eqFBs


-- ===================== --
--	Auxiliary Functions: --
-- ===================== --
--------------------
--	"proper_eq":  --
--------------------
proper_eq :: Int -> Int -> Bool
proper_eq    no1    no2 =  (no1 /= noNUMBER) && (no1 == no2)


-- =============== --
--	"elemTypeOf":  --
-- =============== --
--	Simple mapping of Values to Types; produces "Nothing" if the type cannot be
--		determined    (for Records and Unions, type inference can yield invalid
--		results -- e.g. a too general type of a dependent Record):
--
elemTypeOf::  AldorValue    -> Maybe AldorType

elemTypeOf    TrivVal       =  Just TrivT
elemTypeOf   (TypeVal    _) =  Just TypeT
elemTypeOf   (BooleanVal _) =  Just BooleanT
elemTypeOf   (IntegerVal _) =  Just IntegerT
elemTypeOf   (FloatVal   _) =  Just FloatT
elemTypeOf   (StringVal  _) =  Just StringT
elemTypeOf   (LambdaVal    proto@(FunctionT _ _) _) = Just proto
elemTypeOf   (BuiltInOp  _ proto _)				    = Just proto

--	"ADT_Val"s have type information built-in:
--
elemTypeOf   (ADT_Val adt _)=  Just adt

--	For all others, return Nothing:
elemTypeOf    _	=  Nothing


-- ========================= --
--	"findTypesEnclFPTargs":  --
-- ========================= --
--	Given  a UseNo and NameBinds, this function searches for all definitions of
--		that UseNo which are given by FParmFBs.    It then returns pairs (Type,
--		Index) where Type is a "complex" type (e.g. FunctionT, RecordT, UnionT)
--		and Index is the position (counted from 0) of the target FParmFB in the
--		FB list contained in that type.
--	Therefore, this is in fact an elementary type-checking function which, like
--		"elemTypeOf", doesn't perform any evaluations, so it can be called from
--		within the Evaluation Engine without the risk of infinite recursion:
--
findTypesEnclFPTargs :: NameBinds -> UseNo -> [(BindTarget, AldorType, Int)]

findTypesEnclFPTargs    binds        use_no = res
	where
	--	Get all targets of this "use_no" which are "FP_Targ"s, and return the
	--		corresp "FParmFB"s:
	--
	fparms  :: [(BindTarget, AldorFBlock)]
	fparms  =  [(targ, fparm_fb) |
				targ@(FP_Targ fparm_fb) <- lookupBinds binds use_no]

	--	Determine the "complex" "AldorType"s (currently "RecordT"s, "UnionT"s
	--		and "FunctionT"s) to which each of the "ftargs" belongs:
	--
	t3s     :: [Maybe (BindTarget, AldorType, Int)]
	t3s     =  [findEnclType  (getFBNo fparm) targ  fparm |
				(targ, fparm) <- fparms]

	res     :: [(BindTarget,  AldorType, Int)]
	res     =  catMaybes      t3s

	--
	--	"findEnclType":
	--
	--	Initially invoked on an FParmFB. Goes up-link until NOT an FParmFB is
	--		encountered; VoidFB must not occur in this process. Once a Simple
	--		or NestedFB is encountered, it is searched inside for the Complex
	--		Type containing the original FParmFB (whose FBNo is carried along
	--		by this function). "Nothing" is returned if the type is not found:
	--
	findEnclType  :: FBNo -> BindTarget -> AldorFBlock ->
				  -- origFBNo  origTarg    currFB
					 Maybe  (BindTarget,   AldorType, Int)

	findEnclType origFBNo     targ  (FParmFB _ _ uplink) =
		findEnclType origFBNo targ  uplink
	--
	--	Hmm... In fact, "FParmFB"s CAN be PHYSICALLY nested (not only connected
	--		in a scoping chain by up-links), so it MAY be OK to stop here at an
	--		"FParmFB", but we have no way to know whether it's indeed the case,
	--		so proceeed upwards until we encounter non-"FParmFB"...

	findEnclType _  _ hmm@(VoidFB _)    =
		error2 "findEnclType" ((show hmm)++" encountered")

	findEnclType origFBNo  targ otherFB  =
		--
		--	Search the entire contents   of the "otherFB" for a TypeVal Literal
		--		containing the FParmFB with the "origFBNo". The search is lazy,
		--		and it stops as soon as the first result is found since we know
		--		there can be no more than 1 result.
		--
		case snd (foldFB_Gen passFB findByFBNo (undef,[]) otherFB) of
			(t,i) : _ -> Just     (targ, t, i)
			[]		  -> Nothing
		where
		--
		--  "findByFBNo" is a call-back on all AldorExprs encountered by
		--		"foldFB_Gen"; its 1st parm is to be ignored:
		--
		findByFBNo  :: AldorFBlock-> AldorExpr -> [(AldorType, Int)] ->
					   							  [(AldorType, Int)]

		findByFBNo  _ (Literal (TypeVal t@(UnionT  fields)))    curr =
			scanFields curr fields t

		findByFBNo  _ (Literal (TypeVal t@(RecordT fields)))    curr =
			scanFields curr fields t

		findByFBNo  _ (Literal (TypeVal t@(FunctionT parms _))) curr =
			scanFields curr parms  t

		findByFBNo  _ _     curr   = curr

		scanFields ::  [(AldorType, Int)] -> [AldorFBlock] ->  AldorType ->
					   [(AldorType, Int)]

		scanFields    curr  fields  t   =
			--
			--	"t" is what we need iff its "fields" contain FB with "origFBNo":
			--
			case elemIndex origFBNo (map getFBNo fields)  of
				Just fb_index ->    (t, fb_index) : curr
				_			  ->    curr


-- ================ --
--	"CanBeBottom":  --
-- ================ --
class    CanBeBottom  a  where
	isBottom :: a -> Bool


instance CanBeBottom ADE_Res where
	isBottom Bottom_ADER =   True
	isBottom _           =   False


-- =========== --
--	"TC_Res":  --
-- =========== --
--	NB: "TC_Res" data are produced by the Type-Checking Engine.  However, as
--		we do limited type-checking during evaluation, it is defined here.
--	We can do A LOT  of type inference  while type-checking exprs  without a
--		known target.   The inferred type maybe unknown or given by  (in the
--		general case) an Aldor Expression. The (=&&=) operator needs to keep
--		inferred types of the components,  hence the list   in the following
--		definition. In all cases, the list is non-empty:
--
type     InferredType = [Maybe AldorExpr]

noTYPE:: InferredType
noTYPE = [Nothing]


--	Now "TC_Res" itself:
--	NB: All type-checking functions actually return [TC_Res];  [] indicates
--		failure of the check:
--
data TC_Res =
	 -- Check runs to completion and succeeds, yielding more specific binds
	 --		and (possibly) the type inferred:
	 --
	 OK_TCR		 {_bindsR  :: NameBinds,
				  _infType :: InferredType}

	 --	The result is undecided  because a recursion bottom  was hit in the
	 --		process (in the Eval or AD engine or elsewhere). Still get more
	 --		specific binds in this case   (as we know the binds under which
	 --		the bottom was hit):
	 --
   | Bottom_TCR  {_bindsR  :: NameBinds}


------------------------------
--	"WithAttr" on "TC_Res": --
------------------------------
--	This is required to allow "crossing" of "TC_Res"s. Currently, the attr of
--		"TC_Res" is just "NameBinds":
--
instance WithAttr NameBinds TC_Res
	where
	getAttr tcr		  = _bindsR tcr
	setAttr binds tcr =  tcr {_bindsR = binds}

type TCR_CL = ConsoList  NameBinds TC_Res


--------------------------
--	"Show" on "TC_Res": --
--------------------------
instance Show TC_Res
	where
	show (Bottom_TCR binds)		 =
		"_|_,\nBinds => "++(show binds) ++"\n"

	show (OK_TCR binds inf_type) =
		"OK, \nBinds => "++(show binds) ++",\n"++
	    "Inferred Type =>\n"      ++
	    (case inf_type of
			[Just type_expr] -> show type_expr
			[Nothing]		 -> "None"
			hmm 	 		 -> "Incomplete: "++(show hmm)
	    )++"\n"

------------------------
--	"Eq" on "TC_Res": --
------------------------
--	XXX: For the moment, it's simply the string equility on "show" results!
--
instance Eq TC_Res
	where
	(==) tcr1 tcr2 = (show tcr1) == (show tcr2)


-- ============= --
--	"error2"   : --
--	"debug_msg": --
-- ============= --
error2 ::String -> String -> a
error2 = errorM    modName

debug_msg :: ADE_Env -> String -> a -> a
debug_msg    adenv      msg     val =
	if    _debugMode adenv
	then  trace  msg val
	else  val

