-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--							     "DecorFB.hs":								 --
--	  Decoration of Aldor Functional Blocks with Enumerators and UpLinks     --
--						   L.A.Timochouk, 2000--2001						 --
-- ========================================================================= --
module DecorFB
	   (
			Enums3, decorFBlock, ckFBconn,
			ckFBNo, ckDefNo, ckUseNo, ckBINo
	   )
where

import DataTypes
import FButils

modName :: String
modName = "DecorFB"


-- ================ --
--	"decorFBlock*": --
-- ================ --
--	(*) All Definitions and Identifiers used in expressions are assigned valid
--		sequence numbers;
--	(*) Parent References are provided for all FBlocks:
--
--	NB: We could re-use the "foldFB_Gen" function in the "decorFBlock" impleme-
--		ntation, but since the former does not maintain the "parent" ("uplink")
--		ref by itself, there would be little advantages of doing so:
--
type Enums3  = (FBNo, DefNo, UseNo)

decorFBlock  :: AldorFBlock->			-- FBlock to be decorated
				Enums3	   ->			-- Initial enums to be used
			   (AldorFBlock, Enums3)	-- Result: (Decorated FB, Max Enums + 1)

decorFBlock_R:: AldorFBlock->			-- Parent
				Enums3     ->			-- Current (FBNo,  DefNo,  UseNo)
				AldorFBlock->			-- FBlock to be decorated
			   (AldorFBlock,  Enums3)	-- (Decortd_FBlock,
			   							--	       (FBNo', DefNo', UseNo'))

decorFBlock fb init_enums3  =
	decorFBlock_R (VoidFB fbno)  enums3' fb
	where
	(fbno, defno, useno)  = init_enums3
	enums3' :: Enums3
	enums3' =  (fbno+1,  defno,  useno)


--------------------------------------
--	"decorFBlock_R" implementation: --
--------------------------------------
--	"NestedFB" decoration:
--	The DefTable is processed first, then the inner FBlock:
--
decorFBlock_R parent (fb_no,  def_no,  use_no) (NestedFB _ dt fb _) =
			(thisFB, enums3')
	where
	thisFB			= NestedFB      fb_no   dt'   innrFB   parent
	(dt',   enums3t)= decorDT       thisFB  (fb_no+1, def_no, use_no)  dt
	(innrFB,enums3')= decorFBlock_R thisFB  enums3t   fb

--
--	"SimpleFB" decoration:
--	The expression is processed recursively:
--
decorFBlock_R parent (fb_no, def_no, use_no ) (SimpleFB  _ expr _) =
			(thisFB, enums3')
	where
	thisFB			= SimpleFB  fb_no   expr'  parent
	(expr',enums3') = decorExpr thisFB (fb_no+1, def_no, use_no)  expr

--
--	"VoidFB": Still assign an FBNo to it:
--
decorFBlock_R _ enums3 (VoidFB _) = (VoidFB fbno, enums3')
	where
	(fbno, defno, useno)  = enums3
	enums3' :: Enums3
	enums3' =  (fbno+1, defno, useno)

--
--	"FParmFB":
--	We decorate its singleton DefTable:
--
decorFBlock_R parent  (fb_no, def_no, use_no) (FParmFB _ dt _) =
	(thisFB, enums3')
	where
	thisFB        =   FParmFB  fb_no dt' parent
	(dt',enums3') =   decorDT  thisFB   (fb_no+1, def_no, use_no)  dt


-------------------
--	"decorExpr": --
-------------------
--	Parms and Result:
--		parentFB    -> (currEnums3) -> undecorExpr -> (decorExpr, newEnums3)
--
decorExpr ::   AldorFBlock-> Enums3 -> AldorExpr   -> (AldorExpr, Enums3)

--	"VoidExpr":   --
decorExpr  _   enums3 VoidExpr = (VoidExpr, enums3)


--	"Literal" :   --
decorExpr  parent enums3 (Literal val) = (Literal val', enums3')
	where
	(val', enums3') =  decorVal parent enums3 val


--	"Identifier":   --
--	Modifies "use_no":
decorExpr  _   (fb_no, def_no,  use_no) (Identifier  _ str) =
	(Identifier use_no str,     (fb_no,  def_no,  use_no+1))


--	"AndList": --
decorExpr  parent  enums3 (AndList fbs) =  (AndList dfbs, enums3')
	where
	(dfbs, enums3') = decorFBList parent enums3 fbs


--	"OrList" : --
decorExpr  parent  enums3 (OrList  fbs) =  (OrList  dfbs, enums3')
	where
	(dfbs, enums3') = decorFBList parent enums3 fbs


--	"IfThenElse":  --
decorExpr  parent  enums3 (IfThenElse condFB thenFB elseFB) =
	(IfThenElse  dfb1  dfb2  dfb3, enums3')
	where
	(dfb1, enums3t) = decorFBlock_R parent  enums3  condFB
	(dfb2, enums3e) = decorFBlock_R parent  enums3t thenFB
	(dfb3, enums3') = decorFBlock_R parent  enums3e elseFB


--	"FunctionCall": --
decorExpr  parent  enums3 (FunctionCall func args) =
	(FunctionCall  dfunc  dargs,  enums3')
	where
	(dfunc, enums3f)= decorFBlock_R parent enums3  func
	(dargs, enums3')= decorFBList   parent enums3f args


--	"RecordCtor":
decorExpr  parent  enums3 (RecordCtor fields) =
	(RecordCtor fields', enums3')
	where
	(fields', enums3') = decorFBList parent enums3 fields 


--	"UnionCtor":
decorExpr  parent  enums3 (UnionCtor sel val) =
	(UnionCtor  sel' val', enums3')
	where
	(sel', enums3s) = decorExpr		parent enums3  sel
	(val', enums3') = decorFBlock_R parent enums3s val


--	"UnionCase":
decorExpr  parent  enums3 (UnionCase un sel)  =
	(UnionCase  un'  sel', enums3')
	where
	(un',  enums3u) = decorFBlock_R parent enums3  un
	(sel', enums3') = decorExpr     parent enums3u sel


--	"RestrictType": --
decorExpr  parent  enums3 (RestrictType expr_fb type_fb) =
	(RestrictType  expr_fb' type_fb',   enums3')
	where
	(expr_fb', enums3e) =  decorFBlock_R parent enums3  expr_fb
	(type_fb', enums3') =  decorFBlock_R parent enums3e type_fb


--  "ConvertTo":  --
decorExpr  parent  enums3 (ConvertTo    expr_fb type_fb) =
	(ConvertTo     expr_fb' type_fb',   enums3')
	where
	(expr_fb', enums3e) =  decorFBlock_R parent enums3  expr_fb
	(type_fb', enums3') =  decorFBlock_R parent enums3e type_fb


--	"Add*":
decorExpr  parent  enums3 (Add1 _ fb) =  (Add1 def_no fb', enums3')
	where
	(fb_no, def_no, use_no) = enums3
	(fb', enums3')      =  decorFBlock_R parent (fb_no, def_no+1, use_no) fb

decorExpr  parent  enums3 (Add2 _ fb1 fb2) =  (Add2  def_no fb1' fb2', enums3'')
	where
	(fb_no, def_no, use_no) = enums3
	(fb1', enums3' )    =  decorFBlock_R parent (fb_no, def_no+1, use_no) fb1
	(fb2', enums3'')    =  decorFBlock_R parent enums3' fb2

--	"With*":
decorExpr  parent  enums3 (With1 _ fb) =  (With1 def_no fb', enums3')
	where
	(fb_no, def_no, use_no) = enums3
	(fb', enums3')      =  decorFBlock_R parent (fb_no, def_no+1, use_no) fb

decorExpr  parent  enums3 (With2 _ fb1 fb2) = (With2 def_no fb1' fb2', enums3'')
	where
	(fb_no, def_no, use_no) = enums3
	(fb1', enums3' )    =  decorFBlock_R parent (fb_no, def_no+1, use_no) fb1
	(fb2', enums3'')    =  decorFBlock_R parent enums3' fb2


------------------
--	"decorVal": --
------------------
--	This is non-trivial, as it might contain a TypeVal  with nested FBlocks
--		to be decorated (e.g. for a "FunctionT", a "LambdaVal", and so on):
--
decorVal:: AldorFBlock -> Enums3  -> AldorValue -> (AldorValue, Enums3)
--		   parentFB	  currEnums3     undecorVal    (decorVal,newEnums3)

--
--	"TypeVal":
--
decorVal   parent  enums3  (TypeVal tv) = (TypeVal tv', enums3')
	where
	(tv', enums3') = decorType   parent   enums3   tv

--
--	"LambdaVal":
--
decorVal   parent  enums3  (LambdaVal  proto  body) =
						   (LambdaVal dproto dbody, enums3')
	--
	--	NB: the parent for the lambda body is the last parm FB, if the
	--		function has at least one parm,  and "parent" otherwise:
	--	Functions without any parameters are disallowed, so the body's
	--		uplink is always the last of "dparms":
	where
	(dproto, enums3t)   =  decorFProto  parent  enums3  proto
	(dbody,  enums3')   =  decorFBlock_R bparent enums3t body

	--	Extract "dparms"  from "dproto":
	FunctionT dparms _       =  dproto
	bparent = last dparms		-- Not "parent" directly!

--
--	"RecordVal":
--
decorVal  parent  enums3  (RecordVal fields) = (RecordVal fields', enums3')
	where
	(fields', enums3') =  decorValList parent  enums3 fields

--
--	"UnionVal":
--	"UnionVal"s can only be created INTERNALLY (by the Evaluation Engine),
--		and inherit the field selector from the Union Ctor, so this "sel"
--		is not affected by decoration. The contained value may still need
--		to be decorated (???):
--
decorVal  parent  enums3  (UnionVal sel val) =
	(UnionVal sel val', enums3')
	where
	(val', enums3')    =  decorVal  parent enums3 val

--
--	"ADT_Val":
--	HMM... This is tricky: ADT_Val contains an ALIAS of the FBlock enclosing
--		the ADT body itself, so we need to install an alias to the decorated
--		version of that FBlock -- but where to find it? We cannot just deco-
--		rate the FBlock given here, as it would then get enumerators  diffe-
--		rent from the original decorated ADT FBlock.   However, there are no
--		statically-defined "ADT_Val"s (which would require decoration)  any-
--		way; they can only be created during the evaluation process,   so we
--		disallow this case altogether:
--
decorVal  _ _ (ADT_Val _ _) =
	error2 "decorVal--ADT_Val" "Should not occur at all!"

--
--	All other "AldorValue"s are trivial:
--
decorVal  _ enums3  any_val  = (any_val, enums3)


-------------------
--  "decorType": --
-------------------
decorType:: AldorFBlock -> Enums3 -> AldorType -> (AldorType, Enums3)
--			parentFB   currEnums3  undecorType    (decorType, newEnums3)

--
--	"FunctionT":
--
decorType   parent enums3 (proto@(FunctionT _ _)) =  (proto', enums3')
	where
	(proto',  enums3')  = decorFProto  parent  enums3 proto

--
--	"RecordT":
--
decorType   parent enums3 (RecordT fields) = (RecordT fields', enums3')
	where
	(fields', enums3')  = decorFParms  parent  enums3 fields
	--	Record fields are treated in the same way  as function parms!

--
--	"UnionT":
--	Unlike "RecordT", "UnionT" cannot be internally-dependent, so the uplink of
--		each field is just the "parent":
--
decorType   parent enums3 (UnionT  fields) = (UnionT fields', enums3')
	where
	(fields', enums3')  = decorFBList parent  enums3 fields

--
--	"ADT":
--
decorType   parent enums3 (ADT _ rep_fb dt) = (dADT, enums3'')
	where
	--	Install the current "def_no" in the new ADT:
	(fb_no, def_no, use_no) = enums3

	dADT :: AldorType
	dADT  = ADT def_no drep_fb dt'

	(dt', enums3') = decorDT  parent (fb_no, def_no+1, use_no)  dt

	(drep_fb,  enums3'')    = decorFBlock_R  parent enums3' rep_fb


--
--	"CategoryT": Like ADT:
--
decorType  parent enums3 (CategoryT _ dt) = (new_cat, enums3')
	where
	--	Install the current "def_no" in the new Category:
	(fb_no, def_no, use_no) = enums3

	new_cat :: AldorType
	new_cat =  CategoryT def_no dt'

	(dt', enums3') = decorDT parent (fb_no, def_no+1, use_no) dt

--
--	Others: do nothing:
--
decorType _ enums3 any_type = (any_type, enums3)


---------------------
--	"decorFBList": --
---------------------
--	Used in e.g. decorating "AND", "OR", "UnionT" expressions:
--	NB: using "reverse" with "foldl" rather than "foldr" is preferred here, at
--		it keeps the resulting FBlocks enumeration consistent  with their phy-
--		sical order ("foldl" alone would reverse the latter):
--
decorFBList  :: AldorFBlock -> Enums3 -> [AldorFBlock]-> ([AldorFBlock],Enums3)
--				parentFB   currEnums3     undecorFBs     (decorFBs,  newEnums3)

decorFBList     parent  enums3  fbs = (rfbl, enums3r)
	where
	(fbl, enums3r) = foldl action ([], enums3) fbs
	rfbl		   = reverse fbl

	action:: ([AldorFBlock], Enums3)  ->  AldorFBlock -> ([AldorFBlock],Enums3)
	action   (dfbs, enums3') fb = (dfb:dfbs, enums3'')
		where
		(dfb, enums3'') = decorFBlock_R  parent enums3' fb


----------------------
--	"decorValList": --
----------------------
--	Decorates a list of "AldorValue"s whose internal FBlocks  (if any) share a
--		common "parent";
--
decorValList :: AldorFBlock -> Enums3 -> [AldorValue] -> ([AldorValue], Enums3)
--				parentFB   currEnums3   undecorVals      (decorVals, newEnums3)

decorValList    parent  enums3 vals = foldl  action ([], enums3) vals
	where
	action:: ([AldorValue], Enums3)-> AldorValue -> ([AldorValue], Enums3)

	action   (vals', enums3')   val = (val'':vals', enums3'')
		where
		(val'', enums3'') = decorVal parent enums3' val


------------------
--	"decorDT":  --
------------------
--	Parms and Result:
--	NB: "parentFB" is the FB which contains the DT in question.  For function
--		bodies, the parent is the last parmFB if exists,  "parent" otherwise.
--	The Imports Table is also decorated:
--
decorDT :: AldorFBlock ->  Enums3 -> DefTable   -> (DefTable, Enums3)
--		   parentFB    currEnums3   undecorDT      (decorDT,  newEnums3)

decorDT parent enums3 dt = res2
	where
	--
	--	First decorate the DefTable entries themselves:
	--
	res1:: (DefTable, Enums3)
	res1 = foldDT decorFDTE_R (emptyDT, enums3)  dt

	decorFDTE_R:: FlatDTE  -> (DefTable, Enums3) -> (DefTable, Enums3)

	decorFDTE_R  (GenFDTE _  name  type_expr  body_fb)
				 (dt1, (fb_no, def_no, use_no))  =  (dt2,      enums3')
		where
		dt2    =   attachFDTE  dt1 (GenFDTE  def_no name dtype_expr dbody_fb)

		--	Increment "def_no" here to account for this DTE:
		(dtype_expr, enums3t) =
			decorExpr	   parent (fb_no, def_no+1, use_no) type_expr

		(dbody_fb,   enums3') =
			decorFBlock_R bparent  enums3t body_fb

		bparent =
			case dtype_expr  of
				 Literal (TypeVal (FunctionT dparms _))  -> last dparms
					--  Functions without parms are not allowed, so the uplink
					--		here is always the last of "dparms", not  "parent"

				 _	->	parent   -- In all other cases

	decorFDTE_R  (RenFDTE _ name ren_expr) (dt1, (fb_no, def_no, use_no)) =
				 (dt2, enums3')
		where
		dt2   =  attachFDTE  dt1	(RenFDTE def_no name dren_expr)

		(dren_expr, enums3') = decorExpr parent (fb_no,def_no+1,use_no) ren_expr

	--
	--	Then decorate the domain imports:
	--
	res2:: (DefTable, Enums3)
	res2 = foldl decorImport res1 (getImports dt)

	decorImport :: (DefTable, Enums3) -> AldorExpr -> (DefTable, Enums3)
	decorImport    (  cdt,   cenums3)    imp       =
		(attachImport cdt    dimp,   enums3')
		where
		(dimp, enums3') =    decorExpr parent cenums3 imp


---------------------
--	"decorFProto": --
---------------------
decorFProto::  AldorFBlock  ->  Enums3 -> AldorType -> (AldorType,     Enums3)
--			   parentFB		currEnums3    undecorProto (decorProto, newEnums3)

decorFProto    parent  enums3  (FunctionT parms retype) =
	(FunctionT parms'' retype', enums3')
	where
	--	Decorate the parms first:
	(parms'',  enums3p) = decorFParms parent enums3 parms

	--	The retype FB is linked to last parms'' (which are already properly
	--		reversed), so is the function body, see "decorDT" above.  It is
	--		assumed that there are no functions without parms at all, so the
	--		"retype" is never uplinked directly to the "parent":
	--		to the "parent":
	--
	(retype',  enums3') = decorFBlock_R  rparent enums3p retype
	rparent				= last parms''


decorFProto _ _ hmm = error2
					  "decorFProto" ("Type is not FunctionT: "++ (show hmm))

---------------------
--	"decorFParms": --
---------------------
--	Used in "decorFProto", but also in decorating "Record" types etc:
--
decorFParms:: AldorFBlock -> Enums3 -> [AldorFBlock] -> ([AldorFBlock], Enums3)
--			  parentFB	 currEnums3    undecorParms     (decorParms, newEnums3)

decorFParms   parent enums3  parms  =  (parms'', enums3p)
	where
	--	"parms" are to be uplinked from right to left, and the left-most
	--		one is linked  to the "parent".  To achieve this, we have to
	--      use  "foldl" to get the uplinks in the correct order, and to
	--		reverse the resulting dparms to get them "physically" in the
	--		right order. Note that "foldr" would not do;  the rules are:
	--	"foldl": links = orig, phys =~orig
	--	"foldr": links =~orig, phys = orig
	--
	parms''              = reverse   parms'
	(parms', _, enums3p) = foldl decorFParm ([], parent, enums3)  parms

	decorFParm:: ([AldorFBlock], AldorFBlock, Enums3) -> AldorFBlock ->
				 ([AldorFBlock], AldorFBlock, Enums3)
	--  Carried: (newParms,      currUpLink,  currEnums3)

	decorFParm   (dparms, uplink, curr3) parm = (dparm:dparms, dparm, new3)
		where
		(dparm, new3) = decorFBlock_R uplink curr3 parm


-- =============== --
--	"checkFBconn": --
-- =============== --
--	Checks the consistency of up- and down-link connectivity of FBlocks.

same_as :: AldorFBlock -> AldorFBlock -> Bool

same_as   (VoidFB   lno)       (VoidFB	 rno)		= lno == rno
same_as   (SimpleFB lno _ _  ) (SimpleFB rno _ _  ) = lno == rno
same_as   (NestedFB lno _ _ _) (NestedFB rno _ _ _) = lno == rno
same_as   (FParmFB  lno _ _  ) (FParmFB  rno _ _  ) = lno == rno
same_as   _                     _                   = False


--	The parent of all FBlocks is typically "VoidFB 0", but the FBNo may be
--		different if we are checking a later-decorated extension,  so this
--		FBNo must be passed explicitly:
--
ckFBconn :: AldorFBlock -> FBNo -> Bool
ckFBconn    top_fb         voidNo  = ckFBconnR top_fb (VoidFB voidNo)


ckFBconnR:: AldorFBlock -> AldorFBlock -> Bool
--			known_parent   thisFB

ckFBconnR   _ (VoidFB _) = True
--	VoidFB  does not have an explicit parent, but occurs inside many blocks,
--		so we consider it to be "always connected"


ckFBconnR   parent  thisFB@(NestedFB  _   dt innerFB uplink)
	| parent  `same_as`  uplink =
	  (ckDT thisFB dt)  && (ckFBconnR thisFB innerFB)

	| otherwise =  errFBconn  parent  thisFB

ckFBconnR   parent  thisFB@(SimpleFB _ expr uplink)
	| parent  `same_as`  uplink   = ckExpr    thisFB expr
	| otherwise					  = errFBconn parent thisFB


ckFBconnR   parent thisFB@(FParmFB  _ _ uplink)
	| parent  `same_as`  uplink   = ckFDTE    thisFB (getFParm thisFB)
	| otherwise					  = errFBconn parent thisFB


--------------
--	"ckDT": --
--------------
ckDT :: AldorFBlock -> DefTable ->  Bool

ckDT parent dt  = res1 && res2
	where
	--	Check the DefTable entries themselves:
	res1  :: Bool
	res1  =  foldDT action True  dt

	action:: FlatDTE  -> Bool -> Bool
	action   fdte curr=  curr && (ckFDTE  parent fdte)

	--	Check the list of imported domains:
	res2  :: Bool
	res2  =  foldl impAction True (getImports dt)

	impAction :: Bool -> AldorExpr -> Bool
	impAction    curr    imp       =  curr && (ckExpr parent imp)

----------------
--	"ckExpr": --
----------------
ckExpr:: AldorFBlock -> AldorExpr -> Bool

ckExpr parent (Literal val) = ckVal  parent val

ckExpr parent (AndList fbs) = and (map (ckFBconnR parent) fbs)

ckExpr parent (OrList  fbs) = and (map (ckFBconnR parent) fbs)

ckExpr parent (IfThenElse condFB thenFB elseFB)  =
			   (ckFBconnR parent condFB) &&
			   (ckFBconnR parent thenFB) &&
			   (ckFBconnR parent elseFB)

ckExpr parent (FunctionCall  _ fbs) = and (map (ckFBconnR parent) fbs)

ckExpr parent (RecordCtor   fields) = and (map (ckFBconnR parent) fields)

ckExpr parent (UnionCtor   sel val) =
	(ckExpr parent sel)  &&(ckFBconnR parent val)

ckExpr parent (UnionCase   un  sel) =
	(ckFBconnR parent un)&&(ckExpr    parent sel)

ckExpr parent (RestrictType exprFB typeFB) =
			   (ckFBconnR   parent exprFB) &&
			   (ckFBconnR   parent typeFB)

ckExpr parent (ConvertTo    exprFB typeFB) =
			   (ckFBconnR   parent exprFB) &&
			   (ckFBconnR   parent typeFB)

ckExpr parent (Add1 _  fb)	= ckFBconnR parent fb

ckExpr parent (Add2 _  fb1 fb2) =
			  (ckFBconnR parent fb1) && (ckFBconnR parent fb2)

ckExpr parent (With1 _ fb)  = ckFBconnR parent fb

ckExpr parent (With2 _ fb1 fb2) =
			  (ckFBconnR parent fb1) && (ckFBconnR parent fb2)

ckExpr _ _  = True -- For all other expressions: nothing to check


----------------
--  "ckVal":  --
----------------
ckVal :: AldorFBlock -> AldorValue -> Bool

ckVal  parent (TypeVal tv) =  ckType  parent tv

ckVal  parent (LambdaVal proto body) = ckLambda parent proto body

ckVal  parent (RecordVal fields)     = foldl  ckField True fields
	where
	ckField:: Bool -> AldorValue -> Bool
	ckField   curr    field      =  curr &&  (ckVal parent field)

ckVal  parent (UnionVal _   val) =  ckVal  parent val
--	No need to check the 1st "UnionVal" component

ckVal  _      (ADT_Val  _ _)     =
	error2   "ckVal--ADT_Val" "Should not occur at all!"
	--	Like decorVal on ADT_Val...

ckVal  _ _  = True	-- All other "AldorValue"s are trivial


----------------
--	"ckType": --
----------------
ckType:: AldorFBlock -> AldorType  -> Bool

ckType parent (proto@(FunctionT _ _))  = ckFProto parent proto

ckType parent (RecordT fields)		   = ckFParms parent fields
	--  Record Fields are treated like Function Parms!

ckType parent (UnionT  fields)		   = and (map  (ckFBconnR parent) fields)
	--	Unlike "RecordT", all "fields" have their uplinks
	--		pointing directly to the "parent"

ckType parent (ADT _ rep_fb dt) = (ckFBconnR parent rep_fb) && (ckDT parent dt)

ckType parent (CategoryT _ dt ) =  ckDT parent dt

ckType _ _ =  True  -- All other "AldorType"s  are trivial


-----------------
--	"ckFDTE":  --
-----------------
--	For Function  types, the connection graph  must be:
--		parent <- parm[0] <- ... <- parm[N] <- body_fb
--											<- retype_fb
--	The same for Record and Union types.
--
--	For other definitions, "type_expr" and "body_fb" have the same "parent",
--		and can be verified independently:
--
ckFDTE:: AldorFBlock  -> FlatDTE ->  Bool

ckFDTE   parent (GenFDTE  _ _ type_expr body_fb)  =
	case type_expr   of
		Literal (TypeVal proto @(FunctionT _ _)) ->
			ckLambda parent proto  body_fb
		_ 										 ->
			(ckExpr  parent type_expr) && (ckFBconnR parent body_fb)

ckFDTE  parent (RenFDTE  _ _ ren_expr)   =     ckExpr    parent ren_expr


------------------------------------
--	Checking FProtos and Lambdas: --
------------------------------------
------------------
--	"ckLambda": --
------------------
--	We rely upon the fact that there is always at least 1 function parameter:
--
ckLambda ::  AldorFBlock -> AldorType -> AldorFBlock -> Bool
--			 Parent			LamType  	 LamBodyFB

ckLambda      parent proto   @ (FunctionT parms _) body =
	(ckFProto parent proto) && (ckFBconnR bparent  body)
	where
	bparent = last parms	-- Not "parent" directly!

ckLambda _ hmm _ = error2 "ckLambda" ("Type is not FunctionT: "++(show hmm))


------------------
--	"ckFProto": --
------------------
--  We rely upon the fact that there is always at least 1 function parameter:
--
ckFProto  :: AldorFBlock -> AldorType -> Bool

ckFProto     parent  (FunctionT parms retype) = resP && resR
	where
	resP  =  ckFParms    parent parms
	resR  =  ckFBconnR  rparent retype
	rparent  = last parms   --  Not "parent" directly!

ckFProto _ hmm = error2 "ckFProto" ("Type is not FunctionT: "++(show hmm))


------------------
--	"ckFParms": --
------------------
--	Used by "ckFPoto" and also by "ckType" on e.g. "RecordT":
--
ckFParms  :: AldorFBlock -> [AldorFBlock] -> Bool

ckFParms  parent parms =    snd    (foldl  ckParm  (parent, True) parms)
	--
	--     Use "foldl" here to check the parms in the same order as they
	--			were decorated:
	where
	ckParm ::  (AldorFBlock, Bool) -> AldorFBlock -> (AldorFBlock,Bool)
	--
	--	NB: here "thisFB" must always be an "FParmFB":
	--
	ckParm (left,   stat)    thisFB @(FParmFB _ _ _) =
		   (thisFB, stat && (ckFBconnR left thisFB))

	ckParm  _ _  =  error2 "ckParm" "Not an \"FParmFB\""


--------------------
--	"errFBconn":  --
--------------------
--	If disconnectivity was encountered:
--
errFBconn:: AldorFBlock -> AldorFBlock -> Bool
errFBconn   parent         thisFB      =
	  error2 "ckFBconn"
			 ("FBlock No."  ++(show (getFBNo parent)) ++ ":\n\n" ++
			 (show  parent) ++"\n\nis not uplink from FBlock No."++
			 (show (getFBNo thisFB))++":\n\n" ++(show    thisFB) ++
			  "\n\nas it should be. The actual (wrong) uplink is FBlock No."++
			 (show (getFBNo realUp))++":\n\n" ++(show    realUp) ++"\n")
	  where
	  realUp = getUpLinkFB thisFB


-- ========================================== --
--	Checking enumerators for not "noNUMBER":  --
-- ========================================== --
--  Return the original No if it is valid, or error otherwise:
--
ckFBNo :: FBNo  -> FBNo
ckFBNo    fb_no =
	if    fb_no  /= noNUMBER then fb_no  else error2 "ckFBNo"  "noNUMBER"

ckDefNo:: DefNo -> DefNo
ckDefNo   def_no=
	if    def_no /= noNUMBER then def_no else error2 "ckDefNo" "noNUMBER"

ckUseNo:: UseNo -> UseNo
ckUseNo   use_no=
	if	  use_no /= noNUMBER then use_no else error2 "ckUseNo" "noNUMBER"

ckBINo :: BINo  -> BINo
ckBINo    bi_no =
	if	  bi_no  /= noNUMBER then bi_no  else error2 "ckBINo"  "noNUMBER"


-- =========== --
--  "error2":  --
-- =========== --
error2:: String-> String -> a
error2 = errorM modName

