-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--								 "EvalLib.hs":								 --
--																			 --
--			Data types and helper functions for the "EvalX" module.			 --
--		The purpose of "EvalLib" is basically to keep the size of "EvalX"	 --
--						   within reasonable limits.						 --
--																			 --
--							L.A.Timochouk, 2000--2002						 --
-- ========================================================================= --
module EvalLib
	   (
			ETC_Env (..),  EvalRes  (..),  EvalRes_CL,

			FBorDT  (..),  TypeOrFB (..),  Proto_Res (..), FBList_Res (..),
			DT_Res  (..),  FDTE_Res (..),				   ER_HomoList(..),
			ADT_Res (..),  OpenedDynTarg (..),

			EvalStackFrame (..),    EvalStack,
			pushES,   pushES_EE,    exportES,    exportES_EE,
			lookupES, lookupES_EE,

			EvalCacheKey   (..),    EvalCache,   toCache,  fromCache,

			openDynTarget,  getUltADTrep,

			setPartialFBNo, homogenise_ER_List,

			fb2er, er2fb,   er2expr,   er2adtr,  evalFB_2DT,
			updateTCE,		updateEE,  etc2ade,  fbdt2fb,

			decr_rem_depth, newEnum,

			tcExpr_T,		tcFB_T,
			weaklyConforms, stronglyConforms,    conforms,

			debug_msg,		(!!!)
	   )
where

import DataTypes
import ParseAST
import FButils
import NameBinds
import AlphaDelta

import Maybe
import Data.FiniteMap
import IORef
import List
import IOExts

modName :: String
modName =  "EvalLib"


-- ==================================== --
-- Evaluation Environment and Results:  --
-- ==================================== --
-----------------
--	"EvalRes": --
-----------------
--	The result of an expr or FBlock evaluation:
--	Evaluation result is NOT just "AldorValue". In the presence of names over-
--		loading, an evaluation result can only be obtained   in the context of
--		given NameBinds. The latter are modified in the course  of  evaluation
--		as some bindings may fail and would be removed.
--	Alternatively, an expr/FB may only be partially evaluatable  (e.g. if they
--		refer to formal function parms). In this case the result is in general
--		an FBlock, again with appropriate binds.
--	So:
--
data    EvalRes    =
	Bottom_ER	NameBinds				-- Evaluation failed (undefined result)
  | Partial_ER  NameBinds AldorFBlock	-- Partially-evaluated result
  | Full_ER     NameBinds AldorValue	-- Fully-evaluated result


type    EvalRes_CL = ConsoList  NameBinds  EvalRes
--
--	Since "EvalRes" is a "WithAttr" type...


------------------------------
--	Auxiliary result types: --
------------------------------
data  FBorDT     =
	  Bottom_FBDT   NameBinds
	| FB1		    NameBinds AldorFBlock
	| DT1		    NameBinds DefTable

data  TypeOrFB   =
	  Bottom_TFB    NameBinds
	| Type2		    NameBinds AldorType
	| FB2    	    NameBinds AldorFBlock

data  Proto_Res  =
	  Bottom_PR     NameBinds
	| OK_PR         NameBinds  AldorType

data  FBList_Res =
	  Bottom_FBL    NameBinds
	| OK_FBL        NameBinds [AldorFBlock]

data  DT_Res	 =
	  Bottom_DTR    NameBinds
	| OK_DTR        NameBinds DefTable

data  FDTE_Res	 =
	  Bottom_FDTER  NameBinds
	| OK_FDTER      NameBinds FlatDTE

data  OpenedDynTarg=
      Bottom_ODT    NameBinds
	| OK_ODT        NameBinds BindTarget
	deriving (Show)

--	The following types do NOT contain NameBinds:

data  ER_HomoList=
	  Bottom_ERL
	| Partial_ERL   [AldorFBlock]
	| Full_ERL      [AldorValue]

data  ADT_Res    =
	  OK_ADT        AldorType -- Must be an ADT
	| OK_DT         DefTable


-------------------
--	"EvalStack": --
-------------------
--	Here function call arguments are stored. The stack is a list whose elements
--		are parameters for each nested call, stored as maps "DefNo => EvalRes",
--		where DefNo is that of the corresponding formal parm  and "EvalRes"  is
--		the respective (single-valued) actual arg  along with the NameBinds for
--		which is was computed:
--
newtype EvalStackFrame =  EvalStackFrame (FiniteMap DefNo EvalRes)
type    EvalStack      = [EvalStackFrame]

-----------------
--	"ETC_Env": --
-----------------
--	EVALUATION AND TYPE-CHECKING ENVIRONMENT.
--	A record with named fields is used. NB: NameBinds stored in an ETC_Env are
--		typically passed DOWN the recursive evaluation calls, whereas those in
--		EvalRes are returned BACK UP:
--	The "_remDepth" field is for recursion control:  this is the maximum total
--		number of "evalExpr" and "evalFB" applications allowed in the course of
--		obtaining the evaluation result:
--
data ETC_Env =
	 ETC_Env {_bindsET     :: NameBinds,
			  _evalStack   :: EvalStack,

			  -- Recursion control during evaluation:
			  _remEvalDepth:: Int,
			  _maxADEpth   :: Int,

			  -- Flags for type-checking:
			  _inferTypes  :: Bool,
			  _noEval	   :: Bool,

			  -- Other flags:
			  _debugMode   :: Bool,
			  _strongEvalTC:: Bool,
			  _useCaching  :: Bool,

			  -- Functions from non-importable (due to recursive deps) modules:
			  -- For evaluation:
			  _evalFB	   :: ETC_Env -> AldorFBlock-> [EvalRes],
			  _evalExpr    :: ETC_Env -> AldorExpr  -> [EvalRes],
			  _evalDT	   :: ETC_Env -> DefTable   -> [DT_Res],

			  -- For type-checking:
			  _tcFB_T      :: ETC_Env -> AldorType  -> AldorFBlock -> [TC_Res],
			  _tcExpr_T    :: ETC_Env -> AldorType  -> AldorExpr   -> [TC_Res],
			  _tcFB_E      :: ETC_Env -> AldorExpr  -> AldorFBlock -> [TC_Res],
			  _tcFB_R      :: ETC_Env -> TC_Res     -> AldorFBlock -> [TC_Res]
			 }

-------------------
--	"updateTCE": --
-------------------
--	"ETC_Env" data are passed only DOWN into type-checking functions; "TC_Res"
--		data are passed only UP (as return values).  The "updateTCE" function
--		installes the env-related data from a "TC_Res" into a "ETC_Env" so the
--		updated env can be passed further:
--
updateTCE :: TC_Res -> ETC_Env -> ETC_Env
updateTCE    tcr	   tcenv  =  setBinds (getAttr tcr) tcenv

-------------------
--	"updateEE" : --
-------------------
--	Same as "updateTCE", but uses an "EvalRes" to update the binds of an "ETC_
--		Env":
updateEE  :: EvalRes-> ETC_Env -> ETC_Env
updateEE     er        ee      =  setBinds (getAttr er) ee

-----------------
--	"etc2ade": --
-----------------
--	Builds an "ADE_Env" from a given "ETC_Env":
--
etc2ade:: ETC_Env -> ADE_Env
etc2ade   tcenv   =  mkADEnv
					 (getBinds tcenv) (_maxADEpth tcenv) (_debugMode tcenv)


------------------
--	"pushES*":  --
------------------
--	"pushES":
--
--	Pushing function args on the EvalStack (as a part of an ETC_Env).  Only one
--		version of each arg is pushed at a time  (see the "FunctionCall" eval).
--		The args pushed are "EvalRes"s (i.e. with NameBinds).
--	There is no "popES": unwinding of stack frames is done simply by  returning
--		from a Haskell evaluation function to the previous environment:
--
pushES :: EvalStack -> [DefNo] -> [EvalRes] -> EvalStack

pushES    es           def_nos    args
	--
	--	Check that the number of formal parms (from which "def_nos" come) is
	--		the same as  that of actual ones  ("args"):
	--
	| length def_nos  == length args = newFrame : es

	| otherwise = error2 "pushES"
						 ("Incompatible args list: "++
						 (show (length def_nos ))   ++ " formals, " ++
						 (show (length args    ))   ++ " actuals")
	where
	--
	--	"newFrame": Made by "args"   (already evaluated, fully or partially;
	--		there should be no Bottoms  as they are removed BEFORE the args
	--		are pushed on the stack):
	--
	newFrame  ::EvalStackFrame
	newFrame  = foldl insertArg (EvalStackFrame emptyFM) (zip def_nos args)

	insertArg ::EvalStackFrame ->  (DefNo, EvalRes) -> EvalStackFrame
	insertArg  (EvalStackFrame fm) (def_no, arg)
		--
		--	If this "def_no" is already in the frame, something is wrong
		--		with enumerators!
		| elemFM def_no fm = error2 "pushES" ("DefNo."++(show def_no)++
									" appears more than once in a Proto")
		| otherwise        = EvalStackFrame (addToFM fm def_no arg)

--
--	"pushES_EE":
--
--	Like "pushES", but the 1st arg and the result are "ETC_Env"s (abstract
--		objects containing a EvalStack):
--
pushES_EE :: ETC_Env -> [DefNo] -> [EvalRes] -> ETC_Env

pushES_EE    ee         def_nos    args	     =
	ee {_evalStack   =  pushES (_evalStack ee)  def_nos args}


-------------------
--	"exportES*": --
-------------------
--	"exportES":
--	Converts a EvalStack (which abstract from the external view)  into a list of
--		frames. Stack top is at the head of the list, each frame is a sublist of
--		pairs (DefNo, EvalRes) in no particular order:
--
exportES   :: EvalStack -> [[(DefNo, EvalRes)]]
exportES      es        =  [fmToList fm | EvalStackFrame fm <- es]

--	"exportES_EE":
--	The same as "exportES", but the arg is an ETC_Env containing an EvalStack:
--
exportES_EE:: ETC_Env   -> [[(DefNo, EvalRes)]]
exportES_EE   ee		=  exportES (_evalStack ee)


--------------------
--	"lookupES*":  --
--------------------
--	Search the EvalStack frames from the top into the depth,   searching  for
--		values corresponding to a given DefNo:
--
lookupES   :: EvalStack -> DefNo -> Maybe EvalRes
lookupES      es		 def_no  =
	--
	--	Use "foldr" rather than "foldl" for traversing the stack frames, then
	--		the top-most instance of the arg in question  will be encountered
	--		first and returned (as it must be):
	--
	case foldr scanFrame []  es  of
		--  When a suitable value is found  on the stack,  lazy search termi-
		--		nates immediately:

		arg: _ -> Just arg	-- OK
		_	   -> Nothing   -- Not found!
	where
	scanFrame:: EvalStackFrame -> [EvalRes] ->  [EvalRes]

	scanFrame  (EvalStackFrame fm) curr =
		case lookupFM  fm  def_no   of
			Just arg_vals ->  arg_vals : curr   --  DefNo found!
			_             ->  curr              --  Scan further...

--
--	"lookupES_EE":
--	The same as "lookupES", but the aeg is an ETC_Env containing an EvalStack:
--
lookupES_EE :: ETC_Env -> DefNo -> Maybe EvalRes
lookupES_EE    ee        def_no =  lookupES (_evalStack ee) def_no


-- ============== --
--	"EvalCache":  --
-- ============== --
--	Used for efficiency.    Evaluated FBlocks are cached according to their
--		"FBNo"s. Evaluated type exprs in DTEs are cached according to their
--		"DefNo"s.
--	NB: "NameBinds" MUST be in the key, as a part of the "ETC_Env"  for which
--		the cached result was obtained. The other critical part, "EvalStack",
--		is not preset, as we only cache results obtained on an empty stack:
--		
data     EvalCacheKey = FB_Key  FBNo  NameBinds | 
						DefKey DefNo  NameBinds
		 deriving (Eq,  Ord)

newtype  EvalCache    = EvalCache  (FiniteMap EvalCacheKey [EvalRes])

--
--	The Cache itself is a Mutable Variable:
--	XXX: "unsafePerformIO" is used here!
--
cacheRef :: IORef EvalCache
cacheRef =  unsafePerformIO (newIORef (EvalCache  emptyFM))

------------------
--	"toCache":  --
------------------
--	Writing to the Cache: returns the value written:
--
toCache   :: EvalCacheKey -> [EvalRes] -> [EvalRes]
toCache      key             ress      =
	--	XXX: we use "unsafePerformIO" here!
	unsafePerformIO  written
	where
	--	Force evaluation of "ress" before storing it in the Cache:
	sress    :: [EvalRes]
	sress    =  seq ress ress

	written  :: IO [EvalRes]
	written  =
		do
		--	currMap  :: FiniteMap  FB_Key   [EvalRes]
			EvalCache   currMap <- readIORef cacheRef

		--	Add or overwrite the value for this key,    and store it in the
		--		mutable var. It may in general be possible to have multiple
		--		values EVEN FOR THE SAME NAME BINDS, so merge the conetents
		--		under the key:
		--	:: IO ()
			writeIORef  cacheRef
						(EvalCache (addToFM_C (\o n-> o++n) currMap key sress))

		--	Return the value written:
			return      sress


--------------------
--	"fromCache":  --
--------------------
--	Returns Nothing if there is no such key in the Cache, otherwise Just a list
--		of results (possibly empty!)
--
fromCache :: EvalCacheKey -> Maybe [EvalRes]
fromCache    key          =
	--	XXX: we use "unsafePerformIO" here!
	unsafePerformIO  got
	where
	got   :: IO (Maybe [EvalRes])
	got   =
		do
		--	currMap  :: FiniteMap  FB_Key   [EvalRes]
			EvalCache   currMap <- readIORef cacheRef

		--	Return the result of look-up:
			return (lookupFM  currMap key)


-- ======================= --
--	"WithAttr" Instances:  --
-- ======================= --
--	In all cases, the "attr" type is just "NameBinds":
--
instance WithAttr NameBinds EvalRes
	where
	getAttr   (Bottom_ER  binds   )		  = binds
	getAttr   (Partial_ER binds  _)		  = binds
	getAttr   (Full_ER    binds  _)		  = binds

	setAttr   binds (Bottom_ER   _    )	  = Bottom_ER   binds
	setAttr   binds (Partial_ER  _ fb )	  = Partial_ER  binds fb
	setAttr   binds (Full_ER     _ val)   = Full_ER     binds val


instance WithAttr NameBinds FBorDT
	where
	getAttr   (Bottom_FBDT binds  )		  = binds
	getAttr   (FB1		   binds _)		  = binds
	getAttr   (DT1		   binds _)		  = binds

	setAttr   binds (Bottom_FBDT _)		  = Bottom_FBDT binds
	setAttr   binds (FB1   _    fb)		  = FB1		 	binds fb
	setAttr   binds (DT1   _	dt)		  = DT1			binds dt


instance WithAttr NameBinds TypeOrFB
	where
	getAttr  (Bottom_TFB   binds  )		  = binds
	getAttr  (Type2		   binds _)		  = binds
	getAttr  (FB2		   binds _)		  = binds

	setAttr  binds (Bottom_TFB   _)		  = Bottom_TFB  binds
	setAttr  binds (Type2 _      t)		  = Type2		binds t
	setAttr  binds (FB2   _		fb)		  = FB2			binds fb


instance WithAttr NameBinds Proto_Res
	where
	getAttr   (Bottom_PR  binds   )		  = binds
	getAttr   (OK_PR      binds  _)		  = binds

	setAttr   binds (Bottom_PR   _)		  = Bottom_PR   binds
	setAttr   binds (OK_PR       _ proto) = OK_PR       binds proto


instance WithAttr NameBinds FBList_Res
	where
	getAttr   (Bottom_FBL  binds  )		  = binds
	getAttr   (OK_FBL      binds _)		  = binds

	setAttr   binds (Bottom_FBL  _)		  = Bottom_FBL  binds
	setAttr   binds (OK_FBL      _ fbs)   = OK_FBL      binds fbs


instance WithAttr NameBinds DT_Res
	where
	getAttr   (Bottom_DTR  binds  )		  = binds
	getAttr   (OK_DTR      binds _)		  = binds

	setAttr   binds (Bottom_DTR   _	  )   = Bottom_DTR  binds
	setAttr   binds (OK_DTR       _ dt)	  = OK_DTR      binds dt


instance WithAttr NameBinds FDTE_Res
	where
	getAttr  (Bottom_FDTER  binds   )	  = binds
	getAttr  (OK_FDTER      binds  _)	  = binds

	setAttr  binds (Bottom_FDTER   _)	  = Bottom_FDTER  binds
	setAttr  binds (OK_FDTER      _ fdte) = OK_FDTER      binds fdte


instance WithAttr NameBinds OpenedDynTarg
	where
	getAttr  (Bottom_ODT    binds   )     = binds
	getAttr  (OK_ODT        binds  _)     = binds

	setAttr  binds (Bottom_ODT     _)     = Bottom_ODT    binds
	setAttr  binds (OK_ODT  _   targ)     = OK_ODT        binds targ


-- =============== --
--	Bottom checks: --
-- =============== --
instance CanBeBottom   EvalRes
	where
	isBottom (Bottom_ER _)    = True
	isBottom _				  = False

instance CanBeBottom  FBorDT
	where
	isBottom (Bottom_FBDT _)  = True
	isBottom _				  = False

instance CanBeBottom  TypeOrFB
	where
	isBottom (Bottom_TFB  _)  = True
	isBottom _				  = False

instance CanBeBottom  Proto_Res
	where
	isBottom (Bottom_PR	  _)  = True
	isBottom _				  = False

instance CanBeBottom  FBList_Res
	where
	isBottom (Bottom_FBL _)   = True
	isBottom _				  = False

instance CanBeBottom  DT_Res
	where
	isBottom (Bottom_DTR _)   = True
	isBottom _				  = False

instance CanBeBottom  FDTE_Res
	where
	isBottom (Bottom_FDTER _) = True
	isBottom _				  = False

instance CanBeBottom  ER_HomoList
	where
	isBottom Bottom_ERL		  = True
	isBottom _				  = False

instance CanBeBottom  OpenedDynTarg
	where
	isBottom (Bottom_ODT _)   = True
	isBottom _				  = False


-- ============================= --
--	Functions calling "eval*":   --
-- ============================= --
--  Evaluation functions are lifted from objects in ETC_Env. Don't export
--		them, or they will conflict with their original definitions!
--
evalFB  :: ETC_Env -> AldorFBlock -> [EvalRes]
evalFB     ee         fb		  =  (_evalFB   ee)  ee  fb

evalExpr:: ETC_Env -> AldorExpr   -> [EvalRes]
evalExpr   ee		  expr		  =  (_evalExpr ee)  ee  expr

evalDT  :: ETC_Env -> DefTable    -> [DT_Res]
evalDT     ee		  dt		  =  (_evalDT   ee)  ee  dt


-----------------------
--	"openDynTarget": --
-----------------------
--	Evaluates all necessary "intermediate" data structures (nested ADTs) and
--		returns a non-dynamic final Target and the corresp NameBinds for the
--		obtained evaluation of the given dynamic Target, or a Bottom    (the
--		"OpenDyTarg" data type). The eval result may not be unique, hence the
--		list in the result type.
--	This function is implemented via "openDynTarget_R"  which is a recursive
--		function requiring recursion control. The end result is then wrapped
--		into "AddedTarg" so it can withstand intersections.  Don't forget to
--		replace the DynTarg for the given "use_no"      with the constructed
--		"AddedTarg":
--
openDynTarget  :: ETC_Env ->  UseNo -> BindTarget -> [OpenedDynTarg]

openDynTarget     ee          use_no   dtarg@(DynTarg _ _) =
	map wrapIntoAdded (openDynTarget_R ee [dtarg])
	where
	--	In the result, if the actual target "t" is constructed, it is wrapped
	--		into an "AddedTarg" to survive future intersections:
	--
	wrapIntoAdded :: OpenedDynTarg      -> OpenedDynTarg
	wrapIntoAdded    bot@(Bottom_ODT _) =  bot

	wrapIntoAdded    (OK_ODT binds   t) =  OK_ODT binds' added_targ
		where
		old_targs :: [BindTarget]
		old_targs =  lookupBinds binds use_no

		binds'    :: NameBinds
		binds'    =  modifyBinds binds use_no
								(added_targ : (delete dtarg old_targs))

		added_targ:: BindTarget
		added_targ=  mkAddedTarg t


openDynTarget _ _ hmm =
	error2 "openDynTarget" ("Not a DynTarget: "++(show hmm))


-------------------------
--	"openDynTarget_R": --
-------------------------
openDynTarget_R:: ETC_Env -> [BindTarget] -> [OpenedDynTarg]
				  -- 		 history list    ultimate targets and binds

openDynTarget_R   eeT
				  history@((DynTarg (qhead:qtail) anchor_targ) : _) =
	ntargs
	where
	--	Get the type constraint and the body of the "anchor":
	--	The "anchor_targ" was already verified at creating time, so the
	--		following match will not fail:
	--
	anchor_type:: AldorExpr
	anchor_body:: AldorFBlock
	Just (anchor_type, anchor_body) = getTargFBE anchor_targ

	--	Evaluate "type_constr" and "anchor_fb" and cross  the results:
	--
	tress  :: [EvalRes]
	tress  =  evalExpr eeT      anchor_type

	aress  :: [EvalRes]
	aress  =  evalFB   eeT      anchor_body

	xress  :: [EvalRes_CL]
	xress  =  crossWithAttrs [tress, aress]

	ntargs :: [OpenedDynTarg]
	ntargs =  foldl  descend []		 xress

	-----------------
	--	"descend": --
	-----------------
	descend:: [OpenedDynTarg] -> EvalRes_CL -> [OpenedDynTarg]

	descend   curr     xres  =
		case  getCList xres of
			--	(0) At least one Bottom -- produce a Bottom:
			--
			[Bottom_ER _, _]-> (Bottom_ODT xbinds) : curr
			[_, Bottom_ER _]-> (Bottom_ODT xbinds) : curr

			--	(1) Have an ADT only, no type-constraining value  (since there
			--		was a VoidExpr in type constraint -- can it happen???) Or,
			--		the type constraint was just "TypeT":
			--
			[Full_ER _   TrivVal,
			 Full_ER _  (TypeVal adt@(ADT _ _ _))]  ->
				curr ++ (findInADT  eeX adt)

			[Full_ER _  (TypeVal TypeT),
			 Full_ER _  (TypeVal adt@(ADT _ _ _))]  ->
				curr ++ (findInADT  eeX adt)

			--	(2) Have a Category type constraint only, no body (so the body
			--		evaluated to a Partial VoidFB):
			--	We can in fact do more here:  if the eveluated type constraint
			--		is NOT a Category Literal,  but rather a "With2" expr with
			--		one part being an explicit DefTable, we can "descend" into
			--		that part as well:
			--
			[Full_ER    _  (TypeVal  (CategoryT _ dt)),
			 Partial_ER _  (VoidFB _)]  -> curr++ (findInDT_Cat eeX dt)

			[Partial_ER _  (SimpleFB _ (With2 _ c1 c2) _),
			 Partial_ER _  (VoidFB _)]  ->
				case (getJustDT c1, getJustDT c2) of
					 (Just dt1, Nothing)-> curr++ (findInDT_Cat eeX dt1)
					 (Nothing, Just dt2)-> curr++ (findInDT_Cat eeX dt2)
					 (Nothing,  Nothing)-> curr
					 (Just  _, Just   _)->
						--	Why have not we obtained a CategoryT Literal here???
						error2 "openDynTarget_R"
							   ("Incomplete evaluation of "++(show anchor_type))

			--	(3) Have both a Category type expr and an ADT body:
			--		Install results of ADT search ONLY if the corresp Category
			--		search yields a non-[] result (XXX is this requirement too
			--		relaxed?)
			[Full_ER _  (TypeVal (CategoryT _ dt)),
			 Full_ER _  (TypeVal adt@(ADT _ _ _))]    ->
				let
					adt_ress :: [OpenedDynTarg]
					adt_ress =  findInADT eeX adt

					--	Now, for each "adt_ress",   do a Category search under
					--		the new, restricted binds, and retain  that result
					--		only if the Category search yields non-[]:
					ok_ress   :: [OpenedDynTarg]
					ok_ress  =   foldl searchCat [] adt_ress

					searchCat :: [OpenedDynTarg]-> OpenedDynTarg->
								 [OpenedDynTarg]
					searchCat    curr_odts        odt
						| length (findInDT ee' dt) /= 0  =  odt : curr_odts
						| otherwise						 =  curr_odts
						where
						ee' :: ETC_Env
						ee' =  setBinds (getAttr odt) eeX
				in
					curr ++ ok_ress

			--	(4) All other cases: cannot get anything new from them!
			_			   ->  curr
		where
		xbinds :: NameBinds
		xbinds =  getAttr  xres

		eeX    :: ETC_Env
		eeX    =  setBinds xbinds eeT

	-----------------------
	--	"findInDT_Cat":  --
	-----------------------
	--	It is like the "findInDT" defined below,  but also finds all "%" Idents
	--		in the "dt" entries  and replaces their targs by the "anchor_targ".
	--		The latter, if contains only the type expr and no body, corresponds
	--		to a particular declared instance of this Category:
	--
	findInDT_Cat :: ETC_Env -> DefTable -> [OpenedDynTarg]
	findInDT_Cat    ee   dt =  odts'
		where
		--	Find all "%" Idents in the "dt". Since this is a Category DT, it is
		--		supposed to have declarations only, no def bodies -- so we scan
		--		just type exprs:
		--
		pcids:: [UseNo]
		pcids=  nub (foldDT findPerCentIds1 [] dt)

		findPerCentIds1 :: FlatDTE ->  [UseNo] -> [UseNo]
		findPerCentIds1    g@(GenFDTE _ _ type_expr body)  curr =
			case body of
				--	No Def body, as it should be. Find all "%" Idents in the
				--		"type_expr"; enclosing FBlocks are all ignored,   so
				--		passing "voidFB" for the outer-most enclosing FB  is
				--		OK:
				VoidFB _-> foldExpr findPerCentIds2 curr voidFB type_expr

				_       -> error2 "finPerCentIds1"
								  ("Category DTE has non-Void body: "++
								  (show g))

		findPerCentIds1   (RenFDTE _ _ _) curr = curr

		findPerCentIds2 :: AldorFBlock -> AldorExpr -> [UseNo] -> [UseNo]
		findPerCentIds2    _ (Identifier use_no ["%"]) curr = use_no : curr
		findPerCentIds2    _ _						   curr = curr

		--
		--	Construct all "OpenedDynTarg"s, and for each of them, modify the
		--		bind targs for "pcids":
		--
		odts, odts' :: [OpenedDynTarg]
		odts =  findInDT ee dt
		odts'=  map updatePerCentTargs odts

		updatePerCentTargs :: OpenedDynTarg -> OpenedDynTarg
		updatePerCentTargs    odt = setAttr binds' odt
			where
			binds,   binds':: NameBinds
			binds =  getAttr  odt
			binds'=  foldl addAnchorTarg binds pcids

			addAnchorTarg :: NameBinds -> UseNo -> NameBinds
			addAnchorTarg    curr         pcid  =
				modifyBinds  curr pcid    [mkAddedTarg anchor_targ]


	-------------------
	--	"findInADT": --
	-------------------
	--		Search the DefTable of an ADT (all entries for the "qhead").
	--		Alternatively,  if "qhead" is "Rep", search the rep clause:
	--
	findInADT:: ETC_Env -> AldorType -> [OpenedDynTarg]

	findInADT   eeA       (ADT _ rep_fb dt)  =
		if      qhead == "Rep"
		then    findInRep eeA rep_fb
		else    findInDT  eeA     dt

	findInADT _ hmm  = error2 "findInADT" ("Not an ADT: "++(show hmm))


	-------------------
	--	"findInDT":  --
	-------------------
	findInDT :: ETC_Env -> DefTable  -> [OpenedDynTarg]

	findInDT    eeA        dt        =
		foldl (findInFDTE eeA) [] fdtes
		where
		fdtes :: [FlatDTE]
		fdtes =  map (dte2fdte qhead) (lookupDT  dt  qhead)


	-------------------
	--	"findInRep": --
	-------------------
	findInRep::ETC_Env -> AldorFBlock-> [OpenedDynTarg]

	findInRep  eeR rep_fb =
		case qtail of
			[] ->  -- Return a RepTarg (what else can we return here!?) wrapped
				   [OK_ODT (getBinds eeR) (RepTarg rep_fb)]

			_  ->  -- Descend into "rep_fb" by calling "openDynTarget"  on  a
				   --	synthetic target. The new "anchor_targ" is a "RepTarg"
				   -- 	enveloping the given "rep_fb":
				   --
				   openDynTarget_R eeR
								   ((DynTarg qtail (RepTarg rep_fb)):history)

	--------------------
	--	"findInFDTE": --
	--------------------
	--	(1) A "GenFDTE" for the "qhead":
	--
	findInFDTE::  ETC_Env-> [OpenedDynTarg] -> FlatDTE -> [OpenedDynTarg]

	findInFDTE  eeD  curr fdte@(GenFDTE _ _ _ _)   =
		case qtail of
			--	Empty "qtail": the final def has been found. Produce the resul-
			--		ting (synthetic) DefTarg  (a RepTarg cannot occur here, as
			--		we actually found a DEFINITION):
			--
			[]  ->
				(OK_ODT (getBinds eeD) (DefTarg fdte)) : curr

			--	Otherwise: recurse down the "def_body". For that, make a synth-
			--		etic "DynTarg", and append its results to "curr":
			--
			_   ->  (openDynTarget_R eeD
								   ((DynTarg qtail (DefTarg fdte)):history)) ++
					curr

	--
	--	(2) A "RenFDTE" for the "qhead":
	--		All renamings are supposed to be already resolved in the binds, so
	--		just get its target and follow it.
	--		NB: recursion control is applied here to detect possible  indirect
	--		cyclic renamings (can they occur at all?):
	--
	findInFDTE  eeD   curr (RenFDTE _ _ (Identifier ruse_no _)) =
		(concat (map followRenTarg ok_targs)) ++ curr
		where
		--	Get the targets (there can be more than one)  of the renaming
		--		Identifier. Remove those targets which are already in the
		--		"history list" (recursion control):
		--
		rentargs   :: [BindTarget]
		rentargs    = lookupBinds  (getBinds eeD)  ruse_no

		ok_targs   :: [BindTarget]
		ok_targs    = filter (\ t -> not (elem t history)) rentargs

		-----------------------
		--	"followRenTarg": --
		-----------------------
		--	Each of the "ok_targ"s can, in general, have any BinTarget ctor
		--		but there are some restrictions, as shown below:
		--
		followRenTarg :: BindTarget -> [OpenedDynTarg]

		--	(A) "DefTarg":
		--		It can only be a GenFDTE, as chain renamings have already been
		--		resolved during the construction of NameBinds.  Delegate this
		--		to "findInFDTE" again which will take action depending on whe-
		--		ther "qtail" is empty or not:
		--
		followRenTarg   (DefTarg   fdte@(GenFDTE _ _ _ _)) =
			findInFDTE eeD  curr   fdte

		followRenTarg   (DefTarg   fdte@(RenFDTE _ _ _  )) =
			error2 "followRenTarg"
				   ("Still unresolved chain renaming: "++ (show fdte))

		--	(B) "FP_Targ":
		--		Similar to "DefTarg", but "qtail" MUST be [] here -- there is
		--		no way down since the corresp DTE has no body,   so we have a
		--		more simple case:
		--
		followRenTarg  fp_targ@(FP_Targ (FParmFB _ _ _)) =
			case qtail of
				[] ->  (OK_ODT (getBinds eeD) fp_targ) : curr
				_  ->  curr

		followRenTarg  (FP_Targ hmm) =
			error2 "followRenTarg" ("Invalid FP_Targ: "++(show hmm))

		--	(C) "PerCentTarg":
		--		Can it happen? -- Probably not:
		--
		followRenTarg  (PerCentTarg _) =
			error2 "followRenTarg" "PerCentTarg encountered"

		--	(D) "BI_Targ":
		--		This is OK provided that "qtail" is empty:
		--
		followRenTarg  bi_targ @ (BI_Targ _) =
			case qtail of
				[]  -> (OK_ODT (getBinds eeD) bi_targ) : curr
				_   -> curr      -- Simply drop an improper case!

		--	(E) "RepTarg":
		--		Search it using "findInRep":
		--
		followRenTarg  (RepTarg rep_fb) = findInRep eeD rep_fb

		--	(F) "DynTarg":
		--		A nested "DynTarg" encountered -- OK, process it as well!
		--
		followRenTarg  (another_dyn@(DynTarg _ _)) =
			openDynTarget_R  eeD  (another_dyn : history)

		--	(G) "CondTarg":
		--		Should not occur here, as it is dynamically generated and so
		--		cannot be found in the RHS of a statically-defined renaming:
		--
		followRenTarg  hmm@(CondTarg _)  =
			error2 "followRenTarg" ("CondTarg encountered: " ++(show hmm))

		--	(H) "AddedTarg":
		--		Should not occur here -- cannot be a renaming target,  as it
		--		is dynamically generated -- similar to "CondTarg" above:
		--
		followRenTarg  hmm@(AddedTarg _) =
			error2 "followRenTarg" ("AddedTarg encountered: "++(show hmm))


	findInFDTE _ _ (RenFDTE _ _ expr) =
		error2 "findInFDTE" ("Not an Identifier in RenFDTE: "++(show expr))


--	It's an error if "openDynTarget_R" is called with an empty history list,
--		or the history head is not a DynTarg:
--
openDynTarget_R _ hmm =
	error2 "openDynTarget_R" ("Invalid arg: "++(show hmm))


----------------------
--	"getUltADTrep": --
----------------------
--	Get the ULTIMATE evaluated rep of an ADT recursively (as the immediate rep
--		may itself be an ADT). Since evaluation is performed, multiple results
--		are possible:
--
getUltADTrep  :: ETC_Env -> AldorType -> [EvalRes]

getUltADTrep     ee (ADT _  rep_fb _) =
	concat    [descend res | res   <- evalFB ee rep_fb]
	where
	descend:: EvalRes -> [EvalRes]
	--
	--	The rep is again an ADT:
	--
	descend  (Full_ER        binds'    (TypeVal  inner@(ADT _ _ _))) =
		getUltADTrep  (setBinds binds' ee) inner
	--
	--	The rep is a type but not an ADT: bottom reached:
	--
	descend  final@(Full_ER _ (TypeVal _)) =  [final]
	--
	--	The rep is not a TypeVal at all:  regarded as an error:
	--
	descend  hmm  =
		error2 "getUltADTrep"
			   ("ADT rep evaluated into\n"++(show hmm)++"\n, not a TypeVal")

getUltADTrep _ hmm  = error2 "getUltADTrep" ("Not an ADT: "++(show hmm))


-- ======================= --
--	Misc Helper Functions: --
-- ======================= --
---------------------------------------
--	"WithBinds" "ETC_Env" instance:  --
---------------------------------------
instance WithBinds ETC_Env
	where
	getBinds ee		  = _bindsET ee
	setBinds binds ee = ee {_bindsET = binds}


------------------------
--	"setPartialFBNo": --
------------------------
--	Set a given FBNo on an FBlock occurreng within a Partial_ER:
--
setPartialFBNo :: FBNo -> EvalRes -> EvalRes
setPartialFBNo    fb_no  (Partial_ER binds fb) =
						  Partial_ER binds (setFBNo fb fb_no)
setPartialFBNo    _       other   =  other


-----------------------------
--	"homogenise_ER_List":  --
-----------------------------
--	Converts a list of "EvalRes"s into a list of "AldorVal"s (if all elements
--		of the original list are actually Vals) or into a list of FBlocks (if
--		all original elements are not Bottoms  but there was at least one not
--		fully evaluated element). The [] list is considered to be of Vals. If
--		at least one Bottom is present, "Bottoms" is returned.   The original
--		order of the elements is preserved in the result:
--	NB: This function DOES NOT unify the binds -- they are  not included into
--		the result:
--
homogenise_ER_List:: [EvalRes] -> ER_HomoList

homogenise_ER_List    ers  =  separateCtors ers  [] []
	where
	separateCtors::   [EvalRes] -> [AldorFBlock] -> [AldorValue] -> ER_HomoList

	--
	--	When there is no more input, examine the the output:
	--
	separateCtors   []  fbs vals
		--	Each "val" occurs in "vals" and also in "fbs" (lifted), so if there
		--		is an equal number of "vals" and "fbs",  there are in fact only
		--		fully evaluated values. NB: the lists accumulated in "vals" and
		--		"fbs" are in the reverse order compared to the input!:
		--
		| length fbs == length vals = Full_ERL     (reverse vals)
		| otherwise					= Partial_ERL  (reverse fbs )

	--
	--	When there is still some input: examine it:
	--
	separateCtors    (er_head : er_tail) fbs vals =
		case  er_head   of
			Bottom_ER  _     -> Bottom_ERL -- Full stop: Bottom encountered

			Partial_ER _ fb  -> separateCtors er_tail (fb:fbs) vals
								-- Attach "fb" as is
			Full_ER    _ val ->
				let
					lifted_val::  AldorFBlock
					lifted_val =  simpleFB (Literal val)
				in
					separateCtors er_tail  (lifted_val : fbs) (val : vals)
								  		-- Attach "val" and its wrapped repr

---------------
--	"er2fb": --
---------------
er2fb:: EvalRes ->   AldorFBlock

er2fb  (Bottom_ER  _    ) = error2 "er2FB" "Bottom encountered"
er2fb  (Partial_ER _ fb ) = fb
er2fb  (Full_ER    _ val) = simpleFB (Literal val)


---------------
--	"fb2er": --
---------------
--	XXX: Are "VoidFB" and "FParmFB" allowed here?
--
fb2er:: NameBinds -> AldorFBlock -> EvalRes

fb2er   binds        (SimpleFB _ (Literal val) _) = Full_ER    binds val
fb2er   binds        fb                           = Partial_ER binds fb


------------------
--	"er2adtr":  --
------------------
--	NB: It does NOT accept Bottoms -- they must be filtered out BEFORE this
--		function is applied:
--
er2adtr :: EvalRes -> Maybe ADT_Res

er2adtr   (Full_ER    _ (TypeVal adt@(ADT _ _ _))) = Just (OK_ADT adt)
er2adtr   (Full_ER    _  _)						   = Nothing

er2adtr   (Partial_ER _ fb) =
	case getJustDT fb of
		Just dt ->
			case mkADT_fromDT dt of
				Just adt -> Just (OK_ADT adt)
				_        -> Just (OK_DT   dt)   -- Still a DT!

		_		-> Nothing

er2adtr   (Bottom_ER _)  =  error2 "er2adtr" "Bottom encountered"


---------------------
--	"evalFB_2DT":  --
---------------------
--	Evaluates an FBlock into a Deftable, if possible. Used in dynamic Category
--		and ADT extensions:
--
evalFB_2DT :: ETC_Env -> AldorFBlock -> [FBorDT]

--	(1) NestedFB without an inner FB: Just evaluate its FB:
--		
evalFB_2DT    ee        (NestedFB _ dt (VoidFB _) _) =
	map dtr2fbdt (evalDT ee dt)
--
--	(2) Other FBs: must yield a Category for positive result:
--
evalFB_2DT    ee    fb = map er2fbdt (evalFB ee fb)


dtr2fbdt:: DT_Res ->  FBorDT
dtr2fbdt  (OK_DTR     binds dt) = DT1		  binds dt
dtr2fbdt  (Bottom_DTR binds)    = Bottom_FBDT binds


er2fbdt:: EvalRes ->  FBorDT

er2fbdt  (Full_ER     binds (TypeVal (CategoryT _ dt))) = DT1 binds dt
er2fbdt  (Full_ER     binds other_val)					=
	FB1   binds (simpleFB   (Literal other_val))

er2fbdt  (Partial_ER  binds fb) 						= FB1 binds fb
er2fbdt  (Bottom_ER   binds)							= Bottom_FBDT binds


-----------------
--	"er2expr": --
-----------------
--	Extract an expr from a Partial, or encapsulate a Full into an expr:
--
er2expr :: EvalRes      ->  AldorExpr

er2expr   (Bottom_ER  _ ) = error2 "er2expr" "Bottom encountered"

er2expr   (Partial_ER _ (SimpleFB _ expr _)) = expr

er2expr   (Partial_ER _ another_fb)		   =
	error2 "er2expr" ("Unexpected FB:\n"++(show another_fb))

er2expr   (Full_ER    _ val)			   = Literal val


------------------
--	"fbdt2fb":  --
------------------
fbdt2fb :: FBorDT -> AldorFBlock
fbdt2fb   (FB1 _ fb)	  = fb
fbdt2fb   (DT1 _ dt)	  = nestedFB dt voidFB  -- NB: UNDECOR!
fbdt2fb   (Bottom_FBDT _) = error2 "fbdt2fb" "Bottom_FBDT encountered"


------------------------
--	"decr_rem_depth": --
------------------------
decr_rem_depth:: ETC_Env -> ETC_Env

decr_rem_depth   ee =
	if   curr_rem_depth <= 0
	then error2 "decr_rem_depth" "Hit the recursion bottom"
	else ee {_remEvalDepth = curr_rem_depth - 1}
	where
	curr_rem_depth :: Int
	curr_rem_depth =  _remEvalDepth ee


------------------
--	"newEnum":  --
------------------
--	XXX: for the moment, it just propagates the old emunerator:
--
newEnum:: ETC_Env -> NameBinds -> Int -> Int
newEnum   _          _            i   =  i


-- =========== --
--	"show"s:   --
-- =========== --
-----------------
--	"showESF": --
-----------------
--	Shows "EvalStackFrame"s. However,  the latter cannot simply be made member
--		of the "Show" class, since extra parms may be needed.
--	NB: This function should be used with caution  -- it can produce VERY LONG
--		output!..
--
showESF ::  EvalStackFrame ->  Bool      -> String

showESF    (EvalStackFrame fm) debugMode =
	"\n{"++(foldFM showStackPos "" fm)   ++"\n}"
	where
	showStackPos::  DefNo -> EvalRes -> String -> String
	showStackPos    def_no   arg        curr   =
		curr    ++
		"\n\t#" ++ (show def_no)++ " => "++ (showArg arg) ++"\n\t"

	--
	--	"showArg":
	--	Replaces the standard "show" on "EvalRes", as we normally do not
	--		want to print NameBinds here (unless "debugMode" is set):
	--
	showArg  :: EvalRes  -> String

	showArg     res@(Bottom_ER  _    ) =
		if debugMode then show res else "_|_"

	showArg     res@(Partial_ER _ fb ) =
		if debugMode then show res else show fb

	showArg     res@(Full_ER    _ val) =
		if debugMode then show res else show val

-----------------------------
--	"EvalStackFrame" show: --
-----------------------------
--	Calls "showESF" with debugging turned off:
--
instance Show EvalStackFrame  where
	show esf = (showESF  esf  False)++"\n"

----------------------
--	"EvalRes" show: --
----------------------
instance Show EvalRes
	where
	show (Bottom_ER  binds    ) = "_|_,\n"		 ++
								  "Binds   => "  ++(show binds)

	show (Partial_ER binds fb ) = "Partial_ER,\n"++
								  "Binds   => "  ++(show binds)
												 ++",\n"++
								  "NormForm=> "  ++(show fb)

	show (Full_ER    binds val) = "Full_ER,\n"	 ++
								  "Binds   => "  ++(show binds)
												 ++",\n"++
								  "Val     => "  ++(show val)
--------------------
--	"FBList_Res": --
--------------------
instance Show FBList_Res
	where
	show (Bottom_FBL  binds)	= "_|_\n,"        ++
								  "Binds   => "   ++(show binds)

	show (OK_FBL      binds fbs)= "OK_FBL,\n"     ++
								  "Binds   => "   ++(show binds)
												  ++",\n"++
								  "FBs     => "   ++(show fbs)


-- ============================================= --
--	Type-Checking Functions used in Evaluation:  --
-- ============================================= --
----------------------------------
--	Lifted from the "ETC_Env":  --
----------------------------------
tcExpr_T :: ETC_Env -> AldorType -> AldorExpr  ->  [TC_Res]
tcExpr_T    ee  lhs    rhs       =  (_tcExpr_T ee) ee lhs rhs

tcFB_T   :: ETC_Env -> AldorType -> AldorFBlock->  [TC_Res]
tcFB_T		ee  lhs    rhs		 =  (_tcFB_T   ee) ee lhs rhs


------------------------
--	"weaklyConforms": --
------------------------
--	A "Weak" Type Conformance Check:
--	"Nothing" (undecided conformance) can occur in 2 cases here:
--		(1) the actual type of the "val" may be "Nothing" (unknown);
--		(2) the type comparison result   may be "Nothing".
--	HERE the "Nothing" result is considered to be positive ("True"),
--		that's why we call it a "WEAK" type conformance check:
--
weaklyConforms:: ETC_Env -> AldorValue -> AldorType -> Bool

weaklyConforms   ee         val           mustbe_type
	| isNothing (act_type) =  True
	| otherwise			   =  eq_res
	where
	act_type :: Maybe AldorType
	act_type =  elemTypeOf  val

	eq_res   :: Bool
	eq_res   =  case  adeq (etc2ade ee) mustbe_type (fromJust act_type) of
		Yes_ADER  _ -> True  -- XXX: Binds are thrown away!
		No_ADER     -> False
		Bottom_ADER -> True  -- Undecided, treated as conformant


---------------------------
--	"stronglyConforms":  --
---------------------------
--	A "Strong" Type Conformance Check, based on full type-checking WITHOUT
--		evaluating the target. The positive result is returned iff type-
--		checking returns a non-empty result list.
--	XXX: PROBLEMS WITH THIS APPROACH:
--	(1) "Bottom_TCR"s are silently ACCEPTED,  since there is no way to in-
--		corporate them into the result explicitly, so the check is WEAKENED;
--	(2) Restricted NameBinds  resulting from the type-checking are dropped
--		and not used further, for the same reason;  this WEAKENS the subse-
--		quent multi-evaluation (makes it more general);
--	(3) on the other hand, type-checking without further evaluation may be
--		more STRICT than necessary...
--
stronglyConforms:: ETC_Env -> AldorValue -> AldorType -> Bool

stronglyConforms   ee         val           mustbe_type  =
	case  tcExpr_T ee  mustbe_type (Literal val)  of
		[] -> False
		_  -> True

-------------------
--	"conforms":  --
-------------------
--	Either "weaklyConforms", or "stronglyConforms":
--
conforms :: ETC_Env -> AldorValue -> AldorType  ->  Bool

conforms    ee         val           mustbe_type
	| _strongEvalTC    ee = stronglyConforms ee val mustbe_type
	| otherwise			  = weaklyConforms   ee val mustbe_type


-- ============== --
--	"debug_msg":  --
--	(!!!)	   :  --
-- ============== --
debug_msg:: ETC_Env -> String -> a   -> a
debug_msg   ee         msg       val =
	if   _debugMode ee
	then trace msg val
	else val


(!!!) :: Show a => [a] -> Int -> a

as !!! i
	| i >= 0 && i < (length as) = as !! i
	| otherwise = error ("Unvalid Index: "++(show i)++" in "++(show as))


-- =========== --
--  "error2":  --
-- =========== --
error2 ::String -> String -> a
error2 = errorM modName

