-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--									"FoldFB":								 --
--		   Generic folding of Aldor FBs, Exprs, and other data structures	 --
--							L.A.Timochouk, 2000--2002						 --
-- ========================================================================= --
module FButils
	   (
			foldFB_Gen, foldExpr_Gen, passFB,   passExpr,
			foldFB,     foldExpr,	  foldDT_inclImports,
			getDT,		getJustDT,	  getExpr,  getFBNo,  getUpLinkFB,
			setFBNo,    mergeDTs,	  getPerCentDefNo,	  getPerCentDefNo_U,
			setPerCentDefNo_T,		  setPerCentDefNo_E
	   )
where

import DataTypes
import PrPrint

modName :: String
modName =  "FButils"


-- =============== --
--	"foldFB_Gen":  --
-- =============== --
--	Generic traversing function for FBlocks. It performs user-specified Actions
--		on all FBlocks and all Exprs encountered. These Actions are totally un-
--		related, but the action on Exprs also receives the enclosing FBLock  as
--		an extra arg, as it is useful in many applications.    The "foldFB_Gen"
--		function returns a pair of the cumulative results of both actions:
--
--	Arguments:		  FB_Action, Expr_Action, Init_Vals_Pair,  Starting_FB

foldFB_Gen :: (AldorFBlock -> a -> a) ->
			  (AldorFBlock -> AldorExpr -> b -> b) -> (a,b)  ->
		       AldorFBlock ->   (a,b)

--
--	(1) "NestedFB":
--
foldFB_Gen    (actionFB   ::  AldorFBlock -> a -> a)
			  (actionExpr ::  AldorFBlock -> AldorExpr -> b -> b)
			  (curr_pair  ::  (a,b))
			  thisFB@(NestedFB _ dt innerFB _) =  resFF
	where
	--	First of all, apply "actionFB" to "thisFB" itself:
	resA  ::  a
	resA  =   actionFB  thisFB  (fst curr_pair)

	--	Then traverse the DefTable -- use "foldDT_inclImports", not "foldDT",
	--		as the latter does not do anything with Imports. The DT is Stand-
	--		Alone, not a part of an ADT or Category:
	resDT ::  (a,b)
	resDT =   foldDT_inclImports actionFB actionExpr thisFB
								 (resA, snd curr_pair)   dt

	--	Finally, traverse the "innerFB":
	resFF ::  (a,b)
	resFF =   foldFB_Gen actionFB actionExpr resDT innerFB

--
--	(2) "SimpleFB":
--
foldFB_Gen    (actionFB   ::  AldorFBlock -> a -> a)
			  (actionExpr ::  AldorFBlock -> AldorExpr -> b -> b)
			  (curr_pair  :: (a,b))
			  thisFB@(SimpleFB _ expr _)  =  resEE
	where
	--	First of all, apply "actionFB" to "thisFB" itself:
	resA  ::  a
	resA  =   actionFB      thisFB    (fst  curr_pair)

	--	Then search for other FBlocks in the depth of "expr":
	resEE ::  (a,b)
	resEE =   foldExpr_Gen
			  actionFB  actionExpr (resA, snd curr_pair)  thisFB  expr

--
--	(3) "VoidFB":
--
foldFB_Gen    (actionFB   ::  AldorFBlock -> a -> a)   _  curr_pair
			  thisFB@(VoidFB _) =
	(actionFB thisFB (fst curr_pair), snd (curr_pair))

--
--	(4) "FBlockFB":
--
foldFB_Gen    (actionFB   ::  AldorFBlock -> a -> a)
			  (actionExpr ::  AldorFBlock -> AldorExpr -> b -> b)
			  (curr_pair  :: (a,b))
			  thisFB@(FParmFB _ _ _)  =  resEE
	where
	--	First of all, apply "actionFB" to "thisFB" itself:
	resA  ::  a
	resA   =  actionFB  thisFB  (fst curr_pair)

	--	Then process the type expr in the Definition. There is no body here
	--		(no default parms):
	resEE ::  (a,b)
	GenFDTE _ _ type_expr _ = getFParm  thisFB
	resEE  =  foldExpr_Gen
			  actionFB actionExpr (resA, snd curr_pair)  thisFB   type_expr


-- ================= --
--	"foldExpr_Gen":  --
-- ================= --
--	Used by "foldFB_Gen", so its "action"s are the same as that of "foldFB_Gen".
--	In all equations of this function,  "actionExpr" is first applied to "this"
--		expr:

foldExpr_Gen :: (AldorFBlock -> a -> a)  ->
				(AldorFBlock -> AldorExpr-> b -> b) -> (a,b) ->
				 AldorFBlock -> AldorExpr			-> (a,b)

--
--	(1): "Literal":
--		 There may be quite a few non-trivial cases here,  so factor this code
--		 into "foldVal":
--
foldExpr_Gen     (actionFB   ::  AldorFBlock -> a -> a)
				 (actionExpr ::  AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  ::  (a,b))			encl_fb
				 thisExpr@(Literal val)  =

	foldVal_Gen  actionFB actionExpr resE0 encl_fb val
	where
	resE0 :: (a,b)
	resE0  = (fst curr_pair, actionExpr encl_fb thisExpr (snd curr_pair))

--
--	(2): "AndList":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(AndList fbs)  =

	foldl actionFBL resE0 fbs
	where
	resE0 :: (a,b)
	resE0 =  (fst curr_pair, actionExpr encl_fb thisExpr (snd curr_pair))

	--	Use "foldl" to fold all operands in the original order, although this
	--		may not be essential...
	--
	actionFBL :: (a,b)  -> AldorFBlock -> (a,b)
	actionFBL =  foldFB_Gen   actionFB  actionExpr

--
--	(3): "OrList"
--		 Like "AndList":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(OrList  fbs)  =

	foldl actionFBL resE0    fbs
	where
	resE0 :: (a,b)
	resE0 =  (fst curr_pair, actionExpr encl_fb thisExpr (snd curr_pair))

	--	Use "foldl" to fold all operands in the original order, although this
	--		may not be essential...
	--
	actionFBL :: (a,b) -> AldorFBlock -> (a,b)
	actionFBL =  foldFB_Gen  actionFB  actionExpr

--
--	(4): "IfThenElse":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(IfThenElse  cond_fb  then_fb else_fb) =
	resE3
	where
	resE0, resE1, resE2, resE3 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr   encl_fb  thisExpr  (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB  actionExpr  resE0  cond_fb
	resE2 =  foldFB_Gen      actionFB  actionExpr  resE1  then_fb
	resE3 =  foldFB_Gen      actionFB  actionExpr  resE2  else_fb

--
--	(5): "FunctionCall":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(FunctionCall  func args)  =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr  encl_fb thisExpr (snd curr_pair))
	resE1 =  foldFB_Gen		 actionFB actionExpr resE0 func
	resE2 =  foldl			 actionFBL           resE1 args

	--	Use "foldl" here to process all  args  in the original order, although
	--		this may not be essential...
	--
	actionFBL :: (a,b) -> AldorFBlock -> (a,b)
	actionFBL =  foldFB_Gen  actionFB  actionExpr

--
--	(6) "RecordCtor":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(RecordCtor fields)=
	resE1
	where
	resE0, resE1 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr  encl_fb  thisExpr  (snd curr_pair))
	resE1 =  foldl			 actionFBL   resE0    fields

	--  Use "foldl" here to process all fields in the original order, although
	--		this may not be essential...
	--
	actionFBL :: (a,b)   ->  AldorFBlock -> (a,b)
	actionFBL =  foldFB_Gen  actionFB  actionExpr

--
--	(7) "UnionCtor":
--
foldExpr_Gen	 (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(UnionCtor selector  val_fb) =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb  thisExpr  (snd curr_pair))
	resE1 =  foldExpr_Gen    actionFB   actionExpr  resE0   encl_fb  selector
	resE2 =  foldFB_Gen      actionFB   actionExpr  resE1   val_fb

--
--	(8):  "UnionCase":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))         encl_fb
				 thisExpr@(UnionCase  union selector)  =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb  thisExpr  (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0  union
	resE2 =  foldExpr_Gen    actionFB   actionExpr  resE1  encl_fb   selector

--
--	(9): "RestrictType":
--
foldExpr_Gen	 (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		    encl_fb
				 thisExpr@(RestrictType expr_fb type_fb) =
	resE2
	where
	resE0, resE1, resE2  ::  (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb    thisExpr  (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr resE0 expr_fb
	resE2 =  foldFB_Gen      actionFB   actionExpr resE1 type_fb

--
--	(10): "ConvertTo":
--
foldExpr_Gen	 (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 thisExpr@(ConvertTo   expr_fb type_fb)  =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb     thisExpr   (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0 expr_fb
	resE2 =  foldFB_Gen      actionFB   actionExpr  resE1 type_fb

--
--	(11): "Add*":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))         encl_fb
				 thisExpr@(Add1 _ fb) =
	resE1
	where
	resE0, resE1 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb     thisExpr   (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0      fb


foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))         encl_fb
				 thisExpr@(Add2 _ fb1 fb2) =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb     thisExpr   (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0      fb1
	resE2 =  foldFB_Gen      actionFB   actionExpr  resE1      fb2

--
--	(12): "With*":
--
foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))         encl_fb
				 thisExpr@(With1 _ fb) =
	resE1
	where
	resE0, resE1 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb     thisExpr   (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0      fb


foldExpr_Gen     (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))         encl_fb
				 thisExpr@(With2 _ fb1 fb2) =
	resE2
	where
	resE0, resE1, resE2 :: (a,b)

	resE0 =  (fst curr_pair, actionExpr encl_fb     thisExpr   (snd curr_pair))
	resE1 =  foldFB_Gen      actionFB   actionExpr  resE0      fb1
	resE2 =  foldFB_Gen      actionFB   actionExpr  resE1      fb2


--
--	(13): In all other cases, just apply "actionExpr" to this expr:
--
foldExpr_Gen _ actionExpr curr_pair encl_fb thisExpr =
	(fst curr_pair,      actionExpr encl_fb thisExpr (snd curr_pair))


---------------------
--	"foldVal_Gen": --
---------------------
--	Recursively searches for FBs and Exprs inside "AldorVal"s:
--
foldVal_Gen ::   (AldorFBlock -> a -> a) ->
				 (AldorFBlock -> AldorExpr   -> b -> b) -> (a,b) ->
				  AldorFBlock -> AldorValue  ->   (a,b)
--
--	(1) "TypeVal":
--
foldVal_Gen       actionFB  actionExpr curr_pair encl_fb   (TypeVal tv)   =
	foldType_Gen  actionFB  actionExpr curr_pair encl_fb   tv

--
--	(2) "LambdaVal":
--
foldVal_Gen       actionFB  actionExpr curr_pair _ (LambdaVal proto body) =
	res2
	where
	res1 =  foldFProto_Gen  actionFB   actionExpr curr_pair proto
	res2 =  foldFB_Gen      actionFB   actionExpr res1      body

--
--	(3) "RecordVal":
--
foldVal_Gen      (actionFB   :: AldorFBlock -> a -> a)
				 (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair  :: (a,b))		   encl_fb
				 (RecordVal  fields)		=
	foldl  actionL curr_pair fields
	where
	--	NB: "fields" are searched in their declaration order due to "foldl":
	actionL:: (a,b) -> AldorValue -> (a,b)
	actionL   curr     val        =
		foldVal_Gen    actionFB actionExpr curr encl_fb val

--
--	(4) "UnionVal":
--		Similar to "UnionCtor":
--
foldVal_Gen       actionFB  actionExpr curr_pair encl_fb (UnionVal sel val) =
	resE2
	where
	resE1 =  foldExpr_Gen   actionFB   actionExpr  curr_pair encl_fb  sel
	resE2 =  foldVal_Gen    actionFB   actionExpr  resE1     encl_fb  val

--
--	(5) "ADT_Val":
--		NB: the ADT ref itself is an "alias" only, so it is not followed:
--
foldVal_Gen       actionFB  actionExpr curr_pair encl_fb (ADT_Val  _ repr) =
	foldVal_Gen   actionFB  actionExpr curr_pair encl_fb repr

--
--	(6) Others: Do nothing:
--
foldVal_Gen _ _ curr_pair _ _ = curr_pair


-- ================= --
--	"foldType_Gen":  --
-- ================= --
--	Traverses non-trivial "AldorType"s. Also requires an "encl_fb" (for ADT):
--
foldType_Gen:: (AldorFBlock -> a -> a) ->
			   (AldorFBlock -> AldorExpr   -> b -> b) -> (a,b) ->
				AldorFBlock -> AldorType			  -> (a,b)
--
--	(1) "FunctionT":
--
foldType_Gen        actionFB   actionExpr  curr_pair _ proto@(FunctionT _ _) =
	foldFProto_Gen  actionFB   actionExpr  curr_pair   proto

--
--	(2) "RecordT":
--
foldType_Gen	    actionFB   actionExpr  curr_pair _ (RecordT fields)      =
	foldFParms_Gen  actionFB   actionExpr  curr_pair   fields
	--	Record Fields are treated like Function Parms!

--
--	(3) "UnionT":
--		Similar to "RecordT":
--
foldType_Gen        actionFB   actionExpr  curr_pair _ (UnionT  fields)	     =
	foldFParms_Gen  actionFB   actionExpr  curr_pair   fields

--
--	(4) "ADT":
--		Fold its representation FBlock and the DefTable.  For the latter,  we
--		must use "foldDT_inclImports"  and mark the DT as belonging to an ADT:
--
foldType_Gen       (actionFB   :: AldorFBlock -> a -> a)
				   (actionExpr :: AldorFBlock -> AldorExpr -> b -> b)
				   (curr_pair  :: (a,b))		 encl_fb
				   (ADT _ rfb  dt)			  =  resE1
	where
	resE0, resE1:: (a,b)
	resE0= foldDT_inclImports actionFB actionExpr encl_fb curr_pair dt
	resE1= foldFB_Gen         actionFB actionExpr         resE0    rfb

--
--	(5) "CategoryT":
--		Again, use "foldDT_inclImports"; mark the DT as belonging to a Category:
--
foldType_Gen        actionFB   actionExpr curr_pair encl_fb (CategoryT _ dt) =
	foldDT_inclImports
					actionFB   actionExpr encl_fb   curr_pair  dt

--
--	(6)  Others: Do nothing:
--
foldType_Gen _ _ curr_pair _ _ = curr_pair


------------------------
--	"foldFProto_Gen": --
------------------------
--	Used by "foldVal_Gen" (on "LambdaVal") and "foldType_Gen".
--	NB:		 It currently DOES NOT require an  "encl_fb" parm:
--
foldFProto_Gen:: (AldorFBlock -> a -> a)   ->
				 (AldorFBlock -> AldorExpr -> b -> b) -> (a,b) ->
			      AldorType   ->   (a,b)

foldFProto_Gen    actionFB  actionExpr curr_pair (FunctionT parms retype) =
	res2
	where
	res1 =   foldFParms_Gen actionFB   actionExpr curr_pair parms
	res2 =   foldFB_Gen     actionFB   actionExpr res1      retype

foldFProto_Gen _ _ _ _  =   error2 "foldFProto_Gen" "Not a FunctionT"


------------------------
--	"foldFParms_Gen": --
------------------------
--	Used by "foldFProto" and "foldType" on "RecordT", "UnionT" etc.
--	NB: It currently DOES NOT require   an "encl_fb" parm:
--
foldFParms_Gen ::(AldorFBlock -> a -> a)    ->
				 (AldorFBlock -> AldorExpr  -> b -> b) -> (a,b) ->
				 [AldorFBlock]->   (a,b)

foldFParms_Gen   (actionFB    :: AldorFBlock -> a -> a)
				 (actionExpr  :: AldorFBlock -> AldorExpr -> b -> b)
				 (curr_pair   ::(a,b))   fbs =
	foldl actionL curr_pair  fbs
	where
	--	Parms are processed in their order of declaration due to "foldl":
	--
	actionL:: (a,b) ->   AldorFBlock  -> (a,b)
	actionL=  foldFB_Gen actionFB   actionExpr


-- ======================= --
--	"foldDT_inclImports":  --
-- ======================= --
foldDT_inclImports :: (AldorFBlock -> a -> a)    ->
					  (AldorFBlock -> AldorExpr  -> b -> b) ->
					  AldorFBlock  ->
					  (a,b)		   -> DefTable   ->   (a,b)

foldDT_inclImports    (actionFB    :: AldorFBlock-> a -> a)
					  (actionExpr  :: AldorFBlock-> AldorExpr -> b -> b)
					  encl_fb
					  (curr_pair   :: (a,b))   dt =
	resDT
	where
	--	Imports are processed in their order of declaration due to "foldl":
	resI  ::  b
	resI  =   foldl actionExprL (snd  curr_pair) (getImports dt)

	actionExprL :: b -> AldorExpr -> b
	actionExprL    curr expr      =  actionExpr encl_fb expr curr

	--	Now process the DTEs:
	resDT :: (a,b) 
	resDT = foldDT     (mkDTact  actionFB   actionExpr  encl_fb)
					   (fst    curr_pair,   resI)    dt

-----------------
--	"mkDTact": --
-----------------
--	Converts generic actions on FBlocks and Exprs  into a unified action
--		suitable for "FoldDT". The FB enclosing the DefTable in question
--		is also specified,  and is passed to the action on type exprs in
--		all DTEs:
--
mkDTact :: (AldorFBlock -> a -> a)     ->
		   (AldorFBlock -> AldorExpr   ->  b -> b)   ->
		   AldorFBlock  ->
		   (FlatDTE     -> (a,b)       ->   (a,b))

mkDTact    (actionFB   ::  AldorFBlock ->  a -> a)
		   (actionExpr ::  AldorFBlock ->  AldorExpr -> b -> b)
		   encl_fb      =

	\ fdte (curr_pair  ::  (a,b))      ->
	case  fdte  of
		GenFDTE _ _ type_expr body_fb  ->
			let
				res21,  res22 :: (a,b)
				res21 = foldExpr_Gen actionFB actionExpr curr_pair
						encl_fb      type_expr
				res22 = foldFB_Gen   actionFB actionExpr res21 body_fb
			in
				res22

		RenFDTE _ _ ren_expr ->
			--
			--	The "ren_expr" is an Identifier, so don't forget to traverse it
			--		as well:
			--
			foldExpr_Gen  actionFB actionExpr curr_pair encl_fb ren_expr

-- ====================== --
--	Simplified "fold*"s:  --
-- ====================== --
--	Trivial actions which can be used with "fold*_Gen":
--
passFB  ::  AldorFBlock -> a -> a
passFB      _              a =  a

passExpr::  AldorFBlock -> AldorExpr -> b -> b
passExpr    _  _           b =  b

--
--	"foldFB":
--		Folding FBs only; for Exprs, no action is taken, and arbitrary data
--		(e.g. undefined) is carried along:
--			 
foldFB  ::  (AldorFBlock -> a -> a) -> a -> AldorFBlock -> a

foldFB      actionFB   a   fb =
	fst     (foldFB_Gen actionFB passExpr  (a,undef) fb)

--
--	"foldExpr":
--		Folding Exprs only; for FBs, no action is taken, and arbitrary data
--		(e.g. undefined) is carried along. Still need an "encl"_fb" for the
--		outer-most expr:
--
foldExpr::  (AldorFBlock -> AldorExpr -> b -> b) -> b ->
			 AldorFBlock -> AldorExpr		   	 -> b

foldExpr    actionExpr  b   encl_fb expr  =
	snd     (foldExpr_Gen passFB actionExpr (undef, b) encl_fb expr)


-- ====================== --
--	Auxiliary Functions:  --
-- ====================== --
---------------------
--	"getDT"		 : --
--	"getJustDT"  : --
--	"getExpr"	 : --
--	"getFBNo"    : --
--	"setFBNo"    : --
--  "getUpLinkFB": --
---------------------
getDT    :: AldorFBlock -> DefTable
getDT       (NestedFB _ dt _ _) = dt
getDT       hmm		           =
	error2 "getDT" ("Not a NestedFB:\n"++(show hmm))
--	"getDT" returns a DT and discards the Inner FB


getJustDT:: AldorFBlock -> Maybe DefTable
getJustDT   (NestedFB _ dt (VoidFB _) _) = Just dt
getJustDT   _						     = Nothing
--	"getJustDT" returns a DT from a NestedFB only if the Inner FB is Void,
--		otherwise returns Nothing


getExpr :: AldorFBlock -> AldorExpr
getExpr    (SimpleFB _ expr _)      = expr
getExpr    (NestedFB _ _ innerFB _) = getExpr innerFB

getExpr    (FParmFB  fbno _ _)		=
	error2 "getExpr" ("FParmFB No."++(show fbno)++" encountered")

getExpr    (VoidFB	 fbno)			=
	error2 "getExpr" ("VoidFB  No."++(show fbno)++" encountered")


--	NB: "getFBNo" is defined even on VoidFBs (as it may be  required for
--		navigating through FBlocks).   However,  it produces an error if
--		it encounters any undecorated FBlock (with FBNo being "noNUMBER"):
--
getFBNo :: AldorFBlock  -> FBNo
getFBNo    fb =  checkFBNo fb (getFBNo_U fb)


checkFBNo  :: AldorFBlock -> FBNo -> FBNo
checkFBNo     fb             fb_no
	| fb_no /= noNUMBER  =   fb_no
	| otherwise          =
	  error2 "getFBNo" ("UnDecorated FBlock:\n"++(show fb))


setFBNo    :: AldorFBlock -> FBNo -> AldorFBlock

setFBNo       (SimpleFB _     expr       uplink) fb_no =
			   SimpleFB fb_no expr       uplink

setFBNo       (NestedFB _     dt innerFB uplink) fb_no =
			   NestedFB fb_no dt innerFB uplink

setFBNo       (FParmFB  _     dt         uplink) fb_no =
			   FParmFB  fb_no dt         uplink

setFBNo       (VoidFB   _)    fb_no =    VoidFB  fb_no


getUpLinkFB:: AldorFBlock -> AldorFBlock
getUpLinkFB   (SimpleFB _ _   uplink) = uplink
getUpLinkFB   (NestedFB _ _ _ uplink) = uplink
getUpLinkFB   (FParmFB  _ _   uplink) = uplink
getUpLinkFB   (VoidFB   fbno) =
	 error2  "getUpLinkFB" ("VoidFB No."++(show fbno)++" encountered")


---------------------------
--	"getPerCentDefNo*":  --
---------------------------
--	Extracts DefNo from ADT- and Category-valued exprs:
--
getPerCentDefNo_U :: AldorExpr -> Maybe DefNo

getPerCentDefNo_U   (Literal (TypeVal (ADT       def_no _ _))) = Just def_no
getPerCentDefNo_U   (Literal (TypeVal (CategoryT def_no _  ))) = Just def_no
getPerCentDefNo_U   (Add1    def_no _  )					   = Just def_no
getPerCentDefNo_U   (Add2    def_no _ _)					   = Just def_no
getPerCentDefNo_U   (With1   def_no _  )					   = Just def_no
getPerCentDefNo_U   (With2   def_no _ _)					   = Just def_no
getPerCentDefNo_U   _   									   = Nothing


getPerCentDefNo   :: AldorExpr -> DefNo
getPerCentDefNo      expr      =
	case getPerCentDefNo_U expr of
		Just def_no -> def_no
		Nothing		-> error2 "getPerCentDefNo"
				       ("Not an ADT- or Category-valued expr: "++(show expr))


---------------------------
--	"setPerCentDefNo*":  --
---------------------------
--	Set DefNo in ADT and Category Type Literals ("_T") and "Add*", "With*"
--		exprs ("_E"):
--
setPerCentDefNo_T:: AldorType -> DefNo -> AldorType

setPerCentDefNo_T  (ADT _ rep_fb dt) nw_no = ADT       nw_no rep_fb dt
setPerCentDefNo_T  (CategoryT  _ dt) nw_no = CategoryT nw_no        dt

setPerCentDefNo_T  hmm _ =
	error2 "setPerCentDefNo_T" ("Invalid object: "++(show hmm))


setPerCentDefNo_E:: AldorExpr -> DefNo -> AldorExpr

setPerCentDefNo_E  (Add1  _ fb)      nw_no  =  Add1  nw_no fb
setPerCentDefNo_E  (Add2  _ fb1 fb2) nw_no  =  Add2  nw_no fb1 fb2
setPerCentDefNo_E  (With1 _ fb)		 nw_no  =  With1 nw_no fb
setPerCentDefNo_E  (With2 _ fb1 fb2) nw_no  =  With2 nw_no fb1 fb2

setPerCentDefNo_E  hmm _ =
	error2 "setPerCentDefNo_E" ("Invalid expr: "++(show hmm))


-------------------
--	"mergeDTs":  --
-------------------
mergeDTs :: DefTable -> DefTable -> DefTable

mergeDTs    dt1         dt2      =  dt4
	where
	--	Traverse "dt2" and insert its entries into "dt1":
	dt3  :: DefTable
	dt3  =  foldDT (\ currDT fdte -> attachFDTE fdte currDT) dt1 dt2

	--	Merge the Imports of "dt2" into "dt3":
	dt4  :: DefTable
	dt4  =  foldl  attachImport dt3 (getImports dt2)



----------------
--	"error2": --
----------------
error2 :: String -> String -> a
error2 =  errorM modName

