-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--								  "Main.hs":								 --
--	  The top-level module for Aldor-- Evaluation and Type-Checking Engine   --
--						  L.A.Timochouk, 2000--2002							 --
-- ========================================================================= --
module Main (main) where

import DataTypes
import ParseAST
import DecorFB
import NameBinds
import TypeCheckN
import MkEnvs

import IO
import IOExts
import System
import GetOpt

-----------------------------
--	Command-Line Options:  --
-----------------------------
arg_descrs:: [OptDescr CmdL_Arg]

arg_descrs = [Option [] ["debug"]
					 (NoArg Debug_Flag)
					 "Debugging mode",

			  Option [] ["no-eval-cache"]
					 (NoArg NoCache_Flag)
					 "Don't use internal caching of evaluation results",

			  Option [] ["strong-eval-tc"]
					 (NoArg Strong_EvalTC_Flag)
					 "Apply strong type-checking during evaluation",

			  Option [] ["max-eval-depth"]
					 (ReqArg (\s -> Max_Eval_Depth (read s)) "POSITIVE_INT")
					 "Maximum evaluation stack depth",

			  Option [] ["max-ade-depth"]
					 (ReqArg (\s -> Max_ADE_Depth  (read s)) "POSITIVE_INT")
					 "Maximum depth of Alpha,Delta-Equivalence checks"
			 ]

---------------------
--	"parse_args":  --
---------------------
parse_args :: [String] -> [CmdL_Arg]
parse_args    argv     =
	case getOpt RequireOrder arg_descrs argv of
		(pargs, [ast_file], [])->
			(AST_File ast_file) : pargs

		(_, unparsed, errors)  ->
			trace ("Invalid command-line arg(s) specified:\n"++
				  (concat unparsed)++"\n"++ (concat errors)  ++"\n")

				  (error (usageInfo "PARAMETERS: [Options...] AST_FILE"
				  arg_descrs))

--------------------
--	"parse_AST":  --
--------------------
parse_AST :: [CmdL_Arg] -> IO AldorFBlock

parse_AST    pargs =
	case [ast_file | AST_File ast_file <- pargs] of
		 []			-> error "No AST file specified"

		 [ast_file] -> do
						 ast_fd <- openFile  ast_file ReadMode
						 ast_str<- hGetContents ast_fd
						 -- DON'T CLOSE the file -- its contents is a LAZY list,
						 --	  so if we close it now, the list will be truncated
						 --	  when we use it later!

						 -- NB: Fine type inference here ("read" is overloaded,
						 --		but we know its required result type!)
						 return (mkFBlock (read ast_str))

		 _ -> error "Multiple AST files specified"

---------------
--	"main":  --
---------------
main:: IO()
main=  do
	--
	--	Get the command-line parms. The only non-optional parm is the AST file:
	--	argv :: [String]
	argv <- getArgs

	--	parsed_args:: ([CmdL_Arg], [NoMatch_String], [ErrMsg_String])
	parsed_args <- return (parse_args argv)

	--
	--	Parse the AST:
	--
	fb0   <- parse_AST parsed_args

	putStrLn ("\n")
	putStrLn ("***************************************")
	putStrLn ("*** Top-Level FBlock (undecorated): ***")
	putStrLn ("***************************************")

	putStrLn (show fb0)

	--
	--	Make the Envs:
	--
	putStrLn ("************************************")
	putStrLn ("*** Top-Level FBlock (decorated):***")
	putStrLn ("************************************")

	(fb1, enums1) <- return (decorFBlock fb0 (0,0,0))
	putStrLn (show fb1)

	putStrLn ("************************")
	putStrLn ("*** Integrity check: ***")
	putStrLn ("************************")

	putStrLn (show (ckFBconn (fb1) 0))

	putStrLn ("************************")
	putStrLn ("*** Name   Bindings: ***")
	putStrLn ("************************")

	binds1 <- return (mkBinds  fb1 enums1)
	tcenv  <- return (mkETC binds1 parsed_args)
	putStrLn (show binds1)

	putStrLn ("***************************")
	putStrLn ("*** Root FB Type-Check: ***")
	putStrLn ("***************************")

	tcrs  <- return (typeCheckFB_N tcenv fb1)
	putStrLn (show tcrs)

