-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--								  "ParseAST.hs":							 --
--				           Construction of FBs from ASTs					 --
--						     L.A.Timochouk, 2000--2001						 --
-- ========================================================================= --
module ParseAST
	   (
		  mkFBlock,
		  voidFB, simpleFB, nestedFB, fparmFB, mkADT_fromFB, mkADT_fromDT
	   )
where

import AldorAST
import DataTypes
import FButils
import qualified BuiltIns

modName::  String
modName = "ParseAST"

-- ================================================= --
-- Translation of the Aldor AST into the AldorBlock: --
-- ================================================= --
------------------------------------------
-- Overview of the Translation Process: --
-----------------------------------------
--	At the top level, a NestedFB with empty DefTable and the trivial sub-FBlock
--		(SimpleFB VoidExpr) is created when "mkFBlock" invokes "mkFBlock_R" 1st
--		time. "mkFBlock_R" carries the FBlock under construction thru'  all AST
--		nodes, modifying the former according the AST nodes met.   For "Define"
--		nodes, the DefTable of the FBlock is modified; for "Where" nodes,  both
--		the DefTable and the sub-FBlock are modified; for other nodes, the sub-
--		FBlock is modified. Modification of the sub-FBlock is essentially over-
--		writing it, and can only be done if that block was originally "SimpleFB
--		VoidExpr" (the function "toDefTable" checks this condition),      as we
--		allow no more than 1 value-returning expression in each scope. When the
--		sub-block is over-written, it is replaced by a "NestedFB"   (usually by
--		calling "mkFBlock" recursively) if it can contain nested contexts  dir-
--		ectly  attached  to the current FBlock,  and by a "SimpleFB" otherwise.
--		When all sub-blocks are constructed, "mkFBlock" performs some  term re-
--		writing to replace, wherever possible, "NestedFB"s with empty DefTables
--		by "SimpleFB"s.

--	Internal recursive AldorBlock generator: carries a (possible) hierarchy  of
--		FBlocks under construction along the way:
--
mkFBlock_R  :: AldorFBlock -> AST -> AldorFBlock


----------------------------------------
--	Access to AldorFBlock components: --
----------------------------------------
isNestedFB :: AldorFBlock       -> Bool
isNestedFB   (NestedFB _ _ _ _) =  True
isNestedFB    _			        =  False

--	We'v made sure that "show" on FBs does not itself make use of any of the
--		following functions:
--
getNestedFB:: AldorFBlock 	    -> AldorFBlock
getNestedFB  (NestedFB _ _ fb _)=  fb
getNestedFB   fb = error2 "getNestedFB" ("Not a NestedFB:\n"++(show fb))

getDefTable:: AldorFBlock	    -> DefTable
getDefTable	 (NestedFB _ dt _ _)=  dt
getDefTable   fb = error2 "getDefTable" ("Not a NestedFB:\n"++(show fb))


------------------------------------------------------
--	Important trivial objects and simplified ctors: --
------------------------------------------------------
--	"simpleFB" constructs a "SimpleFB" with "VoidFB" uplink and "noNUMBER" FBNo
--		(from a given Expr)
--	"nestedFB" constructs a "NestedFB" with "VoidFB" uplink and "noNUMBER" FBNo
--		(from a DefTable and innerFB)
--	"fparmFB"  constructs an "FParmFB" with "VoidFB" uplink and "noNUMBER" FBNo
--		(from  a single GenFDTE)
--	"emptyNFB" is a NestedFB with empty DefTable and a VoidFB inner FBlock:

voidFB     :: AldorFBlock
voidFB     =  VoidFB      noNUMBER

simpleFB   :: AldorExpr-> AldorFBlock
simpleFB      expr     =  SimpleFB  noNUMBER expr voidFB

nestedFB   :: DefTable -> AldorFBlock ->   AldorFBlock
nestedFB      dt          innerFB     =
			  NestedFB    noNUMBER dt innerFB   voidFB

emptyNFB   :: AldorFBlock
emptyNFB   =  nestedFB    emptyDT  voidFB

fparmFB    :: FlatDTE  -> AldorFBlock
fparmFB       fdte     =  FParmFB  noNUMBER (singltnDT fdte) voidFB

-------------------
--	"mkFBlock":  --
-------------------
--  Externally-visible AldorFBlock  generator:
--  Generation of a new FBlock is always started from an empty NestedFB
--		(so there is indeed a way of attaching  nested structures to be
--		encountered by "mkFBlock_R"),  but when the latter returns, the
--		FBlock we got may be simplified:
-- We re-combine (in "simplifyFB") "NestedFB _ emptyDT fb _"  into just
--		"fb".
-- NB: We do not re-adjust the "uplink" ("parent") ref as  it is always
--		"Void" at this stage:
--
mkFBlock::  AST ->  AldorFBlock
mkFBlock    ast =   simplifyFB
					(mkFBlock_R  emptyNFB ast)
	where
	simplifyFB   :: AldorFBlock -> AldorFBlock
	simplifyFB fb
		| (isNestedFB fb) && (isEmptyDT (getDefTable fb)) = getNestedFB fb
		| otherwise	= fb

--
--	"attachFB": If the LHS ("root") is a NestedFB with VoidFB as the innerFB,
--		the latter will be over-written, i.e. replaced by the "child".  This
--		function is used by "mkFBlock_R" for building  hierarchical  FBlocks:
--
attachFB  :: AldorFBlock -> AldorFBlock -> AldorFBlock
attachFB 	(NestedFB _  dt (VoidFB _) _)  child = nestedFB dt child

attachFB  fb _ =  error2 "attachFB" ("Improper \"root\":\n"++(show fb))


-- ============================= --
--	"mkFBlock_R" implementation: --
-- ============================= --
--	"mkFBlockR" both implments and uses (recursively) "mkFBlock":

-------------------
--	"Empty" AST: --
-------------------
mkFBlock_R _ Null =  voidFB

--------------------------
--  "Sequence" of ASTs: --
--------------------------
--	Thread the FBlock under construction through generator calls for all nodes
--		in the sequence in question.    If any node(s) in the Sequence are not
--		translatable, the FBlock under construction is not modified by them:
--	NB: "Sequence" is folded into the resulting FBlock,  NOT translated into a
--		list of BLocks!
--
mkFBlock_R  fb (Sequence [])	      = fb
mkFBlock_R  fb (Sequence (ast: asts)) =
	mkFBlock_R (mkFBlock_R fb   ast ) (Sequence asts)

---------------
--	"Ident": --
---------------
mkFBlock_R  fb (Ident name) = attachFB fb (simpleFB (translIdent name))
	where
	translIdent:: String  ->  AldorExpr
	translIdent   nm
		--	For some reason, Aldor has two special Idents ("0" and "1") which
		--		are translated into Integer Literals rather than Identifiers:
		| nm == "0"	= Literal (IntegerVal 0)
		| nm == "1"	= Literal (IntegerVal 1)

		--	We also convert a number of other "well-known" Idents directly to
		--		Literals. Standard Type Literals come here:
		| nm == "Type"      = Literal (TypeVal TypeT)

		| nm == "Category"  = Literal (TypeVal TypeT)
		--	NB: "Category" itself (when it `appears as a type expr in the LHS
		--		of a definition) is just "TypeT",  but its RHS  (the Category
		--		value) would then become a "CategoryT" TypeVal Literal

		| nm == "Boolean"   = Literal (TypeVal BooleanT)

		| nm == "Integer"   = Literal (TypeVal IntegerT)

		| nm == "BLInteger" = Literal (TypeVal IntegerT)

		| nm == "Float"	    = Literal (TypeVal FloatT)

		| nm == "BMFloat"   = Literal (TypeVal FloatT)

		| nm == "String"    = Literal (TypeVal StringT)

		| nm == "BLString"  = Literal (TypeVal StringT)

		| nm == "Exit"		= Literal (TypeVal ExitT)

		--	Boolean Literals:
		| nm == "false"     = Literal (BooleanVal False)

		| nm == "true"      = Literal (BooleanVal True )

		--	Generic case: Create an Identifier (with a simple name):
		| otherwise		    = Identifier noNUMBER [nm]

-----------------
--	"Qualify": --
-----------------
--	In a qualified name, the LHS is an Ident and the RHS is a qualified name
--		or an Ident.
--	NB: the order of name components in Aldor and in our program is OPPOSITE:
--		we put the outer-most domain name first:
--
mkFBlock_R   fb (Qualify head_ast tail_ast) =
	attachFB fb (simpleFB (Identifier noNUMBER (tnames ++ [hname])))
	where
	head_fb =  mkFBlock head_ast

	hname ::   String
	hname   =  case head_fb of
		SimpleFB _ (Identifier _ [hname'] ) _ -> hname'
		_   -> error2 "mkFBlock_R--Qualify" ("Invalid LHS: "++ (show head_ast))

	tail_fb =  mkFBlock tail_ast

	tnames :: [String]
	tnames  =  case tail_fb of
		SimpleFB _ (Identifier _ tnames') _ -> tnames'
		_	-> error2 "mkFBlock_R--Qualify" ("Invalid RHS: "++ (show tail_ast))

------------------
--	"Literal"s: --
------------------
mkFBlock_R   fb (LitInt    int)     =
	attachFB fb (simpleFB (Literal (IntegerVal int)))

mkFBlock_R   fb (LitFloat  double)  =
	attachFB fb (simpleFB (Literal (FloatVal   double)))

mkFBlock_R   fb (LitString string)  =
	attachFB fb (simpleFB (Literal (StringVal  string)))

---------------
--	"Where": --
---------------
--	Create a nested context ("emptyNFB")  and fill in its DefTable by threading
--		it through "defs", and then over-writing its sub-block:
--
mkFBlock_R   fb (Where defs ast) = attachFB fb  new_fb
	where
	new_fb = mkFBlock_R  (mkFBlock_R  emptyNFB  defs)  ast

---------------
--	"Label": --
---------------
--	Currently all labels are just removed, and the corresp unlabelled FBlocks
--		translated:
mkFBlock_R fb (Label _    ast) = attachFB fb (mkFBlock ast)

-----------------------
--	"If" and "Test": --
-----------------------
mkFBlock_R 	 fb (If cond_expr then_expr else_expr) =
	attachFB fb (simpleFB
					(IfThenElse
					(mkFBlock cond_expr) (mkFBlock then_expr)
					(mkFBlock else_expr)))

--
--	"Test" means that its argument MUST be evaluated to Boolean, but we will
--		do it anyway if the context requires that (e.g. in If), so we simply
--		remove "Test" nodes:
--
mkFBlock_R	 fb (Test any_node) = mkFBlock_R fb any_node

--------------------------
--	Logical operations: --
--------------------------
--	In Aldor, these are separate AST node types rather than just function calls,
--		as AND / OR may take a list of arguments:
--
mkFBlock_R fb (And asts)
	| length list >=2  = attachFB fb (simpleFB (AndList list))
	| otherwise		   = error2 "mkFBlock_R--And" "Less than 2 arguments"
	where
	list :: [AldorFBlock]
	list  = map mkFBlock asts

mkFBlock_R fb (Or  asts)
	| length list >= 2 = attachFB fb (simpleFB (OrList  list))
	| otherwise		   = error2 "mkFBlock_R--Or " "Less than 2 arguments"
	where
	list :: [AldorFBlock]
	list  = map mkFBlock asts

-----------------------------
--	"Lambda" and "Comma":  --
-----------------------------
--	Contain a function prototype (even if it is already declared in the corresp
--		"Define" node -- but "Lambda" can also be stand-alone!) and a body:
--	So "Lambda" is translated into a "SimpleFB" with a "Literal LambdaVal":
--
mkFBlock_R   fb (Lambda (Comma fparms) retype  body) =
	attachFB fb
	(simpleFB   (Literal(LambdaVal (FunctionT parms' retype') body')))
	where
	parms' ::   [AldorFBlock]
	parms' =    translFParms fparms

	retype'::   AldorFBlock
	retype'=    mkFBlock retype

	body'  ::   AldorFBlock
	body'  =    mkFBlock body

mkFBlock_R	 _  (Lambda _ _ _) =
	error2 "mkFBlock_R--Lambda" "Missing Comma node"

----------------
--	"Apply":  --
----------------
--	It may represent   a Function Prototype,  a Ctor for a Built-In Type or a
--		general Function Application. And they all have different formats!
--
--	(1) 1-Parm Proto    : ["->",     Parm,  ReType]
--	(2) Multi-Parm Proto: ["->",     Comma, ReType]
--	Here Parm is a Declare or Define node (the latter if there are default args,
--		but we deliberately don't recognise them), and a Comma contains Parms:
--
mkFBlock_R   fb  (Apply [Ident "->", hmm, retype]) =
	attachFB fb  (simpleFB  (Literal (TypeVal
				 (FunctionT (translFParms parms) (mkFBlock retype)) )))
	where
	--	Check whether "hmm" is a Comma or a  single parm,
	--		to distinguish between 1-parm and multi-parms protos:
	parms =
		case hmm of
			Comma hmm_parms ->  hmm_parms
			hmm_single		-> [hmm_single]


--	All other cases: Function application: [FuncName, Args]
--	This also includes Built-In Type and Value Ctors. The Type Ctors cannot be
--		regarded as just built-in functions,   as their "args" are FBlocks and
--		not "AldorValue"s (as we have to support dependent  types).  The Value
--		Ctors  could be  polymorphic built-in functions  but we  treat them as
--		"AldorExpr"s to keep built-in functions simple.
--	The function (if not a known Type Ctor) being applied  is encapsulated  in
--		an FBlock,  as the function may be given by any expression  (e.g. with
--		nested context) returning:
--		(*) Lambda Literal;
--		(*) an Identifier  referring  to  a user-defined  function.
--
mkFBlock_R    fb (Apply (func : args)) = attachFB fb (simpleFB hmm)
	where
	hmm = case func of
		Ident "Record" ->
			-- This is a "Record" TYPE ctor:
			-- The record fields are translated exactly in the same
			--		way as function parms:
			--
			Literal (TypeVal (RecordT (translFParms args)))

		Ident "Union"  ->
			--	Similar to the "Record" type ctor:
			--
			Literal (TypeVal (UnionT  (translFParms args)))

		Ident "record" ->
			--  A Record Val (not type!) Ctor:
			--
			RecordCtor 		 (map mkFBlock args)

		Ident "bracket"->
			--	The same as "record":
			--
			RecordCtor		 (map mkFBlock args)

		Ident "union"  ->
			--	A Union Val (not type!) ctor:
			--	The AST must be of the form Ident == Val:
			let
				(sel, val) = case args of
					[Define s@(Ident _) v] -> (s, v)

					_  -> error2 "mkFBlock_R--Apply"
								("Invalid \"union\" args: "++(show args))
			in
				UnionCtor (getExpr (mkFBlock sel)) (mkFBlock val)

		Ident "case"  ->
			--	A UnionCase expr:
			let
				(union, sel) = case args of
					[u, s@(Ident _)] ->  (u, s)

					_ -> error2 "mkFBlock_R--Apply"
							   ("Invalid \"case\" args: "++(show args))
			in
				UnionCase (mkFBlock union) (getExpr (mkFBlock sel))

		_ ->
			--	The GENERAL CASE:
			--	User-Defined Function Application:
			--
			FunctionCall (mkFBlock func) (map mkFBlock args)

-----------------
--	"Define":  --
-----------------
--	Normally, a definition must contain both a declaration and a body. The only
--		case when a declaration is not required is RENAMING.  The RHS must then
--		translate to an Identifier.
--	(The reverse --  Dcls without Def bodies -- are only allowed in "FParmFB"s,
--		with the Body being actually a VoidFB).
--	A special exception is the "Rep" symbol with no declaration &  an arbitrary
--		RHS (translatable to a full FBlock, not just to an Expr,  so we have to
--		construct a "fictive" "GenFDTE" for it, not a "RenFDTE"):
--
mkFBlock_R (NestedFB _  dt fb _) (Define  (Ident   "Rep") rhs) =
	nestedFB   (attachFDTE dt    (GenFDTE noNUMBER "Rep"  VoidExpr  rhs_fb)) fb
	where
	rhs_fb =   mkFBlock rhs


mkFBlock_R (NestedFB _  dt fb _) (Define  (Ident   idL) rhs) =
	nestedFB   (attachFDTE dt    (RenFDTE noNUMBER idL  rhs_expr)) fb
	where
	--	Here the "rhs" MUST translate to an Identifier:
	rhs_expr = case mkFBlock rhs of
		SimpleFB _ expr @(Identifier _ _) _ -> expr
		_	   ->  error2 "mkFBlock_R--Define" ("Invalid RHS:\n"++(show rhs))


--	Definition with Declaration:
--	The definition body is processed recursively, the declaration is translated
--		by "translDcl", and the result is attached to the DefTable of the FBlock
--		under construction. An error will occur if "fb" has no DefTable (i.e. if
--		it is SimpleFB or VoidFB):
--
mkFBlock_R (NestedFB _   dt fb _) (Define dcl@(Declare _ _)  body) =
	nestedFB (attachFDTE dt (GenFDTE noNUMBER dcl_name def_type  def_body)) fb
	where
	dcl_name		    :: String
	dcl_type, def_type	:: AldorExpr
	def_body			:: AldorFBlock

	(dcl_name, dcl_type) = translDcl    dcl
	(def_type, def_body) = simplifyDef  dcl_name  dcl_type (mkFBlock body)

	--	Simplification of Definitions:
	--	In the (all-important) case of a function definition, the Def Body is a
	--		Lambda Literal.    The latter contains the function prototype which
	--		must be ALPHA(?)-EQUIVALENT to that contained in the function decla-
	--		ration. We cannot really check their equivalence at the AST parsing
	--		stage,  so we just verify that two parm lists have the same length.
	--		If they do,   the one FROM THE LAMBDA-LITERAL is  installed  as the
	--		function type  (NB: the one from the declaration may have different
	--		indets, and that would complicate creating name bindings):
	--
	simplifyDef:: String-> AldorExpr  -> AldorFBlock-> (AldorExpr,  AldorFBlock)
	--			  Name     DCL type	     DEF lambda?    ResType	    ResBody

	simplifyDef   dclName  -- "dclName" is for diagnostics only

			(Literal    (TypeVal (FunctionT fparmsL _ )))

			(SimpleFB _ (Literal (LambdaVal
						 		  lambda_type@(FunctionT fparmsR   _)
								  lambda_body)) _ )
			--
			--	Check that the parm lists have the same length:
			--		"fparmsL" are declaration parms:
			--		"fparmsR" are lambda      parms:
			--	NB: "def_type" is an AldorExpr,  not AldorType,
			--		so it's Literal TypeVal:
			--
			| (length fparmsL)  ==  (length  fparmsR) =
						   (Literal (TypeVal lambda_type), lambda_body)

			| otherwise  = error2 "mkFBlock_R--Define"
						   		  ("Declaration of \""++ dclName++
						   		  "\" does not match its body")
	--
	--	In all other cases, leave the type fb and the body unchanged:
	--
	simplifyDef _ any_type  any_body  = (any_type, any_body)

mkFBlock_R fb ast@(Define _ _) =
	error2	"mkFBlock_R--Define"
			("Invalid FB\n"++(show fb)++"\nor AST:\n"++(show ast))


------------------
--	"Declare":  --
------------------
--	Declaration alone:
--
mkFBlock_R (NestedFB _   dt fb _)  dcl@(Declare _ _)  =
	nestedFB (attachFDTE dt (GenFDTE noNUMBER dcl_name dcl_type voidFB)) fb
	where
	dcl_name		    :: String
	dcl_type   			:: AldorExpr
	(dcl_name, dcl_type) = translDcl    dcl


--------------------
--	"PretendTo":  --
--	"RestrictTo": --
--------------------
mkFBlock_R   fb (RestrictTo expr type_expr) =
	attachFB fb (simpleFB  (RestrictType (mkFBlock expr) (mkFBlock type_expr)))

mkFBlock_R   fb (PretendTo  expr type_expr) =
	attachFB fb (simpleFB  (ConvertTo    (mkFBlock expr) (mkFBlock type_expr)))

--------------
--	"Add":  --
--------------
--	Creation and extension of Domains (ADTs and Packages; but only ADTs are
--		currently supported.  Constructs either an ADT Literal, or an "Add1"
--		expr, or an "Add2" expr.
--	If possible, we do explicit ADT extensions here,   as it will be useful
--		for constructing NameBinds later:
--
mkFBlock_R   fb (Add lhs rhs) = attachFB fb addFB
	where
	addFB::  AldorFBlock
	addFB =  case lhs of
		ABNothing ->
			--	There is no LHS, so the RHS is either an ADT Literal,  or any
			--		expr evaluating (hopefully) to an ADT (the latter constr-
			--		aint is expresed by an "Add1" expr):
			--
			case mkADT_fromFB rhsFB of
					Just rhsADT ->  simpleFB (Literal (TypeVal rhsADT))
					_		 	->  simpleFB (Add1    noNUMBER rhsFB)

		_ ->
			--	LHS exists. If it is an ADT Literal, and the RHS is just a
			--		DefTable WITHOUT special entries,  then we merge these
			--		DefTables right here to construct a single ADT Literal.
			--		Otherwise, we construct an "Add" expr:
			let
				lhsFB:: AldorFBlock
				lhsFB = mkFBlock lhs
			in
				case mkADT_fromFB lhsFB of
					Just lhsADT ->
						--	OK, do we have a DT (but not necesserily an ADT)
						--		in the RHS?
						case getJustDT rhsFB of
							Just rhsDT ->
								-- Attempt to merge the LHS and the RHS into
								--		a single ADT. This will work only if
								--		the RHS is NOT and ADT, otherwise we
								--		need to check that their "Rep"s  are
								--		identical, and that requires evalua-
								--		tion:
								case mkADT_fromDT rhsDT of
									Just rhsADT ->
										-- We have to produce "Add" here:
										simpleFB
										(Add2	  noNUMBER
										(simpleFB (Literal (TypeVal lhsADT)))
										(simpleFB (Literal (TypeVal rhsADT))))

									_ ->
										--  RHS is NOT an ADT: Merge it into
										--		the LHS:
										simpleFB (Literal
												 (TypeVal
												 (extendADT_byDT lhsADT rhsDT)))
							_ ->
								--	The RHS is not even a DT: Produce "Add":
								simpleFB (Add2   noNUMBER
										 (simpleFB (Literal (TypeVal lhsADT)))
										 rhsFB)
					_ ->
						--	The LHS is not a ready ADT object:
						--	Produce the most general form of "Add":
						simpleFB (Add2 noNUMBER lhsFB rhsFB)
	--	Uff!
	--	"rhsFB" is required in all cases:
	rhsFB:: AldorFBlock
	rhsFB = mkFBlock rhs


---------------
--	"With":  --
---------------
--	This resembles "Add", but more is simple, since we don't need to look for
--		any specific symbols in the DTs which may occur here.     It produces
--		either a Category Type Literal, or a "With1" expr, or a "With2" expr.
--	If possible, we do explicit Category extensions here, since this will be
--		useful for constructing NameBinds later:
--
mkFBlock_R   fb (With lhs rhs) = attachFB fb catFB
	where
	catFB :: AldorFBlock
	catFB =  case lhs of
		ABNothing ->
			--	LHS is empty: new Category definition (if the RHS is a DT):
			case getJustDT rhsFB of
				Just rhsDT ->
					--	Yes, it's a Category object:
					simpleFB (Literal (TypeVal (CategoryT noNUMBER rhsDT)))
				_ ->
					--  Not a ready Category object:  leave it as is, and it
					--		should evaluate to a Category (the latter const-
					--		raint given by "With1"):
					simpleFB (With1 noNUMBER rhsFB)

		_ ->
			--	LHS is not empty: Category extension:
			let
				lhsFB :: AldorFBlock
				lhsFB =  mkFBlock lhs
			in
				case getJustDT lhsFB of
					Just lhsDT ->
						case getJustDT rhsFB of
							Just rhsDT ->
								-- Two DTs: merge them:
								simpleFB (Literal
										 (TypeVal (CategoryT   noNUMBER
												  (mergeDTs lhsDT rhsDT))))
							_ ->
								-- RHS is not a DT: produce "With2":
								simpleFB (With2
										  noNUMBER
										 	(simpleFB
										 	(Literal (TypeVal
												(CategoryT noNUMBER lhsDT))))
										  rhsFB)
					_ ->
						--	LHS is not a DT:
						--	The most general case of "With2":
						simpleFB (With2 noNUMBER lhsFB rhsFB)

	rhsFB :: AldorFBlock
	rhsFB =  mkFBlock rhs


----------------
--	"Import": --
----------------
--	Currently recognised in a limited form only. Results in updating the
--		Imports list of the current DefTable:
--
mkFBlock_R orig@(NestedFB _ dt fb _) (Import ABNothing dom) =
	--
	--	Here the AST "dom" can be  an Ident, a "Qualifi"ed expr, etc,  so we
	--		cannot directly extract its name (without translating it first).
	--		It can be translated into an Identifier (the only valid case), a
	--		Literal (e.g. <Boolean>), and so on. If it is an Identifier,  we
	--		check it against the "BuiltIns.dontImport" list;   if it is not,
	--		we temporarily replace it with an empty QName,  and ignore it as
	--		well (no error!):
	--
	if  not (domName `elem` BuiltIns.dontImport || domName == [])
	then   nestedFB (attachImport dt  domExpr) fb
	else   orig
	where
	domExpr  = getExpr (mkFBlock dom)
	domName  = case  domExpr of
		Identifier _ domName' -> domName'
		_					  -> []

---------------------------
--	Any other AST nodes: --
---------------------------
--	We ignore what we do not understand, and return the unmodified root "fb":
mkFBlock_R fb _ = fb


-- ================================ --
--	"Helper" translation functions: --
-- ================================ --
--------------------
--	"translDcl":  --
--------------------
translDcl:: AST -> (String, AldorExpr)

--	"lhs" of a Declare node is supposed to be a Name, and the  "rhs" is an expr
--		evaluating to AldorType (encapsulated into an FBlock):
--
translDcl  (Declare  lhs rhs)  = (name, type_expr)
	where
	name	  = getIdent lhs
	type_expr = getExpr  (mkFBlock rhs)

translDcl  (Define _ _) =
	error2 "translDcl" "Expected declaration, got definition (default arg?)"

translDcl  hmm= error2 "translDcl" ("Invalid node:\n"++(show hmm))


----------------------
--	"translFParms": --
----------------------
--	Used where function parms are expected (in Lambda or function declaration).
--		Parms are translated into a list of FParmFBs  -- empty parm lists  are
--		NOT allowed:
--
translFParms :: [AST] -> [AldorFBlock]
translFParms    parms =
	case parms of
		[] -> error2 "translFParms" "Empty parms list"

		_  -> [fparmFB   (GenFDTE noNUMBER parm_name  parm_type voidFB) |
					 	 (parm_name, parm_type) <-(map translDcl parms)]

-----------------
-- "getIdent": --
-----------------
--	Gets an Identifier if we know it's there.  It's a short-cut introduced for
--		efficiency only. A more consistent approach would be to use "mkFBlock"
--		and pattern-match on its contents to extract the Identifier's name:
--
getIdent:: AST -> String
getIdent  (Ident  name) = name
getIdent  hmm  =  error2 "getIdent" ("Not an Ident node:\n"++(show hmm))

-- ========================= --
--	ADT and Category Ctors:  --
-- ========================= --
----------------------
--	"mkADT_fromDT": --
--	"mkADT_fromFB": --
----------------------
--	Attempting to construct an ADT object from a DefTable or an FBlock:
--
mkADT_fromDT:: DefTable -> Maybe AldorType
mkADT_fromDT   dt =
	--	Does the "dt" really contain "%" and "Rep"?
	if  isADT  dt
	then
		--	ADT: Create an ADT type literal and wrap it into a SimpleFB:
		--	NB : The DefTable in this newly-constructed ADT does not yet
		--		contain the "%" self-reference:
		--	The "Rep" is removed from  the "domDT" and installed separately:
		let
			[GenDTE _ _ repFB] = lookupDT   dt "Rep"
			adtDT			   = removeDTEs dt "Rep"
		in
			Just (ADT noNUMBER repFB adtDT)
	else
		--	Not an ADT (a Package? -- but we currently don't support them):
		Nothing


mkADT_fromFB:: AldorFBlock -> Maybe AldorType
mkADT_fromFB   fb =
	--	Do we have a bare "dt" without an Inner FB in a NestedFB?
	case getJustDT fb of
		Just dt -> mkADT_fromDT dt
		_		-> Nothing

----------------
--	"isADT":  --
----------------
--	Determines whether a given DefTable represents an ADT (or Package). It
--		searches for "%" IDENTIFIERS (not Defs -- the "%" DTE is installed
--		later!) and a "Rep" DTE; they must be BOTH 
--		present (in an ADT), or none at all (in a Package):
--
isADT :: DefTable -> Bool
isADT  dt = (hasRepDTE dt) && (hasPerCentIdent dt)

--	"hasRepDTE": True iff the "dt" defines the symbol "Rep", and it is not
--		locally overloaded:
--
hasRepDTE:: DefTable -> Bool
hasRepDTE    dt      =  length (lookupDT dt "Rep") == 1

--	"hasPerCentIdent":  True iff a "%" Identifier occurs somewhere in this
--		"dt". We use "foldFB_Gen" here as the search starts from an FB but
--		concerns Exprs only.   Lazy result list is used here -- the search
--		stops as soon as its first element is encountered:
--
hasPerCentIdent :: DefTable -> Bool
hasPerCentIdent dt =
	case foldFB_Gen  passFB findPerCent (undef,[]) (nestedFB dt voidFB)  of
		(_, True:_) -> True
		_			-> False
	where
	findPerCent:: AldorFBlock -> AldorExpr-> [Bool]  -> [Bool]
	findPerCent   _ (Identifier _ ["%"]) _ = [True]  -- Got it!
	findPerCent   _	_			    hmm    = hmm	 -- No status change...


------------------------
--	"extendADT_byDT": --
------------------------
--	Pre-condition: "dt" does not contain the "Rep" entries; an error is
--		generated otherwise:
--
extendADT_byDT :: AldorType -> DefTable -> AldorType
extendADT_byDT    (ADT _ rep dt1)   dt2
	| hasRepDTE dt2 = error2 "extendADT_byDT"
						 ("RHS already has \"Rep\": "++(show dt2))
	| otherwise  = ADT noNUMBER rep (mergeDTs dt1 dt2)

extendADT_byDT     _    _   =  error2 "extendADT" "Not an ADT"


-- ========== --
--	"error2"  --
-- ========== --
error2:: String-> String -> a
error2 = errorM modName

