-- vim:ts=4:syntax=haskell
-- ==========================================================================--
--								   "PrPrint.hs":							 --
--			     Pretty-Printer for "FBlocks" and Related Types:			 --
--							L.A.Timochouk, 2000--2001						 --
-- ========================================================================= --
--  We use the "Pretty" module by J. Hughes and S. Peyton Jones. All our  types
--		which require pretty-printing are first converted into the "Doc" format
--		and then stringified according to the page layout specified:
--
module PrPrint
	  (
		dotcat -- Also exports "show" instances on FBlocks etc...
	  )
where

import DataTypes
import Pretty

modName  :: String
modName  =  "PrPrint"


class  PrPrintable a where
	toDoc:: a -> Doc

---------------------------
--	Some style settings: --
---------------------------
indent   :: Int
indent   =  4

lineLen  :: Int
lineLen  =  80

--	"Ribbons Per Line": Specifies the portion of "lineLen" (1/ribbons)  to be
--		filled when there is a choice between continuing on the same line and
--		skipping to a new one (???) 1.0 coeff means that we always try to use
--		the whole available lineLen:
ribbons  :: Float
ribbons  =  1.0

debugMode:: Bool
debugMode=  True

-------------------------------
--	PrPrinting "AldorType"s: --
-------------------------------
instance PrPrintable AldorType  where

	toDoc TrivT    = text "<Triv>"
	toDoc ExitT    = text "<Exit>"
	toDoc TypeT    = text "<Type>"
	toDoc BooleanT = text "<Boolean>"
	toDoc IntegerT = text "<Integer>"
	toDoc FloatT   = text "<Float>"
	toDoc StringT  = text "<String>"

	toDoc (FunctionT fparms retype)  =
		sep [parens  dparms, text "->", toDoc retype]
		where
		dparms =  nest indent (sep (punctuate comma (map toDoc fparms)))
		-- "map" prints parms in the same order as they are defined!

	toDoc (BuiltInT  mb_fparms mb_retype)  =
		sep [parens  dparms, text "->", dretype]
		where
		--  Simplified form of "FunctionT" printing:
		dparms = sep   (punctuate comma (map toDoc mb_fparms))
		dretype= toDoc mb_retype

	toDoc (RecordT fields) =
		sep [text "Record", parens  (sep (punctuate comma (map toDoc fields)))]

	toDoc (UnionT  fields) =
		sep [text "Union",  parens  (sep (punctuate comma (map toDoc fields)))]

	toDoc    (ADT  def_no rep_fb dt) =
		vcat [sep [text ("ADT"++str_no),
			       parens (sep [text "Rep ==", toDoc rep_fb])],
			 text "",
			 char '{',  nest indent (toDoc dt), char '}']
		where
		str_no = if  def_no /= noNUMBER then '#':(show def_no) else ""

	toDoc    (CategoryT def_no dt) =
		vcat [text ("CATEGORY"++str_no), text "",
			 char '{',  nest indent (toDoc dt), char '}']
		where
		str_no = if  def_no /= noNUMBER then '#':(show def_no) else ""


-------------------------------
--	PrPrinting "AldorExpr"s: --
-------------------------------
instance PrPrintable AldorExpr  where

	toDoc  VoidExpr			    	= text  "<VoidExpr>"

	toDoc (Literal            val)	= toDoc val

	toDoc (Identifier use_no  idn)	=
		text  ("\""++(dotcat  idn)++"\""++str_no)
		where
		str_no = if  use_no == noNUMBER then "" else '@':(show use_no)

	toDoc (AndList	 ands) =
		sep ((text "AND") : lbrack :
			 (punctuate comma (map  toDoc ands)) ++ [rbrack])

	toDoc (OrList    ors )  =
		sep ((text "OR" ) : lbrack :
			 (punctuate comma (map  toDoc ors )) ++ [rbrack])

	toDoc (IfThenElse cond_fb then_fb  else_fb) =
		sep [text "IF", toDoc cond_fb, text "THEN", nest indent (toDoc then_fb),
									   text "ELSE", nest indent (toDoc else_fb)]

	toDoc (FunctionCall func args) = sep [toDoc func, parens dargs]
		where
		dargs = nest indent (sep (punctuate comma (map toDoc args)))

	toDoc (RecordCtor   fields)    =
		sep [text "record", brackets (sep (map toDoc fields))]

	toDoc (UnionCtor   sel val)	   =
		sep [text "union",  brackets (sep [toDoc sel, text "==", toDoc val])]

	toDoc (UnionCase union sel)    =
		sep [text "case",   brackets (sep [toDoc union, toDoc sel])]

	toDoc (RestrictType expr to_type) =
		sep [toDoc expr, text "\"@\"",   toDoc to_type]

	toDoc (ConvertTo    expr to_type) =
		sep [parens (toDoc expr), text "CONVERT_TO", parens (toDoc to_type)]

	toDoc (Add1  def_no fb) = sep [text ("ADD1"++str_no),  parens (toDoc fb)]
		where
		str_no = if def_no /= noNUMBER then '#':(show def_no) else ""

	toDoc (Add2  def_no fb1 fb2) =
		sep [text ("ADD2"++str_no),  parens
						   (sep (punctuate comma (map toDoc [fb1, fb2])))]
		where
		str_no = if def_no /= noNUMBER then '#':(show def_no) else ""

	toDoc (With1 def_no fb)	= sep [text ("WITH1"++str_no), parens (toDoc fb)]
		where
		str_no = if def_no /= noNUMBER then '#':(show def_no) else ""

	toDoc (With2 def_no fb1 fb2)=
		sep [text ("WITH2"++str_no), parens
						   (sep (punctuate comma (map toDoc [fb1, fb2])))]
		where
		str_no = if def_no /= noNUMBER then '#':(show def_no) else ""


----------------
--	"dotcat": --
----------------
--	Dot-concatenate components of a fully-qualified name:
dotcat :: [String] -> String
dotcat    [ ]	   =  error2  "dotcat" "Nothing to concatenate"
dotcat    [c]      =  c
dotcat    (c:cs)   =  c++"."++(dotcat cs)

--------------------------------
--	PrPrinting "AldorValue"s: --
--------------------------------
instance PrPrintable AldorValue where

	toDoc  TrivVal			=   text "<TrivVal>"
	toDoc (TypeVal    t)	=   toDoc t
	toDoc (BooleanVal b)	=   text ("<"++(show b)++">")
	toDoc (IntegerVal i)	=   text ("<"++(show i)++">")
	toDoc (FloatVal   d)    =   text ("<"++(show d)++">")
	toDoc (StringVal  s)	=   text ("<\""++s++"\">");

	toDoc (LambdaVal  proto  body)   = sep
		  [text "LAMBDA", toDoc proto, text "==", toDoc body]

	toDoc (BuiltInOp  opcode proto _)= sep
		  [text("BI"++opcode++" : " ), toDoc proto]

	toDoc (RecordVal  fields)   =
		brackets (sep (punctuate comma (map toDoc fields)))

	toDoc (UnionVal   use_no val) =
		sep [text "union",
			 brackets (sep [text ((show use_no) ++" => "), toDoc val])]

	toDoc (ADT_Val    _      val) =
		--	There is currently no way of outputting the ADT itself here (it
		--		would not be a good idea to print out the whole FBlock):
		sep [text ("ADT_Val"), toDoc val]

-----------------------------
--	PrPrinting "FlatDTE"s: --
-----------------------------
instance PrPrintable  FlatDTE where

	toDoc fdte =
		sep [hsep  [text  node_type, text ("\""++name++"\""++str_no), char ':'],
			 nest  indent (sep def_lst)]
		where
		node_type  :: String
		node_type  =  case fdte of
			GenFDTE _ _ _ (VoidFB _) -> "DECLARE"
			_						 -> "DEFINE"

		def_no  :: DefNo
		def_no = case fdte of
			GenFDTE def_no' _ _ _ -> def_no'
			RenFDTE def_no' _ _   -> def_no'

		name    :: String
		name   = case fdte of
			GenFDTE _ name' _ _   -> name'
			RenFDTE _ name' _     -> name'

		def_lst :: [Doc]
		def_lst= case fdte of
			GenFDTE  _  _  type_expr body_fb ->
				[toDoc type_expr, text "==", toDoc body_fb]

			RenFDTE _ _ ren_expr -> [text "==", toDoc ren_expr]


		str_no   = if  def_no == noNUMBER then "" else '#':(show def_no)

-------------------------
--	PrPrinting "DTE"s: --
-------------------------
--	Like "FlatDTE"s, but without a name:
--
instance PrPrintable DefTableEntry where

	toDoc (GenDTE  def_no type_expr body) =
		toDoc (GenFDTE def_no "UNSPECIFIED_NAME" type_expr body)

	toDoc (RenDTE  def_no ident_expr)     =
		toDoc (RenFDTE def_no "UNSPECIFIED_NAME" ident_expr)

------------------------------
--	PrPrinting "DefTable"s: --
------------------------------
instance PrPrintable DefTable    where
	--
	--	Inside "vcat", we construct the list of Doc objects for all
	--		"flattened" elements of the DefTable:
	--
	toDoc dt  = vcat (punctuate (char '\n') allDocs)
		where
		allDocs   :: [Doc]
		allDocs=  if   length dimports == 0
				  then bodyDocs
				  else sep [text "IMPORT",
							nest indent (sep (punctuate comma dimports)) ] :
					   bodyDocs

		dimports  :: [Doc]
		dimports  =  map toDoc (getImports dt)

		bodyDocs  :: [Doc]
		bodyDocs  =  foldDT  fdteToDocs [] dt

		fdteToDocs:: FlatDTE -> [Doc] -> [Doc]
		fdteToDocs   fdte  docs		  = (toDoc fdte) : docs


--------------------------
--	PrPrinting FBlocks: --
--------------------------
instance PrPrintable AldorFBlock where

	toDoc (VoidFB fb_no)  = text ("<VoidFB>"++strno)
		where
		strno = if fb_no == noNUMBER then "" else show fb_no


	toDoc (SimpleFB  fb_no expr _)
		| debugMode = sep [text ('{':strno), dexpr, text (strno++"}")]
		| otherwise = dexpr
		where
		strno = if fb_no == noNUMBER then "" else show fb_no
		dexpr = nest indent (toDoc  expr)


	toDoc (NestedFB fb_no dt  innerFB _)
		|   isEmptyDT     dt  =
			sep  [text ('{':strno), toDoc innerFB, text (strno++"}")]

		|	otherwise	     =
			vcat [text ('{':strno), defs_and_body, text (strno++"}")]
		where
		defs_and_body = nest indent
						(vcat [toDoc dt, text "",
							  sep   [text "RETURN", toDoc innerFB]])

		strno		  = if fb_no == noNUMBER then "" else show fb_no


	toDoc thisFB@(FParmFB fb_no _ _)
		| debugMode = sep [text ('{':sfbno), dparm, text (sfbno++"}")]
		| otherwise = dparm
		where
		sfbno = if fb_no == noNUMBER then "" else show fb_no

		GenFDTE    def_no   name  type_fb  body_fb = getFParm thisFB

		sdfno = if def_no == noNUMBER then "" else '#':(show def_no)

		dparm = sep  [text ("\""++name++"\""++(sdfno)++" : "),
                     toDoc  type_fb, dbody]

		dbody = case body_fb of
			VoidFB _ -> empty
			--	NB: We don;t print empty bodies, even if they have "FBNo"s!
			_	     -> sep [text "==",   toDoc body_fb]

--------------------------------------------------------------
--	We can now define "show" on Types, Exprs, ..., FBlocks: --
--------------------------------------------------------------
instance Show  AldorValue    where
	show = rendStr . toDoc

instance Show  AldorType     where
	show = rendStr . toDoc

instance Show  AldorExpr     where
	show = rendStr . toDoc

instance Show  FlatDTE       where
	show = rendStr . toDoc

instance Show  DefTableEntry where
	show = rendStr . toDoc

instance Show  DefTable      where
	show = rendStr . toDoc

instance Show  AldorFBlock   where
	show = rendStr . toDoc

--
--	We need to provide our own renderer, as the standard one (Pretty.rend)
--		is not flexible enough:
--
rendStr :: Doc -> String
rendStr =  fullRender PageMode lineLen ribbons string_txt ""
--
--	"string_text" is the same as in the Pretty module  (duplicated here as
--		it is not exported):
--
string_txt:: TextDetails -> String -> String
string_txt   (Chr c)   s  = c:s
string_txt   (Str s1)  s2 = s1 ++ s2
string_txt   (PStr s1) s2 = s1 ++ s2


-- =========== --
--  "error2":  --
-- =========== --
error2:: String -> String -> a
error2 = errorM modName

