-- vim:ts=4:syntax=haskell
-- ========================================================================= --
--							   "TypeCheckLib.hs":							 --
--	Principal Data Types and Utility Functions for the Type-Checking Engine  --
--						    L.A.Timochouk, 2001--2002						 --
-- ========================================================================= --
module TypeCheckLib
	   (
			er2tcr,    fb2tcr,    expr2tcr,   val2tcr, type2tcr,

			maybeType, maybeExpr, setType, setTypeTCR, okWith,

			inferRecordT,   inferUnionT,
			getFParmEType,  getTargEType,  getDefEType,
			findTargFields,

			TorE   (..)
	   )

where
import DataTypes
import BuiltIns
import ParseAST
import FButils
import NameBinds
import AlphaDelta
import IOExts

--	We must know precisely where calls to the Evaluation Engine occur
--		(to prevent such calls if further evaluation is not allowed),
--		so use explicit qualification of Eval* functions:
import EvalLib
import qualified EvalX

--	"trace" for warnings:
import qualified IOExts

modName :: String
modName =  "TypeCheckLib"


----------------
--	"er2tcr": --
----------------
--	Converts an "EvalRes" (e.g. for a type expr) into a "TC_Res". Since the
--		latter type cannot incorporate FBlocks (exprs only), for NestedFBs,
--		we do recursive evaluation and drop the DefTables-- but they should
--		not be necessary for representing results of types evaluation, any-
--		way. The value of "EvalRes"   is encoded as "Just" inferred type of
--		the "TC_Res" to be constructed:
--
er2tcr:: EvalRes -> TC_Res

er2tcr  (Bottom_ER  binds    )   =  Bottom_TCR     binds
er2tcr  (Partial_ER binds fb )   =
	case fb of
		SimpleFB  _ expr _	     -> OK_TCR  binds  [Just expr]
		NestedFB  _ _ inner_fb _ -> er2tcr (Partial_ER binds inner_fb)
		other					 -> error2 "er2tcr"
										  ("Unexpected FBlock:\n"++(show other))

er2tcr  (Full_ER    binds val)   =  OK_TCR  binds  [Just (Literal val)]


------------------
--	"fb2tcr":   --
--	"expr2tcr": --
--	"val2tcr":  --
--	"type2tcr": --
------------------
--	Similar to "er2tcr" above. Produce "OK_TCR" with "Just" an inferred type
--		(unless an error occurs); that's the difference between eg "type2tcr"
--		and "okWith" (the latter can set the inferred type to Nothing if in-
--		ference is not required):
--
fb2tcr  :: ETC_Env -> AldorFBlock -> TC_Res
fb2tcr   tcenv     fb =
	OK_TCR {_bindsR = getBinds tcenv, _infType = [Just (getExpr fb)]}


expr2tcr:: ETC_Env -> AldorExpr   -> TC_Res
expr2tcr tcenv   expr =
	OK_TCR {_bindsR = getBinds tcenv, _infType = [Just expr]}


val2tcr :: ETC_Env -> AldorValue  -> TC_Res
val2tcr  tcenv   val  =
	OK_TCR {_bindsR = getBinds tcenv, _infType = [Just (Literal val)]}


type2tcr:: ETC_Env -> AldorType   -> TC_Res
type2tcr tcenv   t    =
	OK_TCR {_bindsR = getBinds tcenv, _infType = [Just (Literal (TypeVal t))]}


----------------------------
--	(-&&-) on "[TC_Res]": --
----------------------------
--	This function operates on a pair of MULTI-VERSIONED "TC_Res"s, so (=&&=)
--		can then be applied directly to vectors  of muti-versioned  "TC_Res"
--		WITHOUT the need of crossing all versions over first:
--
instance ANDable [TC_Res] where

	(-&&-)  tcrs1 tcrs2 = tcrs12
		where
		--	Cross the lists as just Vectors; in "ressX", the inner dim is of
		--		length 2, the outer dim is X-versions:
		--
		tcrsX :: [TCR_CL]
		tcrsX =  crossWithAttrs [tcrs1, tcrs2]

		--	Now "and" the inner lists in "ressX":
		--
		tcrs12:: [TC_Res]
		tcrs12 = map andTCRs tcrsX

		-----------------
		--	"andTCRs": --
		-----------------
		andTCRs :: TCR_CL -> TC_Res

		andTCRs    tcrcl  =
			case getCList tcrcl of
				--
				--	Two "OK"s:   yield an "OK" with inferred types stored in a
				--		combined list:
				--
				[OK_TCR _ inf_type1, OK_TCR _ inf_type2] ->
					OK_TCR  (getAttr tcrcl)  (inf_type1 ++ inf_type2)

				--	Remaining cases: two Bottoms, or a Bottom and an OK: yield
				--		a Bottom:
				[Bottom_TCR _, _] -> Bottom_TCR  (getAttr tcrcl)
				[_, Bottom_TCR _] -> Bottom_TCR  (getAttr tcrcl)

				--	Other cases: the list is not of length 2: impossible:
				--
				_ -> error2 "andTCRs" "Internal error"


-- =========================--
--	Setting Inferred Types: --
-- =========================--
--------------------------
--	"maybe[Type|Expr]": --
--------------------------
--	"maybeType":
--	Returns a full InferredType or "noTYPE", depending on whether the
--		"_inferTypes" flag is set in the environment:
--
maybeType :: ETC_Env -> AldorType -> InferredType

maybeType    tcenv     someType  =
	if   _inferTypes   tcenv
	then [Just (Literal (TypeVal someType))]
	else noTYPE

--
--	"maybeExpr":
--	The same as "maybeType", but takes a Type Expr:
--
maybeExpr :: ETC_Env -> AldorExpr -> InferredType

maybeExpr	 tcenv	   typeExpr  =
	if   _inferTypes   tcenv
	then [Just   typeExpr]
	else noTYPE


------------------
--	"setType*": --
------------------
--	Set the inferred type in the [TC_Res] (actually, only in "OK_TCR"s):
--
setType   :: InferredType  ->  [TC_Res]  -> [TC_Res]

setType inf_type res = map (setTypeTCR inf_type) res


setTypeTCR:: InferredType  ->   TC_Res   ->  TC_Res

setTypeTCR   inf_type 	(OK_TCR binds _) =   OK_TCR binds inf_type
setTypeTCR   _			other_tcr		 =   other_tcr


----------------
--	"okWith": --
----------------
--	For trivial checks which succeed and yield an inferred type (if necessary):
--
okWith :: ETC_Env -> AldorType -> [TC_Res]
okWith    tcenv     someType  =
	[OK_TCR {_bindsR = getBinds tcenv, _infType = maybeType tcenv someType}]


-- ===================================== --
--	Some Type Inference and Evaluation:  --
-- ===================================== --
----------------------
--	"inferRecordT": --
----------------------
--	Given a "TC_Res" with a LIST of types in the "_infType",   treat those types
--		as Record Fields Types and produce a "TC_Res" with a consolidated Record
--		inferred type. This function is typically applied to a result of (=&&=):
--
inferRecordT ::  ETC_Env -> Int ->  TC_Res -> TC_Res

inferRecordT     tcenv	  nFields  (OK_TCR binds inf_type) =

	if    length ftypes  == nFields  && _inferTypes tcenv
	then  OK_TCR binds  [Just  (Literal (TypeVal (RecordT ffbs)))]
	else  OK_TCR binds  noTYPE
	where
	ftypes   ::  [AldorExpr]
	ftypes   =   [ftype | Just ftype <- inf_type]

	-- XXX: "FParmFB"s in the inferred "RecordT" are UNDECORATED -- when
	--		should we decorate them?
	-- The field names are empty here,  --  this should not be a problem
	--		since these names will not be used anyway:
	--
	ffbs     :: [AldorFBlock]
	ffbs	 =  [fparmFB (GenFDTE noNUMBER "" type_expr voidFB)  |
											  type_expr <- ftypes]
inferRecordT _ _ other_tcr = other_tcr


----------------------
--	"inferUnionT":  --
----------------------
--	Returns a list of "consolidated" lists [UnionType_TCR, FieldType_TCR]
--		where the 2nd element corresponds to the field given by the UseNo
--		argument:
--
inferUnionT :: ETC_Env -> AldorExpr -> [ConsoList NameBinds TC_Res]

inferUnionT	   tcenv      (Identifier use_no _) = resUT
	where
	--	Get all "complex" types enclosing  "FParmFB"s to which "use_no" can be
	--		bound:
	t3s     :: [(BindTarget, AldorType, Int)]
	t3s     =  findTypesEnclFPTargs (getBinds tcenv) use_no

	--	Select "UnionT"s only in "tpairs",     and get the "FParmFB"s by their
	--		indices:
	--
	upairs  :: [(AldorType, AldorFBlock)]
	upairs  =  [(ut, fields !! index) | (_, ut@(UnionT fields), index) <- t3s]

	--	For each element of "upairs", evaluate the 1st component, and get and
	--		evaluate the type of the 2nd one.  The results are represented as
	--		TCRs, and they are crossed.
	--	However, no evaluation is done if the "_noEval" flag is set in "tcenv":
	--
	resUT   :: [ConsoList NameBinds TC_Res]
	resUT   =  concat  (map evalUnionT upairs)

	evalUnionT :: (AldorType, AldorFBlock)  -> [ConsoList NameBinds TC_Res]
	evalUnionT    (ut,   fp)  = crossWithAttrs [ut_tcrs, fp_tcrs]
		where
		ut_tcrs:: [TC_Res]
		ut_tcrs = if   _noEval tcenv
				  then -- No further evaluation of "ut":
					   [type2tcr tcenv ut]

				  else -- Do full evaluation of "ut" (it can make difference if
					   --	it's e.g. an ADT):
					   map er2tcr
						   (EvalX.evalExpr tcenv (Literal(TypeVal ut)))

		fp_tcrs:: [TC_Res]
		fp_tcrs = getFParmEType   tcenv fp
		--	"getFParmEType" takes care itself of whether to do evaluation


inferUnionT _ hmm = error2 "inferUnionT"
						  ("Not a Union selector: "++(show hmm))

-----------------------
--	"getFParmEType": --
-----------------------
--	Returns the EVALUATED type expr from an FParmFB; the result is represented
--		as [TC_Res], UNLESS the "_noEval" flag is set in the "tcenv":
--
getFParmEType :: ETC_Env -> AldorFBlock -> [TC_Res]

getFParmEType    tcenv  fb@(FParmFB _ _ _) = tcrs
	where
	parm_type :: AldorExpr
	GenFDTE  _ _ parm_type _ = getFParm fb    -- Always GenFDTE here!

	tcrs 	  :: [TC_Res]
	tcrs      =  if  _noEval tcenv
				 then -- No evaluation of "parm_type", just take what we got:
					 [expr2tcr tcenv parm_type]

				 else -- Do full evaluation of "parm_type" expr:
					 map  er2tcr  (EvalX.evalExpr tcenv  parm_type)
	--
	--	NB: We DO NOT check here  that exprs returned in "tcrs"  are valid
	--		type-forming ones; this will be done when checks are performed
	--		against them.

getFParmEType    _ _ = error2 "getFParmEType" "Not an FParmFB"


----------------------
--	"getTargEType": --
----------------------
--	Restrict bindings for the given Identifier' UseNo to the given Target,  and
--		get the evaluated type expression(s) of that Target. For that, a TC_Res
--		list is produced, and the types computed are returned as "_infType"s of
--		its elements.
--	However, no evaluation of Target's type is done if the "_noEval" flag is set
--		in the "tcenv".
--	NB: Like in "getFParmEType",  this function does not verify  that evaluated
--		type exprs to be returned are actually valid type-forming exprs;   this
--		will be done later when checks are performed against these targets:
--
getTargEType:: ETC_Env -> UseNo -> BindTarget -> [TC_Res]

--
--	(1) "DefTarg":
--		Evaluate  the Def's type  in the restricted  binds.
--		Unlike "EvalX.evalTarg", there are no special cases
--		(e.g. re-constructing Lambdas) here,  since we are
--		interested in the type only, not in the body:
--
getTargEType   tcenv  use_no  target@(DefTarg (GenFDTE def_no _ type_expr _))
	| _noEval  tcenv =
		--	No evaluation of "type_expr", just take what we got, but
		--	   with restricted binds:
		--
		[OK_TCR {_bindsR = binds', _infType= [Just type_expr]}]

	| otherwise =
		--	Do full evaluation of "type_expr".
		--	NB: "evalTypeExpr" performs caching of the evaluation results:
		--
		map er2tcr (EvalX.evalTypeExpr tcenv' def_no type_expr)
	where
	binds'::  NameBinds
	binds' =  modifyBinds (getBinds tcenv) use_no [target]

	tcenv'::  ETC_Env
	tcenv' =  setBinds binds' tcenv


getTargEType  _ _    (DefTarg ren_fdte) =
	error2 "getTargEType" ("RenDTE as a  DefTarg: "++(show ren_fdte))

--
--	(2) "FP_Targ":
--		Similar to "DefTarg":
--
getTargEType   tcenv  use_no  target@(FP_Targ fb@(FParmFB _ _ _))
	| _noEval  tcenv =
		--	No evaluation of "type_expr", just take what we got, but
		--	   with restricted binds:
		--
		[OK_TCR {_bindsR = binds', _infType= [Just type_expr]}]

	| otherwise =
		--  Do full evaluation of "type_expr".
		--  XXX: Caching of evaluation results is badly needed here!
		--
		map er2tcr (EvalX.evalTypeExpr tcenv' def_no type_expr)

	where
	type_expr ::  AldorExpr
	GenFDTE def_no _ type_expr _  = getFParm fb

	binds'    ::  NameBinds
	binds'    =   modifyBinds (getBinds tcenv) use_no [target]

	tcenv'    ::  ETC_Env
	tcenv'    =   setBinds binds' tcenv


getTargEType   _ _ (FP_Targ  hmm) =
	error2 "getTargEType" ("Not an FParmFB in FP_Targ: "++(show hmm))

--
--	(3) "PerCentTarg":
--		Don't evaluate it, return it as is:
--
getTargEType  tcenv  use_no  target@(PerCentTarg expr) =
	[OK_TCR {_bindsR = binds', _infType= [Just expr]}]
	where
	binds'    ::  NameBinds
	binds'    =   modifyBinds (getBinds tcenv) use_no [target]

--
--	(4) "BI_Targ":
--		Just look-up the corresp BI def, and return its type along with
--		restricted binds. No evaluation is done here in any case:
--
getTargEType  tcenv  use_no  target@(BI_Targ bi_no) =
	okWith    tcenv' bi_type
	where
	(_, bi_type, _) = biInfo  bi_no

	tcenv' :: ETC_Env
	tcenv'  = setBinds (modifyBinds (getBinds tcenv) use_no [target]) tcenv

--
--	(5) "RepTarg":
--		The type of an Identifier bound to a "RepTarg" must (ultimately) be
--		a "TypeT", although it may be a more specific (e.g. Category) type.
--		We DON'T check it here (but rather where the obtained type is used),
--		just evaluate the "rep_fb" (unless the "_noEval" flag is set):
--
getTargEType      tcenv  use_no target@(RepTarg rep_fb)
	| _noEval tcenv =
		-- Just extract the expr from "rep_fb". It it refers to any local defs,
		--	these refs will anyway be preserved through name binds:
		--
		[OK_TCR {_bindsR = binds', _infType= [Just (getExpr rep_fb)]}]

	| otherwise =
		--	Perform  full  evaluation  of "rep_fb":
		--
		map er2tcr (EvalX.evalFB tcenv' rep_fb)
	where
	binds'::  NameBinds
	binds' =  modifyBinds (getBinds tcenv) use_no [target]

	tcenv'::  ETC_Env
	tcenv' =  setBinds binds' tcenv

--
--	(6) "DynTarg":
--		Get the ultimate non-Dynamic Targets (this requires evaluation through
--		"openDynTarget"), and follow them recursively.
--		If evaluation is disallowed by the "_noEval" flag in "tcenv", the only
--		possible return value is []:
--
getTargEType  tcenv  use_no  target@(DynTarg _ _)
	| _noEval tcenv =
		debug_msg    tcenv
		"\nWARNING: \"getTargEType\" on DynTarg in NoEval Mode" []

	| otherwise		=
--		trace ("\nGETTING THE TYPE OF "++(show target)++":\nGOT "++
--			  (show ress)++"\n\n")
		ress
	where
	ress :: [TC_Res]
	ress =  concat (map   followTarget  ultims)

	--	XXX: do we need to restrict the binds of "use_no"  to just [target]
	--		before going ahead? -- At the end, they will be replaced by the
	--		generated "AddedTarg" anyway. Still do it for efficiency:
	--
	rbinds :: NameBinds
	rbinds =  modifyBinds (getBinds tcenv) use_no [target]

	tcenv' :: ETC_Env
	tcenv' =  setBinds rbinds tcenv

	--	"openDynTarget" is only called if "noEval" is NOT set:
	--
	ultims:: [OpenedDynTarg]
	ultims =  EvalLib.openDynTarget tcenv' use_no target

	followTarget :: OpenedDynTarg -> [TC_Res]

	followTarget   (OK_ODT ubinds ultim ) =
		getTargEType     (setBinds ubinds tcenv')   use_no ultim

	followTarget   (Bottom_ODT    ubinds) = [Bottom_TCR ubinds]

--
--	(7)	"CondTarg":
--		Should not occur at all, as "CondTarg"s are not returned by
--		"lookupBinds", and in any case, they do not carry any type info:
--
getTargEType  _ _ hmm@(CondTarg _) =
	error2 "getTargEType" ("CondTarg encountered: "++(show hmm))

--
--	(8) "AddedTarg":
--		Get what it contains,   BUT in the result produced, preserve the
--		original "AddedTarg" for "use_no", otherwise it will be silently
--		eliminated!
--
getTargEType tcenv use_no added@(AddedTarg t) =
	map   restoreAdded (getTargEType tcenv use_no t)
	where
	restoreAdded :: TC_Res -> TC_Res
	restoreAdded    tcr    =
		setAttr (modifyBinds (getAttr tcr) use_no [added]) tcr


---------------------
--	"getDefEType": --
---------------------
--	Like "getTargEType" on a "DefTarg" or "FP_Targ", but here we have no UseNo
--		to which this Definition is bound:
--
getDefEType :: ETC_Env -> DefTableEntry -> [TC_Res]

getDefEType    tcenv     (GenDTE def_no type_expr _)
	| _noEval  tcenv  =
		--	Just return "type_expr":
		--
		[expr2tcr tcenv  type_expr]

	| otherwise =
		--	Do full evaluation of "type_expr":
		--	NB: "evalTypeExpr" may perform caching of its results!
		--
		case xxx of
			[] -> trace ("\nFAILED TO EVALUATE: "++(show type_expr)++"\n") []
			ok -> map er2tcr ok
	where
	xxx = EvalX.evalTypeExpr tcenv def_no type_expr

getDefEType    _         (RenDTE _ _) =
	error2 "getDefEType" "FIXME -- RenDTEs here are not yet supported"


------------------------
--	"findTargFields": --
------------------------
--	Finds possible targets ("FP_Targ"s) for the given "UseNo" among the given
--		list of "FParmFB"s (which typically come from a "RecordT" or "UnionT"
--		type definition). There may be more than 1 target found  (since field
--		names can be overloaded -- XXX is it OK???),    so a list of possible
--		target sis returned, along with their corresp indices in the original
--		list of "fields":
--
findTargFields :: NameBinds -> UseNo -> [AldorFBlock] -> [(Int, AldorFBlock)]

findTargFields    binds		   use_no   fields		  =  ftargs
	where
	--	All possible FBNos of FParmFBs  to which the Selector "use_no" can be
	--		bound (they do not necessarily correspond to the given "fields"):
	--
	poss_fbnos    :: [FBNo]
	poss_fbnos    =  [fb_no | FP_Targ (FParmFB fb_no _ _) <-
									  lookupBinds binds use_no]

	--	Now select those of "poss_fbnos" which are really within the "fields":
	--
	ftargs :: [(Int, AldorFBlock)]
	ftargs =  findTargFields_R  ([], 0)  fields

	findTargFields_R:: ([(Int, AldorFBlock)], Int) ->
					    [AldorFBlock]    -> [(Int, AldorFBlock)]

	findTargFields_R   (curr, _) [] = curr

	findTargFields_R   (curr, idx) ((fld@(FParmFB fb_no _ _)) :   fld_tail) =
		if  fb_no `elem` poss_fbnos
		then
			findTargFields_R ((idx, fld):curr, idx+1) fld_tail
		else
			findTargFields_R (curr,			   idx+1) fld_tail

	findTargFields_R   _ (other_fb : _) =
		error2 "findTargFields_R"
			   ("Invalid Field Descriptor -- not an FParmFB: "
			   ++(show other_fb)++"\n")


-- ========= --
--	"TorE":  --
-- ========= --
--	Type or Expr:
--
data TorE =
	 T  AldorType
  |  E  AldorExpr

instance AD_Equatable  TorE   where
	adeq adenv (T t1) (T t2) = eqTypes adenv t1 t2
	adeq adenv (E e1) (E e2) = eqExprs adenv e1 e2

	--	We can also mix "T"s and "E"s:
	adeq adenv (T t1) (E e2) = eqExprs adenv (Literal (TypeVal t1)) e2
	adeq adenv (E e1) (T t2) = eqExprs adenv e1 (Literal (TypeVal t2))

	adeq _     _      _      = No_ADER

instance Show   TorE		  where
	show (T t) = show t
	show (E e) = show e


-- =========== --
--	"error2":  --
-- =========== --
error2 :: String -> String -> a
error2 =  errorM modName

